'use strict';

import {errorList} from './lib/errorHandler';
import {
    chatMessageVOTypes,
    inviteeVOidTypes,
    createThreadTypes,
    chatMessageTypes,
    systemMessageTypes,
    imageMimeTypes,
    imageExtentions,
    CHAT_ERRORS,
    SERVICES_PATH,
    chatStickerTypes,
    podspace
} from "./lib/constants";
import Async from 'podasync';
import Utility from './utility/utility.js';
import Mime from 'mime';
import App from "./lib/app";
import ChatMessaging from "./lib/messenger";
import buildConfig from "./buildConfig.json"
import ContactsV2 from "./lib/chat/contacts/contactsV2";

/*
 * Pod Chat Module
 * @module chat
 *
 * @param {Object} params
 */

function Chat(params) {
    const app = new App();

    app.contact = new ContactsV2(app);
    this.contact = {
        add: app.contact.add,
        remove: app.contact.remove,
        get: app.contact.get,
        update: app.contact.update
    };

    /**
     * Defining global variables for Dexie to work in Node ENV
     */

    /*******************************************************
     *          P R I V A T E   V A R I A B L E S          *
     *******************************************************/
    function getUserInfoParams(userInfo) {

        if (typeof userInfo === 'object') {
            if (["id", "name"].every(key => Object.keys(userInfo).includes(key))) {
                return userInfo
            }
        }
        if (userInfo != null) {
            console.error("[SDK] userInfo params is null or not Valid ");
        }
        return null
    }

    app.sdkParams.appId = params.appId || 'PodChat';
    app.sdkParams.deviceId = params.deviceId || undefined;
    app.sdkParams.token = params.token || "111";
    app.sdkParams.mapApiKey = params.mapApiKey || '8b77db18704aa646ee5aaea13e7370f4f88b9e8c';
    app.sdkParams.grantDeviceIdFromSSO = (params.grantDeviceIdFromSSO && typeof params.grantDeviceIdFromSSO === 'boolean')
        ? params.grantDeviceIdFromSSO
        : false;
    app.sdkParams.seenIntervalPitch = params.seenIntervalPitch || 2000;
    app.sdkParams.systemMessageIntervalPitch = params.systemMessageIntervalPitch || 1000;
    app.sdkParams.socketAddress = params.socketAddress;
    app.sdkParams.serverName = params.serverName || '';
    app.sdkParams.wsConnectionWaitTime = params.wsConnectionWaitTime;
    app.sdkParams.connectionRetryInterval = params.connectionRetryInterval;
    app.sdkParams.msgPriority = params.msgPriority || 1;
    app.sdkParams.messageTtl = params.messageTtl || 10000;
    app.sdkParams.reconnectOnClose = params.reconnectOnClose;
    app.sdkParams.asyncLogging = params.asyncLogging;
    app.sdkParams.connectionCheckTimeout = params.connectionCheckTimeout;
    app.sdkParams.httpRequestTimeout = (params.httpRequestTimeout >= 0) ? params.httpRequestTimeout : 30000;
    app.sdkParams.asyncRequestTimeout = (typeof params.asyncRequestTimeout === 'number' && params.asyncRequestTimeout >= 0) ? params.asyncRequestTimeout : 0;
    app.sdkParams.connectionCheckTimeoutThreshold = params.connectionCheckTimeoutThreshold;
    app.sdkParams.httpUploadRequestTimeout = (params.httpUploadRequestTimeout >= 0) ? params.httpUploadRequestTimeout : 0;
    app.sdkParams.actualTimingLog = (params.asyncLogging.actualTiming && typeof params.asyncLogging.actualTiming === 'boolean')
        ? params.asyncLogging.actualTiming
        : false;
    app.sdkParams.consoleLogging = (params.asyncLogging.consoleLogging && typeof params.asyncLogging.consoleLogging === 'boolean')
        ? params.asyncLogging.consoleLogging
        : false;
    app.sdkParams.fullResponseObject = params.fullResponseObject || false;
    app.sdkParams.webrtcConfig = (params.webrtcConfig ? params.webrtcConfig : null);
    app.sdkParams.chatPingMessageInterval = params.chatPingMessageInterval;
    app.sdkParams.protocol = params.protocol || 'websocket';
    app.sdkParams.callOptions = params.callOptions;

    app.sdkParams.generalTypeCode = params.typeCode || 'default';
    app.sdkParams.typeCodeOwnerId = params.typeCodeOwnerId || null;

    params.ssoHost && (app.sdkParams.SERVICE_ADDRESSES.SSO_ADDRESS = params.ssoHost);
    params.platformHost && (app.sdkParams.SERVICE_ADDRESSES.PLATFORM_ADDRESS = params.platformHost);
    params.fileServer && (app.sdkParams.SERVICE_ADDRESSES.FILESERVER_ADDRESS = params.fileServer);
    params.podSpaceFileServer && (app.sdkParams.SERVICE_ADDRESSES.PODSPACE_FILESERVER_ADDRESS = params.podSpaceFileServer);
    params.mapServer && (app.sdkParams.SERVICE_ADDRESSES.MAP_ADDRESS = params.mapServer);


    /**
     * Minimum value for messageTtl is 2000
     */
    if(app.sdkParams.messageTtl < 2000) {
        app.sdkParams.messageTtl = 2000;
        console.warn("[SDK] messageTtl can not be less than 2000, we changed it to 2000")
    }

    if(getUserInfoParams(params.userInfo)) {
        app.store.user.setUser(getUserInfoParams(params.userInfo))
    }

    let
        asyncClient,
        currentModuleInstance = this,
        peerId,
        oldPeerId,
        db,
        cacheSecret = '',
        isTypingInterval,
        queueServers = params.queueServers || [],
        queueReceive = params.queueReceive,
        queueSend = params.queueSend,
        queueConnectionTimeout = params.queueConnectionTimeout,
        sendPingTimeout,
        getUserInfoTimeout,
        config = {
            getHistoryCount: 50
        },
        getUserInfoRetry = 5,
        getUserInfoRetryCount = 0,
        asyncStateTypes = {
            0: 'CONNECTING',
            1: 'CONNECTED',
            2: 'CLOSING',
            3: 'CLOSED'
        },
        chatState = false,
        chatFullStateObject = {},
        minIntegerValue = Number.MAX_SAFE_INTEGER * -1,
        maxIntegerValue = Number.MAX_SAFE_INTEGER * 1,
        chatSendQueue = [],
        chatSendQueueHandlerTimeout,

        /** call variables */
        callTypes = {
            'VOICE': 0x0,
            'VIDEO': 0x1
        },
        currentCallParams = {},
        currentCallId = null,
        newCallId = null,
        callClientType = {
            WEB: 1,
            ANDROID: 2,
            DESKTOP: 3,
            NODE: 4
        },
        callRequestController = {
            callRequestReceived: false,
            callEstablishedInMySide: false,
            iCanAcceptTheCall: function () {
                return callRequestController.callRequestReceived && callRequestController.callEstablishedInMySide;
            }
        },
        callStopQueue = {
            callStarted: false,
        },
        callNoAnswerTimeout = params.callOptions && params.callOptions.callNoAnswerTimeout ? params.callOptions.callNoAnswerTimeout : 0,
        asyncLogCallback = typeof params.asyncLogCallback == "function" ? params.asyncLogCallback : null;

    app.messenger = new ChatMessaging(app, Object.assign(params, {
        asyncClient: asyncClient
    }));

    /*******************************************************
     *            P R I V A T E   M E T H O D S            *
     *******************************************************/

    var init = function () {
            /**
             * Initialize Cache Databases
             */
            // startCacheDatabases(function () {
            if (app.sdkParams.grantDeviceIdFromSSO) {
                var getDeviceIdWithTokenTime = new Date().getTime();
                getDeviceIdWithToken(function (retrievedDeviceId) {
                    if (app.sdkParams.actualTimingLog) {
                        Utility.chatStepLogger('Get Device ID ', new Date().getTime() - getDeviceIdWithTokenTime);
                    }

                    app.sdkParams.deviceId = retrievedDeviceId;

                    initAsync();
                });
            } else {
                initAsync();
            }
            // });
        },

        /**
         * Initialize Async
         *
         * Initializes Async module and sets proper callbacks
         *
         * @access private
         *
         * @return {undefined}
         * @return {undefined}
         */
        initAsync = function () {
            if (app.sdkParams.protocol === 'queue' && (!Array.isArray(queueServers) || queueServers.length == 0)) {
                console.log('Queue Servers are invalid!', queueServers);
                return;
            }
            var asyncGetReadyTime = new Date().getTime();

            if (typeof app.sdkParams.appId == "number") {
                app.sdkParams.appId = app.sdkParams.appId.toString()
            }
            if (typeof app.sdkParams.appId != "string" || app.sdkParams.appId.length > 30) {
                app.errorHandler.raiseError(errorList.INVALID_APP_ID, null, true, {});
                return;
            }

            asyncClient = new Async({
                appId: app.sdkParams.appId,
                protocol: app.sdkParams.protocol,
                servers: queueServers,
                queueReceive: queueReceive,
                queueSend: queueSend,
                queueConnectionTimeout: queueConnectionTimeout,
                socketAddress: app.sdkParams.socketAddress,
                serverName: app.sdkParams.serverName,
                deviceId: app.sdkParams.deviceId,
                wsConnectionWaitTime: app.sdkParams.wsConnectionWaitTime,
                connectionRetryInterval: app.sdkParams.connectionRetryInterval,
                connectionCheckTimeout: app.sdkParams.connectionCheckTimeout,
                connectionCheckTimeoutThreshold: app.sdkParams.connectionCheckTimeoutThreshold,
                messageTtl: app.sdkParams.messageTtl,
                reconnectOnClose: app.sdkParams.reconnectOnClose,
                asyncLogging: app.sdkParams.asyncLogging,
                logLevel: (app.sdkParams.consoleLogging ? 3 : 1),
                asyncLogCallback: asyncLogCallback || null,
            });

            app.messenger.asyncInitialized(asyncClient);

            asyncClient.on('asyncReady', function () {
                if (app.sdkParams.actualTimingLog) {
                    Utility.chatStepLogger('Async Connection ', new Date().getTime() - asyncGetReadyTime);
                }

                peerId = asyncClient.getPeerId();

                if (!app.store.user.get()) {
                    getUserAndUpdateSDKState();
                } else if (app.store.user.get().id > 0) {
                    app.messenger.chatState = true;
                    app.chatEvents.fireEvent('chatReady');
                    chatSendQueueHandler();
                }
            });

            asyncClient.on('stateChange', function (state) {
                app.chatEvents.fireEvent('chatState', state);
                chatFullStateObject = state;

                let st = app.sdkParams.protocol == 'queue' ? state.queueState : state.socketState;

                switch (st) {
                    case 1: // CONNECTED
                        if (
                            app.sdkParams.protocol != 'websocket'
                            || (app.sdkParams.protocol == 'websocket' && state.deviceRegister && state.serverRegister)
                        ) {
                            app.messenger.chatState = true;
                            app.messenger.startChatPing();
                        }
                        break;
                    case 0: // CONNECTING
                        app.messenger.chatState = false;
                        app.messenger.stopChatPing();
                        break;
                    case 2: // CLOSING
                    case 3: // CLOSED
                        app.messenger.chatState = false;
                        app.messenger.stopChatPing();
                        //chatMessaging.sendPingTimeout && clearTimeout(chatMessaging.sendPingTimeout);
                        break;
                }
            });

            asyncClient.on('connect', function (newPeerId) {
                asyncGetReadyTime = new Date().getTime();
                peerId = newPeerId;
                app.chatEvents.fireEvent('connect');
                ping();
            });

            asyncClient.on('disconnect', function (event) {
                oldPeerId = peerId;
                peerId = undefined;
                app.chatEvents.fireEvent('disconnect', event);
            });

            asyncClient.on('reconnect', function (newPeerId) {
                peerId = newPeerId;
                app.chatEvents.fireEvent('reconnect');
            });

            asyncClient.on('message', function (params, ack) {
                receivedAsyncMessageHandler(params);
                ack && ack();
            });

            asyncClient.on('error', function (error) {
                if (error.errorCode) {
                    app.chatEvents.fireEvent('error', {
                        code: error.errorCode,
                        message: error.errorMessage,
                        error: error.errorEvent
                    });
                } else {
                    app.chatEvents.fireEvent('error', {
                        code: 12003,
                        message: CHAT_ERRORS[12003],
                        error: error.errorEvent
                    });
                }
            });
        },
        getUserAndUpdateSDKState = function () {
            var getUserInfoTime = new Date().getTime();
            getUserInfo(function (userInfoResult) {
                if (app.sdkParams.actualTimingLog) {
                    Utility.chatStepLogger('Get User Info ', new Date().getTime() - getUserInfoTime);
                }
                if (!userInfoResult.hasError) {
                    // userInfo = userInfoResult.result.user;
                    app.store.user.setUser(userInfoResult.result.user);
                    // getAllThreads({
                    //     summary: true,
                    //     cache: false
                    // });

                    app.messenger.chatState = true;
                    app.chatEvents.fireEvent('chatReady');
                    chatSendQueueHandler();
                }
            });
        },
        /**
         * Get Device Id With Token
         *
         * If ssoGrantDevicesAddress set as TRUE, chat agent gets Device ID
         * from SSO server and passes it to Async Module
         *
         * @access private
         *
         * @param {function}  callback    The callback function to run after getting Device Id
         *
         * @return {undefined}
         */
        getDeviceIdWithToken = function (callback) {
            var deviceId;

            var params = {
                url: app.sdkParams.SERVICE_ADDRESSES.SSO_ADDRESS + SERVICES_PATH.SSO_DEVICES,
                method: 'GET',
                headers: {
                    'Authorization': 'Bearer ' + app.sdkParams.token
                }
            };

            app.httpRequest.httpRequest(params, function (result) {
                if (!result.hasError) {
                    var devices = JSON.parse(result.result.responseText).devices;
                    if (devices && devices.length > 0) {
                        for (var i = 0; i < devices.length; i++) {
                            if (devices[i].current) {
                                deviceId = devices[i].uid;
                                break;
                            }
                        }

                        if (!deviceId) {
                            app.chatEvents.fireEvent('error', {
                                code: 6000,
                                message: CHAT_ERRORS[6000],
                                error: null
                            });
                        } else {
                            callback(deviceId);
                        }
                    } else {
                        app.chatEvents.fireEvent('error', {
                            code: 6001,
                            message: CHAT_ERRORS[6001],
                            error: null
                        });
                    }
                } else {
                    app.chatEvents.fireEvent('error', {
                        code: result.errorCode,
                        message: result.errorMessage,
                        error: result
                    });
                }
            });
        },

        /**
         * Handshake with SSO to get user's keys
         *
         * In order to Encrypt and Decrypt cache we need a key.
         * We can retrieve encryption keys from SSO, all we
         * need to do is to do a handshake with SSO and
         * get the keys.
         *
         * @access private
         *
         * @param {function}  callback    The callback function to run after Generating Keys
         *
         * @return {undefined}
         */
        generateEncryptionKey = function (params, callback) {
            var data = {
                validity: 10 * 365 * 24 * 60 * 60, // 10 Years
                renew: false,
                keyAlgorithm: 'aes',
                keySize: 256
            };

            if (params) {
                if (params.keyAlgorithm != 'undefined') {
                    data.keyAlgorithm = params.keyAlgorithm;
                }

                if (parseInt(params.keySize) > 0) {
                    data.keySize = params.keySize;
                }
            }

            var httpRequestParams = {
                url: app.sdkParams.SERVICE_ADDRESSES.SSO_ADDRESS + SERVICES_PATH.SSO_GENERATE_KEY,
                method: 'POST',
                data: data,
                headers: {
                    'Authorization': 'Bearer ' + app.sdkParams.token
                }
            };

            app.httpRequest.httpRequest(httpRequestParams, function (result) {
                if (!result.hasError) {
                    try {
                        var response = JSON.parse(result.result.responseText);
                    } catch (e) {
                        console.log(e);
                    }
                } else {
                    app.chatEvents.fireEvent('error', {
                        code: result.error,
                        message: result.error_description,
                        error: result
                    });
                }
            });
        },

        /**
         * Get Encryption Keys by KeyId
         *
         * In order to Encrypt and Decrypt cache we need a key.
         * We can retrieve encryption keys from SSO by sending
         * KeyId to SSO and get related keys
         *
         * @access private
         *
         * @param {function}  callback    The callback function to run after getting Keys
         *
         * @return {undefined}
         */
        getEncryptionKey = function (params, callback) {
            var keyId;

            if (params) {
                if (params.keyId != 'undefined') {
                    keyId = params.keyId;

                    var httpRequestParams = {
                        url: app.sdkParams.SERVICE_ADDRESSES.SSO_ADDRESS + SERVICES_PATH.SSO_GET_KEY + keyId,
                        method: 'GET',
                        headers: {
                            'Authorization': 'Bearer ' + app.sdkParams.token
                        }
                    };

                    app.httpRequest.httpRequest(httpRequestParams, function (result) {
                        if (!result.hasError) {
                            try {
                                var response = JSON.parse(result.result.responseText);
                            } catch (e) {
                                console.log(e);
                            }

                            callback && callback({
                                hasError: false,
                                secretKey: response.secretKey
                            });
                        } else {
                            callback && callback({
                                hasError: true,
                                code: result.errorCode,
                                message: result.errorMessage
                            });

                            app.chatEvents.fireEvent('error', {
                                code: result.errorCode,
                                message: result.errorMessage,
                                error: result
                            });
                        }
                    });
                }
            }
        },


        /**
         * Get User Info
         *
         * This functions gets user info from chat serverName.
         * If info is not retrived the function will attemp
         * 5 more times to get info from erver
         *
         * @recursive
         * @access private
         *
         * @param {function}    callback    The callback function to call after
         *
         * @return {object} Instant function return
         */
        getUserInfo = function getUserInfoRecursive(callback) {
            getUserInfoRetryCount++;

            if (getUserInfoRetryCount > getUserInfoRetry) {
                getUserInfoTimeout && clearTimeout(getUserInfoTimeout);

                getUserInfoRetryCount = 0;

                app.chatEvents.fireEvent('error', {
                    code: 6101,
                    message: CHAT_ERRORS[6101],
                    error: null
                });
            } else {
                getUserInfoTimeout && clearTimeout(getUserInfoTimeout);

                getUserInfoTimeout = setTimeout(function () {
                    getUserInfoRecursive(callback);
                }, getUserInfoRetryCount * 10000);
                return app.messenger.sendMessage({
                    chatMessageVOType: chatMessageVOTypes.USER_INFO,
                    typeCode: params.typeCode
                }, {
                    onResult: function (result) {
                        var returnData = {
                            hasError: result.hasError,
                            cache: false,
                            errorMessage: result.errorMessage,
                            errorCode: result.errorCode
                        };

                        if (!returnData.hasError) {
                            getUserInfoTimeout && clearTimeout(getUserInfoTimeout);

                            var messageContent = result.result;
                            var currentUser = formatDataToMakeUser(messageContent);

                            let resultData = {
                                user: currentUser
                            };

                            returnData.result = resultData;
                            getUserInfoRetryCount = 0;

                            callback && callback(returnData);

                            /**
                             * Delete callback so if server pushes response
                             * before cache, cache won't send data again
                             */
                            callback = undefined;
                        }
                    }
                });
            }
        },

        sendSystemMessage = function (params) {
            return app.messenger.sendMessage({
                chatMessageVOType: chatMessageVOTypes.SYSTEM_MESSAGE,
                subjectId: params.threadId,
                content: params.content,
                uniqueId: params.uniqueId,
                pushMsgType: 3
            });
        },

        /**
         * Chat Send Message Queue Handler
         *
         * Whenever something pushes into cahtSendQueue
         * this function invokes and does the message
         * sending progress throught async
         *
         * @access private
         *
         * @return {undefined}
         */
        chatSendQueueHandler = function () {
            chatSendQueueHandlerTimeout && clearTimeout(chatSendQueueHandlerTimeout);
            if (chatSendQueue.length) {
                var messageToBeSend = chatSendQueue[0];

                /**
                 * Getting chatSendQueue from either cache or
                 * memory and scrolling through the send queue
                 * to send all the messages which are waiting
                 * for chatState to become TRUE
                 *
                 * There is a small possibility that a Message
                 * wouldn't make it through network, so it Will
                 * not reach chat server. To avoid losing those
                 * messages, we put a clone of every message
                 * in waitQ, and when ack of the message comes,
                 * we delete that message from waitQ. otherwise
                 * we assume that these messages have been failed to
                 * send and keep them to be either canceled or resent
                 * by user later. When user calls getHistory(), they
                 * will have failed messages alongside with typical
                 * messages history.
                 */
                if (app.messenger.chatState) {
                    getChatSendQueue(0, function (chatSendQueue) {
                        deleteFromChatSentQueue(messageToBeSend,
                            function () {
                                app.messenger.sendMessage(messageToBeSend.message, messageToBeSend.callbacks, function () {
                                    if (chatSendQueue.length) {
                                        chatSendQueueHandler();
                                    }
                                });
                            });
                    });
                }
            }
        },

        /**
         * Ping
         *
         * This Function sends ping message to keep user connected to
         * chat server and updates its status
         *
         * @access private
         *
         * @return {undefined}
         */
        ping = function () {
            if (app.messenger.chatState && app.store.user.get() !== undefined) {
                /**
                 * Ping messages should be sent ASAP, because
                 * we don't want to wait for send queue, we send them
                 * right through async from here
                 */
                app.messenger.sendMessage({
                    chatMessageVOType: chatMessageVOTypes.PING,
                    pushMsgType: 3
                });
            } else {
                sendPingTimeout && clearTimeout(sendPingTimeout);
            }
        },

        /**
         * Clear Cache
         *
         * Clears Async queue so that all the remained messages will be
         * ignored
         *
         * @access private
         *
         * @return {undefined}
         */
        clearChatServerCaches = function () {
            app.messenger.sendMessage({
                chatMessageVOType: chatMessageVOTypes.LOGOUT,
                pushMsgType: 3
            });
        },

        /**
         * Received Async Message Handler
         *
         * This functions parses received message from async
         *
         * @access private
         *
         * @param {object}    asyncMessage    Received Message from Async
         *
         * @return {undefined}
         */
        receivedAsyncMessageHandler = function (asyncMessage) {
            /**
             * + Message Received From Async      {object}
             *    - id                            {long}
             *    - senderMessageId               {long}
             *    - senderName                    {string}
             *    - senderId                      {long}
             *    - type                          {int}
             *    - content                       {string}
             */

            var content = JSON.parse(asyncMessage.content);
            chatMessageHandler(content);
        },

        /**
         * is Valid Json
         *
         * This functions checks if a string is valid json or not?
         *
         * @access private
         *
         * @param {string}  jsonString   Json String to be checked
         *
         * @return {boolean}
         */
        isValidJson = function (jsonString) {
            try {
                JSON.parse(jsonString);
            } catch (e) {
                return false;
            }
            return true;
        },

        /**
         * Chat Message Handler
         *
         * Manages received chat messages and do the job
         *
         * @access private
         *
         * @param {object}    chatMessage     Content of Async Message which is considered as Chat Message
         *
         * @return {undefined}
         */
        chatMessageHandler = function (chatMessage) {
            if (chatMessage.typeCode && chatMessage.typeCode !== app.sdkParams.generalTypeCode) {
                return;
            }

            var threadId = chatMessage.subjectId,
                type = chatMessage.type,
                // TODO Check this again
                messageContent = (typeof chatMessage.content === 'string' && isValidJson(chatMessage.content))
                    ? JSON.parse(chatMessage.content)
                    : chatMessage.content,
                contentCount = chatMessage.contentCount,
                uniqueId = chatMessage.uniqueId,
                time = chatMessage.time;

            switch (type) {
                /**
                 * Type 1    Get Threads
                 */
                case chatMessageVOTypes.CREATE_THREAD:
                    messageContent.uniqueId = uniqueId;

                    if (app.store.messagesCallbacks[uniqueId]) {
                        createThread(messageContent, true, true);
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    } else {
                        createThread(messageContent, true, false);
                    }

                    break;

                /**
                 * Type 2    Message
                 */
                case chatMessageVOTypes.MESSAGE:
                    newMessageHandler(threadId, messageContent, chatMessage, uniqueId);
                    break;

                /**
                 * Type 3    Message Sent
                 */
                case chatMessageVOTypes.SENT:
                    if (app.store.sendMessageCallbacks[uniqueId] && app.store.sendMessageCallbacks[uniqueId].onSent) {
                        app.store.sendMessageCallbacks[uniqueId].onSent({
                            uniqueId: uniqueId
                        });
                        delete (app.store.sendMessageCallbacks[uniqueId].onSent);
                        app.store.threadCallbacks[threadId][uniqueId].onSent = true;
                    }
                    break;

                /**
                 * Type 4    Message Delivery
                 */
                case chatMessageVOTypes.DELIVERY:
                    // if (fullResponseObject) {
                    //     getHistory({
                    //         offset: 0,
                    //         threadId: threadId,
                    //         id: messageContent.messageId,
                    //         cache: false
                    //     }, function (result) {
                    //         if (!result.hasError) {
                    //             chatEvents.fireEvent('messageEvents', {
                    //                 type: 'MESSAGE_DELIVERY',
                    //                 result: {
                    //                     message: result.result.history[0],
                    //                     threadId: threadId,
                    //                     senderId: messageContent.participantId
                    //                 }
                    //             });
                    //         }
                    //     });
                    // } else {
                    app.chatEvents.fireEvent('messageEvents', {
                        type: 'MESSAGE_DELIVERY',
                        result: {
                            message: messageContent.messageId,
                            threadId: threadId,
                            senderId: messageContent.participantId
                        }
                    });
                    // }

                    sendMessageCallbacksHandler(chatMessageVOTypes.DELIVERY, threadId, uniqueId);
                    break;

                /**
                 * Type 5    Message Seen
                 */
                case chatMessageVOTypes.SEEN:
                    if (app.sdkParams.fullResponseObject) {
                        getHistory({
                            offset: 0,
                            threadId: threadId,
                            id: messageContent.messageId,
                            cache: false
                        }, function (result) {
                            if (!result.hasError) {
                                app.chatEvents.fireEvent('messageEvents', {
                                    type: 'MESSAGE_SEEN',
                                    result: {
                                        message: result.result.history[0],
                                        threadId: threadId,
                                        senderId: messageContent.participantId
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });
                            }
                        });
                    } else {
                        app.chatEvents.fireEvent('messageEvents', {
                            type: 'MESSAGE_SEEN',
                            result: {
                                message: messageContent.messageId,
                                threadId: threadId,
                                senderId: messageContent.participantId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    sendMessageCallbacksHandler(chatMessageVOTypes.SEEN, threadId, uniqueId);
                    break;

                /**
                 * Type 6    Chat Ping
                 */
                case chatMessageVOTypes.PING:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        let result = Utility.createReturnData(false, '', 0, messageContent);
                        result.uniqueId = uniqueId;
                        app.store.messagesCallbacks[uniqueId](result);
                    }
                    break;

                /**
                 * Type 7    Block Contact
                 */
                case chatMessageVOTypes.BLOCK:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }
                    break;

                /**
                 * Type 8    Unblock Blocked User
                 */
                case chatMessageVOTypes.UNBLOCK:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }
                    break;

                /**
                 * Type 9   Leave Thread
                 */
                case chatMessageVOTypes.LEAVE_THREAD:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [threadId]
                        }, function (threadsResult) {
                            if (!threadsResult.cache) {
                                var threads = threadsResult.result.threads;
                                if (threads.length > 0) {
                                    app.chatEvents.fireEvent('threadEvents', {
                                        type: 'THREAD_LEAVE_PARTICIPANT',
                                        result: {
                                            thread: threads[0],
                                            participant: formatDataToMakeParticipant(messageContent, threadId)
                                        },
                                        uniqueId,
                                        typeCode: chatMessage.typeCode
                                    });

                                    app.chatEvents.fireEvent('threadEvents', {
                                        type: 'THREAD_LAST_ACTIVITY_TIME',
                                        result: {
                                            thread: threads[0]
                                        },
                                        uniqueId,
                                        typeCode: chatMessage.typeCode
                                    });
                                } else {
                                    app.chatEvents.fireEvent('threadEvents', {
                                        type: 'THREAD_LEAVE_PARTICIPANT',
                                        result: {
                                            threadId: threadId,
                                            participant: formatDataToMakeParticipant(messageContent, threadId)
                                        },
                                        uniqueId,
                                        typeCode: chatMessage.typeCode
                                    });
                                }
                            }
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_LEAVE_PARTICIPANT',
                            result: {
                                thread: threadId,
                                participant: formatDataToMakeParticipant(messageContent, threadId)
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });

                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_LAST_ACTIVITY_TIME',
                            result: {
                                thread: threadId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }
                    break;

                /**
                 * Type 11    Add Participant to Thread
                 */
                case chatMessageVOTypes.ADD_PARTICIPANT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [messageContent.id]
                        }, function (threadsResult) {
                            var threads = threadsResult.result.threads;

                            if (!threadsResult.cache) {
                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_ADD_PARTICIPANTS',
                                    result: {
                                        thread: threads[0]
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });

                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_LAST_ACTIVITY_TIME',
                                    result: {
                                        thread: threads[0]
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });
                            }
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_ADD_PARTICIPANTS',
                            result: {
                                thread: messageContent
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });

                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_LAST_ACTIVITY_TIME',
                            result: {
                                thread: messageContent
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }
                    break;

                /**
                 * Type 13    Get Contacts List
                 */
                case chatMessageVOTypes.GET_CONTACTS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 14    Get Threads List
                 */
                case chatMessageVOTypes.GET_THREADS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 15    Get Message History of an Thread
                 */
                case chatMessageVOTypes.GET_HISTORY:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 17    Remove sb from thread
                 */
                case chatMessageVOTypes.REMOVED_FROM_THREAD:

                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'THREAD_REMOVED_FROM',
                        result: {
                            thread: threadId
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 18    Remove a participant from Thread
                 */
                case chatMessageVOTypes.REMOVE_PARTICIPANT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [threadId]
                        }, function (threadsResult) {
                            var threads = threadsResult.result.threads;

                            if (!threadsResult.cache) {
                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_REMOVE_PARTICIPANTS',
                                    result: {
                                        thread: threads[0]
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });

                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_LAST_ACTIVITY_TIME',
                                    result: {
                                        thread: threads[0]
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });
                            }
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_REMOVE_PARTICIPANTS',
                            result: {
                                thread: threadId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });

                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_LAST_ACTIVITY_TIME',
                            result: {
                                thread: threadId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }
                    break;

                /**
                 * Type 19    Mute Thread
                 */
                case chatMessageVOTypes.MUTE_THREAD:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [threadId]
                        }, function (threadsResult) {
                            var thread = threadsResult.result.threads[0];
                            thread.mute = true;

                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'THREAD_MUTE',
                                result: {
                                    thread: thread
                                },
                                uniqueId,
                                typeCode: chatMessage.typeCode
                            });
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_MUTE',
                            result: {
                                thread: threadId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    break;

                /**
                 * Type 20    Unmute muted Thread
                 */
                case chatMessageVOTypes.UNMUTE_THREAD:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [threadId]
                        }, function (threadsResult) {
                            var thread = threadsResult.result.threads[0];
                            thread.mute = false;

                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'THREAD_UNMUTE',
                                result: {
                                    thread: thread
                                },
                                uniqueId,
                                typeCode: chatMessage.typeCode
                            });
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_UNMUTE',
                            result: {
                                thread: threadId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }
                    break;

                /**
                 * Type 21    Update Thread Info
                 */
                case chatMessageVOTypes.UPDATE_THREAD_INFO:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [messageContent.id],
                            cache: false
                        }, function (threadsResult) {
                            var thread = formatDataToMakeConversation(threadsResult.result.threads[0]);

                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'THREAD_INFO_UPDATED',
                                result: {
                                    thread: thread
                                },
                                uniqueId,
                                typeCode: chatMessage.typeCode
                            });
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_INFO_UPDATED',
                            result: {
                                thread: messageContent
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }
                    break;

                /**
                 * Type 22    Forward Multiple Messages
                 */
                case chatMessageVOTypes.FORWARD_MESSAGE:
                    newMessageHandler(threadId, messageContent, chatMessage, uniqueId);
                    break;

                /**
                 * Type 23    User Info
                 */
                case chatMessageVOTypes.USER_INFO:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    app.chatEvents.fireEvent('systemEvents', {
                        type: 'SERVER_TIME',
                        result: {
                            time: time
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;

                /**
                 * Type 25    Get Blocked List
                 */
                case chatMessageVOTypes.GET_BLOCKED:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 27    Thread Participants List
                 */
                case chatMessageVOTypes.THREAD_PARTICIPANTS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 28    Edit Message
                 */
                case chatMessageVOTypes.EDIT_MESSAGE:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    chatEditMessageHandler(threadId, messageContent, chatMessage, uniqueId);
                    break;

                /**
                 * Type 29    Delete Message
                 */
                case chatMessageVOTypes.DELETE_MESSAGE:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    if (messageContent.pinned) {
                        unPinMessage({
                            messageId: messageContent.id,
                            notifyAll: true
                        });
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [threadId]
                        }, function (threadsResult) {
                            var threads = threadsResult.result.threads;

                            if (!threadsResult.cache) {
                                app.chatEvents.fireEvent('messageEvents', {
                                    type: 'MESSAGE_DELETE',
                                    result: {
                                        message: {
                                            id: messageContent.id,
                                            pinned: messageContent.pinned,
                                            threadId: threadId
                                        }
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });

                                if (messageContent.pinned) {
                                    app.chatEvents.fireEvent('threadEvents', {
                                        type: 'THREAD_LAST_ACTIVITY_TIME',
                                        result: {
                                            thread: threads[0]
                                        },
                                        uniqueId,
                                        typeCode: chatMessage.typeCode
                                    });
                                }
                            }
                        });
                    } else {
                        app.chatEvents.fireEvent('messageEvents', {
                            type: 'MESSAGE_DELETE',
                            result: {
                                message: {
                                    id: messageContent.id,
                                    pinned: messageContent.pinned,
                                    threadId: threadId
                                }
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });

                        if (messageContent.pinned) {
                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'THREAD_LAST_ACTIVITY_TIME',
                                result: {
                                    thread: threadId
                                },
                                uniqueId,
                                typeCode: chatMessage.typeCode
                            });
                        }
                    }

                    break;

                /**
                 * Type 30    Thread Info Updated
                 */
                case chatMessageVOTypes.THREAD_INFO_UPDATED:
                    var thread = formatDataToMakeConversation(messageContent);

                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'THREAD_INFO_UPDATED',
                        result: {
                            thread: thread
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;

                /**
                 * Type 31    Thread Last Seen Updated
                 */
                case chatMessageVOTypes.LAST_SEEN_UPDATED:
                    var threadObject = messageContent;
                    threadObject.unreadCount = (messageContent.unreadCount) ? messageContent.unreadCount : 0;

                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'THREAD_UNREAD_COUNT_UPDATED',
                        result: {
                            thread: (app.sdkParams.fullResponseObject ? threadObject : messageContent.id),
                            unreadCount: (messageContent.unreadCount) ? messageContent.unreadCount : 0
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });


                    break;

                /**
                 * Type 32    Get Message Delivered List
                 */
                case chatMessageVOTypes.GET_MESSAGE_DELEVERY_PARTICIPANTS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 33    Get Message Seen List
                 */
                case chatMessageVOTypes.GET_MESSAGE_SEEN_PARTICIPANTS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 34    Is Public Group Name Available?
                 */
                case chatMessageVOTypes.IS_NAME_AVAILABLE:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 39    Join Public Group or Channel
                 */
                case chatMessageVOTypes.JOIN_THREAD:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 40    Bot Messages
                 */
                case chatMessageVOTypes.BOT_MESSAGE:
                    app.chatEvents.fireEvent('botEvents', {
                        type: 'BOT_MESSAGE',
                        result: {
                            bot: messageContent
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;

                /**
                 * Type 41    Spam P2P Thread
                 */
                case chatMessageVOTypes.SPAM_PV_THREAD:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }
                    break;

                /**
                 * Type 42    Set Role To User
                 */
                case chatMessageVOTypes.SET_ROLE_TO_USER:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [messageContent.id]
                        }, function (threadsResult) {
                            var threads = threadsResult.result.threads;

                            if (!threadsResult.cache) {
                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_ADD_ADMIN',
                                    result: {
                                        thread: threads[0],
                                        admin: messageContent
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });

                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_LAST_ACTIVITY_TIME',
                                    result: {
                                        thread: threads[0],
                                        admin: messageContent
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });
                            }
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_ADD_ADMIN',
                            result: {
                                thread: threadId,
                                admin: messageContent
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });

                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_LAST_ACTIVITY_TIME',
                            result: {
                                thread: threadId,
                                admin: messageContent
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    break;

                /**
                 * Type 43    Remove Role From User
                 */
                case chatMessageVOTypes.REMOVE_ROLE_FROM_USER:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [messageContent.id]
                        }, function (threadsResult) {
                            var threads = threadsResult.result.threads;

                            if (!threadsResult.cache) {
                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_REMOVE_ADMIN',
                                    result: {
                                        thread: threads[0],
                                        admin: messageContent
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });

                                app.chatEvents.fireEvent('threadEvents', {
                                    type: 'THREAD_LAST_ACTIVITY_TIME',
                                    result: {
                                        thread: threads[0],
                                        admin: messageContent
                                    },
                                    uniqueId,
                                    typeCode: chatMessage.typeCode
                                });
                            }
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_REMOVE_ADMIN',
                            result: {
                                thread: threadId,
                                admin: messageContent
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });

                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_LAST_ACTIVITY_TIME',
                            result: {
                                thread: threadId,
                                admin: messageContent
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    break;

                /**
                 * Type 44    Clear History
                 */
                case chatMessageVOTypes.CLEAR_HISTORY:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }
                    break;

                /**
                 * Type 46    System Messages
                 */
                case chatMessageVOTypes.SYSTEM_MESSAGE:
                    app.chatEvents.fireEvent('systemEvents', {
                        type: 'IS_TYPING',
                        result: {
                            thread: threadId,
                            user: messageContent
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;

                /**
                 * Type 47    Get Not Seen Duration
                 */
                case chatMessageVOTypes.GET_NOT_SEEN_DURATION:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }
                    break;

                /**
                 * Type 48    Pin Thread
                 */
                case chatMessageVOTypes.PIN_THREAD:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [threadId]
                        }, function (threadsResult) {
                            var thread = threadsResult.result.threads[0];

                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'THREAD_PIN',
                                result: {
                                    thread: thread
                                },
                                uniqueId,
                                typeCode: chatMessage.typeCode
                            });
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_PIN',
                            result: {
                                thread: threadId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    break;

                /**
                 * Type 49    UnPin Thread
                 */
                case chatMessageVOTypes.UNPIN_THREAD:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    if (app.sdkParams.fullResponseObject) {
                        getThreads({
                            threadIds: [threadId]
                        }, function (threadsResult) {
                            var thread = threadsResult.result.threads[0];

                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'THREAD_UNPIN',
                                result: {
                                    thread: thread
                                },
                                uniqueId,
                                typeCode: chatMessage.typeCode
                            });
                        });
                    } else {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREAD_UNPIN',
                            result: {
                                thread: threadId
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    break;

                /**
                 * Type 50    Pin Message
                 */
                case chatMessageVOTypes.PIN_MESSAGE:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'MESSAGE_PIN',
                        result: {
                            thread: threadId,
                            pinMessage: formatDataToMakePinMessage(threadId, messageContent)
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 51    UnPin Message
                 */
                case chatMessageVOTypes.UNPIN_MESSAGE:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'MESSAGE_UNPIN',
                        result: {
                            thread: threadId,
                            pinMessage: formatDataToMakePinMessage(threadId, messageContent)
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 52    Update Chat Profile
                 */
                case chatMessageVOTypes.UPDATE_CHAT_PROFILE:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    app.chatEvents.fireEvent('userEvents', {
                        type: 'CHAT_PROFILE_UPDATED',
                        result: {
                            user: messageContent
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 54    Get Participant Roles
                 */
                case chatMessageVOTypes.GET_PARTICIPANT_ROLES:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    app.chatEvents.fireEvent('userEvents', {
                        type: 'GET_PARTICIPANT_ROLES',
                        result: {
                            roles: messageContent
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 60    Get Contact Not Seen Duration
                 */
                case chatMessageVOTypes.GET_CONTACT_NOT_SEEN_DURATION:
                    app.chatEvents.fireEvent('contactEvents', {
                        type: 'CONTACTS_LAST_SEEN',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;

                /**
                 * Type 61      Get All Unread Message Count
                 */
                case chatMessageVOTypes.ALL_UNREAD_MESSAGE_COUNT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent));
                    }

                    app.chatEvents.fireEvent('systemEvents', {
                        type: 'ALL_UNREAD_MESSAGES_COUNT',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 62    Create Bot
                 */
                case chatMessageVOTypes.CREATE_BOT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 63    Define Bot Commands
                 */
                case chatMessageVOTypes.DEFINE_BOT_COMMAND:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 64    Start Bot
                 */
                case chatMessageVOTypes.START_BOT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 65    Stop Bot
                 */
                case chatMessageVOTypes.STOP_BOT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;
                /**
                 * Type 66    Last Message Deleted
                 */
                case chatMessageVOTypes.LAST_MESSAGE_DELETED:
                    delete messageContent.unreadCount;
                    let threadOfDeletedMessage = formatDataToMakeConversation(messageContent);
                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'THREAD_INFO_UPDATED',
                        result: {
                            threadId,
                            thread: threadOfDeletedMessage
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;
                /**
                 * Type 67    Last Message Edited
                 */
                case chatMessageVOTypes.LAST_MESSAGE_EDITED:
                    delete messageContent.unreadCount;
                    let threadOfEditedMessage = formatDataToMakeConversation(messageContent);
                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'THREAD_INFO_UPDATED',
                        result: {
                            threadId,
                            thread: threadOfEditedMessage
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;
                /**
                 * Type 68    Get Bot Commands List
                 */
                case chatMessageVOTypes.BOT_COMMANDS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 69    Get Thread All Bots
                 */
                case chatMessageVOTypes.THREAD_ALL_BOTS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 70    Send Call Request
                 */
                case chatMessageVOTypes.CALL_REQUEST:
                    callRequestController.callRequestReceived = true;
                    callReceived({
                        callId: messageContent.callId
                    }, function (r) {

                    });

                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'RECEIVE_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    if (messageContent.callId > 0) {
                        if (!currentCallId) {
                            currentCallId = messageContent.callId;
                        } else
                            newCallId = messageContent.callId;
                    } else {
                        app.chatEvents.fireEvent('callEvents', {
                            type: 'PARTNER_RECEIVED_YOUR_CALL',
                            result: messageContent,
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    currentCallId = messageContent.callId;

                    break;

                /**
                 * Type 71    Accept Call Request
                 */
                case chatMessageVOTypes.ACCEPT_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'ACCEPT_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 72    Reject Call Request
                 */
                case chatMessageVOTypes.REJECT_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'REJECT_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 73    Receive Call Request
                 */
                case chatMessageVOTypes.RECEIVE_CALL_REQUEST:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    if (messageContent.callId > 0) {
                        app.chatEvents.fireEvent('callEvents', {
                            type: 'RECEIVE_CALL',
                            result: messageContent,
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                        if (!currentCallId) {
                            currentCallId = messageContent.callId;
                        } else
                            newCallId = messageContent.callId;
                    } else {
                        app.chatEvents.fireEvent('callEvents', {
                            type: 'PARTNER_RECEIVED_YOUR_CALL',
                            result: messageContent,
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                    }

                    // currentCallId = messageContent.callId;

                    break;

                /**
                 * Type 74    Start Call Request
                 */
                case chatMessageVOTypes.START_CALL:
                    if (!callRequestController.iCanAcceptTheCall()) {
                        //Do not handle the call if it is already started elsewhere for the same user
                        app.chatEvents.fireEvent('callEvents', {
                            type: 'CALL_STARTED_ELSEWHERE',
                            message: 'Call already started somewhere else..., aborting...',
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });
                        return;
                    }

                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    messageContent.callId = threadId;
                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_STARTED',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    if (typeof messageContent === 'object'
                        && messageContent.hasOwnProperty('chatDataDto')
                        && !!messageContent.chatDataDto.kurentoAddress) {
                    } else {
                        app.chatEvents.fireEvent('callEvents', {
                            type: 'CALL_ERROR',
                            message: 'Chat Data DTO is not present!'
                        });
                    }

                    break;

                /**
                 * Type 75    End Call Request
                 */
                case chatMessageVOTypes.END_CALL_REQUEST:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'END_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    callStop();

                    break;

                /**
                 * Type 76   Call Ended
                 */
                case chatMessageVOTypes.END_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_ENDED',
                        callId: threadId,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    callStop();

                    break;

                /**
                 * Type 77    Get Calls History
                 */
                case chatMessageVOTypes.GET_CALLS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    break;


                /**
                 * Type 90    Contacts Synced
                 */
                case chatMessageVOTypes.CONTACT_SYNCED:
                    app.chatEvents.fireEvent('contactEvents', {
                        type: 'CONTACTS_SYNCED',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    break;

                /**
                 * Type 91    Send Group Call Request
                 */
                case chatMessageVOTypes.GROUP_CALL_REQUEST:
                    callRequestController.callRequestReceived = true;
                    callReceived({
                        callId: messageContent.callId
                    }, function (r) {
                    });

                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    if (messageContent.callId > 0) {
                        if (!currentCallId)
                            currentCallId = messageContent.callId;
                        else
                            newCallId = messageContent.callId;
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'RECEIVE_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    // currentCallId = messageContent.callId;

                    break;

                /**
                 * Type 92    Call Partner Leave
                 */
                case chatMessageVOTypes.LEAVE_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_PARTICIPANT_LEFT',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 93    Add Call Participant
                 */
                case chatMessageVOTypes.ADD_CALL_PARTICIPANT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    break;

                /**
                 * Type 94    Call Participant Joined
                 */
                case chatMessageVOTypes.CALL_PARTICIPANT_JOINED:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_PARTICIPANT_JOINED',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 95    Remove Call Participant
                 */
                case chatMessageVOTypes.REMOVE_CALL_PARTICIPANT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_PARTICIPANT_REMOVED',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 96    Terminate Call
                 */
                case chatMessageVOTypes.TERMINATE_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'TERMINATE_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    callStop();

                    break;

                /**
                 * Type 97    Mute Call Participant
                 */
                case chatMessageVOTypes.MUTE_CALL_PARTICIPANT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }


                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_PARTICIPANT_MUTE',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 98    UnMute Call Participant
                 */
                case chatMessageVOTypes.UNMUTE_CALL_PARTICIPANT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    } else {
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_PARTICIPANT_UNMUTE',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 110    Active Call Participants List
                 */
                case chatMessageVOTypes.ACTIVE_CALL_PARTICIPANTS:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    break;

                /**
                 * Type 111    Kafka Call Session Created
                 */
                case chatMessageVOTypes.CALL_SESSION_CREATED:
                    if (!callRequestController.callEstablishedInMySide)
                        return;

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_SESSION_CREATED',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                    if (!currentCallId)
                        currentCallId = messageContent.callId;
                    else
                        newCallId = messageContent.callId;

                    // currentCallId = messageContent.callId;

                    break;

                /**
                 * Type 113    Turn On Video Call
                 */
                case chatMessageVOTypes.TURN_ON_VIDEO_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'TURN_ON_VIDEO_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 114    Turn Off Video Call
                 */
                case chatMessageVOTypes.TURN_OFF_VIDEO_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'TURN_OFF_VIDEO_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 121    Record Call Request
                 */
                case chatMessageVOTypes.RECORD_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'START_RECORDING_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 122   End Record Call Request
                 */
                case chatMessageVOTypes.END_RECORD_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'STOP_RECORDING_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 126   Destined Record Call Request
                 */
                case chatMessageVOTypes.DESTINATED_RECORD_CALL:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'START_RECORDING_CALL',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 127   Assistant call started
                 *
                 * Tells the assistant that a call started
                 */
                case chatMessageVOTypes.ASSISTANT_CALL_STARTED:
                    app.chatEvents.fireEvent('callEvents', {
                        type: 'ASSISTANT_CALL_STARTED',
                        result: {
                            callId: threadId
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 128   Assistant call ended
                 *
                 * Tells the assistant that a call ended
                 */
                case chatMessageVOTypes.ASSISTANT_CALL_ENDED:
                    app.chatEvents.fireEvent('callEvents', {
                        type: 'ASSISTANT_CALL_ENDED',
                        callId: threadId,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 129   Get Calls To Join
                 */
                case chatMessageVOTypes.GET_CALLS_TO_JOIN:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;

                /**
                 * Type 152    Gives us a json to export for user
                 */
                case chatMessageVOTypes.EXPORT_CHAT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount, uniqueId));
                    }

                    break;


                /**
                 * Type 221  Event to tell us p2p call converted to a group call
                 */
                case chatMessageVOTypes.SWITCH_TO_GROUP_CALL_REQUEST:
                    app.chatEvents.fireEvent('callEvents', {
                        type: 'SWITCH_TO_GROUP_CALL',
                        result: messageContent, //contains: isGroup, callId, threadId
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 222    Call Recording Started
                 */
                case chatMessageVOTypes.RECORD_CALL_STARTED:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    app.chatEvents.fireEvent('callEvents', {
                        type: 'CALL_RECORDING_STARTED',
                        result: messageContent,
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });

                    break;

                /**
                 * Type 237    GET THREAD LIGHT
                 */
                case chatMessageVOTypes.GET_THREAD_LIGHT:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }

                    break;

                /**
                 * Type 239    ADD_REACTION
                 */
                case chatMessageVOTypes.ADD_REACTION:
                    onAddReaction(uniqueId, messageContent, contentCount);
                    break;

                /**
                 * Type 242    REACTION_LIST
                 */
                case chatMessageVOTypes.REACTION_LIST:
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount));
                    }
                    break;


                /**
                 * Type 999   All unknown errors
                 */
                case chatMessageVOTypes.ERROR:
                    if(messageContent.message) {
                        messageContent.message = '[ChatServer] ' + messageContent.message;
                    }
                    if (app.store.messagesCallbacks[uniqueId]) {
                        app.store.messagesCallbacks[uniqueId](Utility.createReturnData(true, messageContent.message, messageContent.code, messageContent, 0));
                    }

                    /**
                     * If error code is 21, Token is invalid &
                     * user should logged out
                     */
                    if (messageContent.code == 21) {
                        // TODO: Temporarily removed due to unknown side-effects
                        // chatState = false;
                        // asyncClient.logout();
                        // clearChatServerCaches();
                    }

                    app.chatEvents.fireEvent('error', {
                        code: messageContent.code,
                        message: messageContent.message,
                        error: messageContent,
                        uniqueId: uniqueId
                    });

                    break;
            }
        },

        /**
         * Send Message Callbacks Handler
         *
         * When you send Delivery or Seen Acknowledgements of a message
         * You should send Delivery and Seen for all the Messages before
         * that message so that you wont have un delivered/unseen messages
         * after seeing the last message of a thread
         *
         * @access private
         *
         * @param {int}     actionType      Switch between Delivery or Seen
         * @param {long}    threadId        Id of thread
         * @param {string}  uniqueId        uniqueId of message
         *
         * @return {undefined}
         */
        sendMessageCallbacksHandler = function (actionType, threadId, uniqueId) {
            switch (actionType) {

                case chatMessageVOTypes.DELIVERY:
                    if (app.store.threadCallbacks[threadId]) {
                        var lastThreadCallbackIndex = Object.keys(app.store.threadCallbacks[threadId])
                            .indexOf(uniqueId);
                        if (lastThreadCallbackIndex !== undefined) {
                            while (lastThreadCallbackIndex > -1) {
                                var tempUniqueId = Object.entries(app.store.threadCallbacks[threadId])[lastThreadCallbackIndex][0];
                                if (app.store.sendMessageCallbacks[tempUniqueId] && app.store.sendMessageCallbacks[tempUniqueId].onDeliver) {
                                    if (app.store.threadCallbacks[threadId][tempUniqueId] && app.store.threadCallbacks[threadId][tempUniqueId].onSent) {
                                        // app.store.sendMessageCallbacks[tempUniqueId].onDeliver(
                                        //     {
                                        //         uniqueId: tempUniqueId
                                        //     });
                                        delete (app.store.sendMessageCallbacks[tempUniqueId].onDeliver);
                                        app.store.threadCallbacks[threadId][tempUniqueId].onDeliver = true;
                                    }
                                }

                                lastThreadCallbackIndex -= 1;
                            }
                        }
                    }
                    break;

                case chatMessageVOTypes.SEEN:
                    if (app.store.threadCallbacks[threadId]) {
                        var lastThreadCallbackIndex = Object.keys(app.store.threadCallbacks[threadId])
                            .indexOf(uniqueId);
                        if (lastThreadCallbackIndex !== undefined) {
                            while (lastThreadCallbackIndex > -1) {
                                var tempUniqueId = Object.entries(app.store.threadCallbacks[threadId])[lastThreadCallbackIndex][0];

                                if (app.store.sendMessageCallbacks[tempUniqueId] && app.store.sendMessageCallbacks[tempUniqueId].onSeen) {
                                    if (app.store.threadCallbacks[threadId][tempUniqueId] && app.store.threadCallbacks[threadId][tempUniqueId].onSent) {
                                        // if (!app.store.threadCallbacks[threadId][tempUniqueId].onDeliver) {
                                        //     app.store.sendMessageCallbacks[tempUniqueId].onDeliver(
                                        //         {
                                        //             uniqueId: tempUniqueId
                                        //         });
                                        //     delete (app.store.sendMessageCallbacks[tempUniqueId].onDeliver);
                                        //     app.store.threadCallbacks[threadId][tempUniqueId].onDeliver = true;
                                        // }

                                        app.store.sendMessageCallbacks[tempUniqueId].onSeen(
                                            {
                                                uniqueId: tempUniqueId
                                            });

                                        delete (app.store.sendMessageCallbacks[tempUniqueId].onSeen);
                                        app.store.threadCallbacks[threadId][tempUniqueId].onSeen = true;

                                        if (app.store.threadCallbacks[threadId][tempUniqueId].onSent &&
                                            app.store.threadCallbacks[threadId][tempUniqueId].onDeliver &&
                                            app.store.threadCallbacks[threadId][tempUniqueId].onSeen) {
                                            delete app.store.threadCallbacks[threadId][tempUniqueId];
                                            delete app.store.sendMessageCallbacks[tempUniqueId];
                                        }
                                    }
                                }

                                lastThreadCallbackIndex -= 1;
                            }
                        }
                    }
                    break;

                default:
                    break;
            }
        },

        /**
         * New Message Handler
         *
         * Handles Event Emitter of a newly received Chat Message
         *
         * @access private
         *
         * @param {long}    threadId         ID of image
         * @param {object}  messageContent   Json Content of the message
         * @param {object}  chatMessage
         * @param {string}  uniqueId
         *
         * @return {undefined}
         */
        newMessageHandler = function (threadId, messageContent, chatMessage, uniqueId) {
            var message = formatDataToMakeMessage(threadId, messageContent);
            // deliver({
            //     messageId: message.id,
            //     ownerId: message.participant.id
            // });

            app.chatEvents.fireEvent('messageEvents', {
                type: 'MESSAGE_NEW',
                cache: false,
                result: {
                    message: message
                },
                uniqueId,
                typeCode: chatMessage.typeCode
            });

            var threadObject = message.conversation;
            var lastMessageVoCopy = Object.assign({}, message);
            lastMessageVoCopy.conversation && delete lastMessageVoCopy.conversation;

            threadObject.lastParticipantImage = (!!message.participant && message.participant.hasOwnProperty('image')) ? message.participant.image : '';
            threadObject.lastMessageVO = lastMessageVoCopy;
            threadObject.lastParticipantName = (!!message.participant && message.participant.hasOwnProperty('name')) ? message.participant.name : '';
            threadObject.lastMessage = (message.hasOwnProperty('message')) ? message.message : '';


            app.chatEvents.fireEvent('threadEvents', {
                type: 'THREAD_UNREAD_COUNT_UPDATED',
                result: {
                    thread: (app.sdkParams.fullResponseObject ? threadObject : messageContent.id),
                    unreadCount: (threadObject.unreadCount) ? threadObject.unreadCount : 0
                },
                uniqueId,
                typeCode: chatMessage.typeCode
            });

            app.chatEvents.fireEvent('threadEvents', {
                type: 'THREAD_LAST_ACTIVITY_TIME',
                result: {
                    thread: (app.sdkParams.fullResponseObject ? threadObject : messageContent.id),
                },
                uniqueId,
                typeCode: chatMessage.typeCode
            });

            for (var i = 0; i < chatSendQueue.length; i++) {
                if (chatSendQueue[i].uniqueId == message.uniqueId) {
                    chatSendQueue.splice(i, 1);
                }
            }
        },

        /**
         * Chat Edit Message Handler
         *
         * Handles Event Emitter of an edited Chat Message
         *
         * @access private
         *
         * @param {long}    threadId         ID of image
         * @param {object}  messageContent   Json Content of the message
         *
         * @return {undefined}
         */
        chatEditMessageHandler = function (threadId, messageContent, chatMessage, uniqueId) {
            var message = formatDataToMakeMessage(threadId, messageContent);

            if (app.sdkParams.fullResponseObject) {
                getThreads({
                    threadIds: [threadId]
                }, function (threadsResult) {
                    var threads = threadsResult.result.threads;

                    if (!threadsResult.cache) {
                        app.chatEvents.fireEvent('messageEvents', {
                            type: 'MESSAGE_EDIT',
                            result: {
                                message: message
                            },
                            uniqueId,
                            typeCode: chatMessage.typeCode
                        });

                        if (message.pinned) {
                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'THREAD_LAST_ACTIVITY_TIME',
                                result: {
                                    thread: threads[0]
                                },
                                uniqueId,
                                typeCode: chatMessage.typeCode
                            });
                        }
                    }
                });
            } else {
                app.chatEvents.fireEvent('messageEvents', {
                    type: 'MESSAGE_EDIT',
                    result: {
                        message: message
                    },
                    uniqueId,
                    typeCode: chatMessage.typeCode
                });

                if (message.pinned) {
                    app.chatEvents.fireEvent('threadEvents', {
                        type: 'THREAD_LAST_ACTIVITY_TIME',
                        result: {
                            thread: threadId
                        },
                        uniqueId,
                        typeCode: chatMessage.typeCode
                    });
                }
            }
        },

        /**
         * Create Thread
         *
         * Makes formatted Thread Object out of given contentCount,
         * If Thread has been newly created, a THREAD_NEW event
         * will be emitted
         *
         * @access private
         *
         * @param {object}    messageContent    Json object of thread taken from chat server
         * @param {boolean}   addFromService    if this is a newly created Thread, addFromService should be True
         *
         * @return {object} Formatted Thread Object
         */
        createThread = function (messageContent, addFromService, showThread) {
            var threadData = formatDataToMakeConversation(messageContent);
            var redirectToThread = (showThread === true) ? showThread : false;

            if (addFromService) {
                app.chatEvents.fireEvent('threadEvents', {
                    type: 'THREAD_NEW',
                    redirectToThread: redirectToThread,
                    result: {
                        thread: threadData
                    },
                    uniqueId: messageContent.uniqueId
                });
            }
            return threadData;
        },

        /**
         * Format Data To Make Linked User
         *
         * This functions re-formats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} linkedUser Object
         */
        formatDataToMakeLinkedUser = function (messageContent) {
            /**
             * + RelatedUserVO                 {object}
             *   - coreUserId                  {long}
             *   - username                    {string}
             *   - nickname                    {string}
             *   - name                        {string}
             *   - image                       {string}
             */

            var linkedUser = {
                coreUserId: (messageContent.coreUserId !== undefined)
                    ? messageContent.coreUserId
                    : messageContent.id,
                username: messageContent.username,
                nickname: messageContent.nickname,
                name: messageContent.name,
                image: messageContent.image
            };

            // return linkedUser;
            return JSON.parse(JSON.stringify(linkedUser));
        },

        /**
         * Format Data To Make Contact
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} contact Object
         */
        formatDataToMakeContact = function (messageContent) {
            /**
             * + ContactVO                        {object}
             *    - id                            {long}
             *    - blocked                       {boolean}
             *    - userId                        {long}
             *    - firstName                     {string}
             *    - lastName                      {string}
             *    - image                         {string}
             *    - email                         {string}
             *    - cellphoneNumber               {string}
             *    - uniqueId                      {string}
             *    - notSeenDuration               {long}
             *    - hasUser                       {boolean}
             *    - linkedUser                    {object : RelatedUserVO}
             */

            var contact = {
                id: messageContent.id,
                blocked: (messageContent.blocked !== undefined)
                    ? messageContent.blocked
                    : false,
                userId: messageContent.userId,
                firstName: messageContent.firstName,
                lastName: messageContent.lastName,
                image: messageContent.profileImage,
                email: messageContent.email,
                cellphoneNumber: messageContent.cellphoneNumber,
                uniqueId: messageContent.uniqueId,
                notSeenDuration: messageContent.notSeenDuration,
                hasUser: messageContent.hasUser,
                linkedUser: undefined
            };

            if (messageContent.linkedUser !== undefined) {
                contact.linkedUser = formatDataToMakeLinkedUser(messageContent.linkedUser);
            }

            // return contact;
            return JSON.parse(JSON.stringify(contact));
        },

        /**
         * Format Data To Make User
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} user Object
         */
        formatDataToMakeUser = function (messageContent) {
            /**
             * + User                     {object}
             *    - id                    {long}
             *    - name                  {string}
             *    - email                 {string}
             *    - cellphoneNumber       {string}
             *    - image                 {string}
             *    - lastSeen              {long}
             *    - sendEnable            {boolean}
             *    - receiveEnable         {boolean}
             *    - contactSynced         {boolean}
             *    - chatProfileVO         {object:chatProfileVO}
             */

            var user = {
                id: messageContent.id,
                name: messageContent.name,
                email: messageContent.email,
                cellphoneNumber: messageContent.cellphoneNumber,
                image: messageContent.image,
                lastSeen: messageContent.lastSeen,
                sendEnable: messageContent.sendEnable,
                receiveEnable: messageContent.receiveEnable,
                contactSynced: messageContent.contactSynced
            };

            if (messageContent.contactId) {
                user.contactId = messageContent.contactId;
            }

            if (messageContent.contactName) {
                user.contactName = messageContent.contactName;
            }

            if (messageContent.contactFirstName) {
                user.contactFirstName = messageContent.contactFirstName;
            }

            if (messageContent.contactLastName) {
                user.contactLastName = messageContent.contactLastName;
            }

            if (messageContent.blocked) {
                user.blocked = messageContent.blocked;
            }

            // Add chatProfileVO if exist
            if (messageContent.chatProfileVO) {
                user.chatProfileVO = messageContent.chatProfileVO;
            }

            // return user;
            return JSON.parse(JSON.stringify(user));
        },

        /**
         * Format Data To Make Blocked User
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} blockedUser Object
         */
        formatDataToMakeBlockedUser = function (messageContent) {
            /**
             * + BlockedUser              {object}
             *    - id                    {long}
             *    - coreUserId            {long}
             *    - firstName             {string}
             *    - lastName              {string}
             *    - nickName              {string}
             *    - profileImage          {string}
             *    - contact               {object: contactVO}
             */

            var blockedUser = {
                blockId: messageContent.id,
                coreUserId: messageContent.coreUserId,
                firstName: messageContent.firstName,
                lastName: messageContent.lastName,
                nickName: messageContent.nickName,
                profileImage: messageContent.profileImage
            };

            // Add contactVO if exist
            if (messageContent.contactVO) {
                blockedUser.contact = messageContent.contactVO;
            }

            // return blockedUser;
            return JSON.parse(JSON.stringify(blockedUser));
        },

        /**
         * Format Data To Make Invitee
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} inviteeData Object
         */
        formatDataToMakeInvitee = function (messageContent) {
            /**
             * + InviteeVO       {object}
             *    - id           {string}
             *    - idType       {int}
             */

            var inviteeData = {
                id: messageContent.id,
                idType: inviteeVOidTypes[messageContent.idType]
            };

            return inviteeData;
        },

        /**
         * Format Data To Make Participant
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} participant Object
         */
        formatDataToMakeParticipant = function (messageContent, threadId) {
            /**
             * + ParticipantVO                   {object}
             *    - id                           {long}
             *    - coreUserId                   {long}
             *    - threadId                     {long}
             *    - sendEnable                   {boolean}
             *    - receiveEnable                {boolean}
             *    - firstName                    {string}
             *    - lastName                     {string}
             *    - name                         {string}
             *    - cellphoneNumber              {string}
             *    - email                        {string}
             *    - image                        {string}
             *    - chatProfileVO                {object}
             *    - myFriend                     {boolean}
             *    - online                       {boolean}
             *    - notSeenDuration              {long}
             *    - contactId                    {long}
             *    - contactName                  {string}
             *    - contactFirstName             {string}
             *    - contactLastName              {string}
             *    - blocked                      {boolean}
             *    - admin                        {boolean}
             *    - auditor                      {boolean}
             *    - keyId                        {string}
             *    - roles                        {list:string}
             *    - username                     {string}
             */

            var participant = {
                id: messageContent.id,
                coreUserId: messageContent.coreUserId,
                threadId: threadId,
                sendEnable: messageContent.sendEnable,
                receiveEnable: messageContent.receiveEnable,
                firstName: messageContent.firstName,
                lastName: messageContent.lastName,
                name: messageContent.name,
                cellphoneNumber: messageContent.cellphoneNumber,
                email: messageContent.email,
                image: messageContent.image,
                myFriend: messageContent.myFriend,
                online: messageContent.online,
                notSeenDuration: messageContent.notSeenDuration,
                contactId: messageContent.contactId,
                contactName: messageContent.contactName,
                contactFirstName: messageContent.contactFirstName,
                contactLastName: messageContent.contactLastName,
                blocked: messageContent.blocked,
                admin: messageContent.admin,
                auditor: messageContent.auditor,
                keyId: messageContent.keyId,
                roles: messageContent.roles,
                username: messageContent.username
            };

            // Add chatProfileVO if exist
            if (messageContent.chatProfileVO) {
                participant.chatProfileVO = messageContent.chatProfileVO;
            }

            // return participant;
            return JSON.parse(JSON.stringify(participant));
        },

        /**
         * Format Data To Make Conversation
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} Conversation Object
         */
        formatDataToMakeConversation = function (messageContent) {

            /**
             * + Conversation                           {object}
             *    - id                                  {long}
             *    - joinDate                            {long}
             *    - title                               {string}
             *    - inviter                             {object : ParticipantVO}
             *    - participants                        {list : ParticipantVO}
             *    - time                                {long}
             *    - lastMessage                         {string}
             *    - lastParticipantName                 {string}
             *    - group                               {boolean}
             *    - partner                             {long}
             *    - lastParticipantImage                {string}
             *    - image                               {string}
             *    - description                         {string}
             *    - unreadCount                         {long}
             *    - lastSeenMessageId                   {long}
             *    - lastSeenMessageTime                 {long}
             *    - lastSeenMessageNanos                {integer}
             *    - lastMessageVO                       {object : ChatMessageVO}
             *    - pinMessageVO                        {object : pinMessageVO}
             *    - partnerLastSeenMessageId            {long}
             *    - partnerLastSeenMessageTime          {long}
             *    - partnerLastSeenMessageNanos         {integer}
             *    - partnerLastDeliveredMessageId       {long}
             *    - partnerLastDeliveredMessageTime     {long}
             *    - partnerLastDeliveredMessageNanos    {integer}
             *    - type                                {int}
             *    - metadata                            {string}
             *    - mute                                {boolean}
             *    - participantCount                    {long}
             *    - canEditInfo                         {boolean}
             *    - canSpam                             {boolean}
             *    - admin                               {boolean}
             *    - mentioned                           {boolean}
             *    - pin                                 {boolean}
             *    - uniqueName                          {string}
             *    - userGroupHash                       {string}
             */

            var conversation = {
                id: messageContent.id,
                joinDate: messageContent.joinDate,
                title: messageContent.title,
                inviter: undefined,
                participants: undefined,
                time: messageContent.time,
                lastMessage: messageContent.lastMessage,
                lastParticipantName: messageContent.lastParticipantName,
                group: messageContent.group,
                partner: messageContent.partner,
                lastParticipantImage: messageContent.lastParticipantImage,
                image: messageContent.image,
                description: messageContent.description,
                unreadCount: messageContent.unreadCount,
                lastSeenMessageId: messageContent.lastSeenMessageId,
                lastSeenMessageTime: (messageContent.lastSeenMessageNanos)
                    ? (parseInt(parseInt(messageContent.lastSeenMessageTime) / 1000) * 1000000000) + parseInt(messageContent.lastSeenMessageNanos)
                    : (parseInt(messageContent.lastSeenMessageTime)),
                lastMessageVO: undefined,
                pinMessageVO: undefined,
                partnerLastSeenMessageId: messageContent.partnerLastSeenMessageId,
                partnerLastSeenMessageTime: (messageContent.partnerLastSeenMessageNanos)
                    ? (parseInt(parseInt(messageContent.partnerLastSeenMessageTime) / 1000) * 1000000000) +
                    parseInt(messageContent.partnerLastSeenMessageNanos)
                    : (parseInt(messageContent.partnerLastSeenMessageTime)),
                partnerLastDeliveredMessageId: messageContent.partnerLastDeliveredMessageId,
                partnerLastDeliveredMessageTime: (messageContent.partnerLastDeliveredMessageNanos)
                    ? (parseInt(parseInt(messageContent.partnerLastDeliveredMessageTime) / 1000) * 1000000000) +
                    parseInt(messageContent.partnerLastDeliveredMessageNanos)
                    : (parseInt(messageContent.partnerLastDeliveredMessageTime)),
                type: messageContent.type,
                metadata: messageContent.metadata,
                mute: messageContent.mute,
                participantCount: messageContent.participantCount,
                canEditInfo: messageContent.canEditInfo,
                canSpam: messageContent.canSpam,
                admin: messageContent.admin,
                mentioned: messageContent.mentioned,
                pin: messageContent.pin,
                uniqueName: messageContent.uniqueName,
                userGroupHash: messageContent.userGroupHash
            };

            // Add inviter if exist
            if (messageContent.inviter) {
                conversation.inviter = formatDataToMakeParticipant(messageContent.inviter, messageContent.id);
            }

            // Add participants list if exist
            if (messageContent.participants && Array.isArray(messageContent.participants)) {
                conversation.participants = [];

                for (var i = 0; i < messageContent.participants.length; i++) {
                    var participantData = formatDataToMakeParticipant(messageContent.participants[i], messageContent.id);
                    if (participantData) {
                        conversation.participants.push(participantData);
                    }
                }
            }

            // Add lastMessageVO if exist
            if (messageContent.lastMessageVO) {
                conversation.lastMessageVO = formatDataToMakeMessage(messageContent.id, messageContent.lastMessageVO);
            }

            // Add pinMessageVO if exist
            if (messageContent.pinMessageVO) {
                conversation.pinMessageVO = formatDataToMakePinMessage(messageContent.id, messageContent.pinMessageVO);
            }

            // return conversation;
            return JSON.parse(JSON.stringify(conversation));
        },

        /**
         * Format Data To Make Reply Info
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} replyInfo Object
         */
        formatDataToMakeReplyInfo = function (messageContent, threadId) {
            /**
             * + replyInfoVO                  {object : replyInfoVO}
             *   - participant                {object : ParticipantVO}
             *   - repliedToMessageId         {long}
             *   - repliedToMessageTime       {long}
             *   - repliedToMessageNanos      {int}
             *   - message                    {string}
             *   - deleted                    {boolean}
             *   - messageType                {int}
             *   - metadata                   {string}
             *   - systemMetadata             {string}
             */

            var replyInfo = {
                participant: undefined,
                repliedToMessageId: messageContent.repliedToMessageId,
                repliedToMessageTime: (messageContent.repliedToMessageNanos)
                    ? (parseInt(parseInt(messageContent.repliedToMessageTime) / 1000) * 1000000000) + parseInt(messageContent.repliedToMessageNanos)
                    : (parseInt(messageContent.repliedToMessageTime)),
                repliedToMessageTimeMiliSeconds: parseInt(messageContent.repliedToMessageTime),
                repliedToMessageTimeNanos: parseInt(messageContent.repliedToMessageNanos),
                message: messageContent.message,
                deleted: messageContent.deleted,
                messageType: messageContent.messageType,
                metadata: messageContent.metadata,
                systemMetadata: messageContent.systemMetadata
            };

            if (messageContent.participant) {
                replyInfo.participant = formatDataToMakeParticipant(messageContent.participant, threadId);
            }

            // return replyInfo;
            return JSON.parse(JSON.stringify(replyInfo));
        },

        /**
         * Format Data To Make Forward Info
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} forwardInfo Object
         */
        formatDataToMakeForwardInfo = function (messageContent, threadId) {
            /**
             * + forwardInfo                  {object : forwardInfoVO}
             *   - participant                {object : ParticipantVO}
             *   - conversation               {object : ConversationSummary}
             */

            var forwardInfo = {
                participant: undefined,
                conversation: undefined
            };

            if (messageContent.conversation) {
                forwardInfo.conversation = formatDataToMakeConversation(messageContent.conversation);
            }

            if (messageContent.participant) {
                forwardInfo.participant = formatDataToMakeParticipant(messageContent.participant, threadId);
            }

            // return forwardInfo;
            return JSON.parse(JSON.stringify(forwardInfo));
        },

        /**
         * Format Data To Make Message
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} message Object
         */
        formatDataToMakeMessage = function (threadId, pushMessageVO, fromCache) {
            /**
             * + MessageVO                       {object}
             *    - id                           {long}
             *    - threadId                     {long}
             *    - ownerId                      {long}
             *    - uniqueId                     {string}
             *    - previousId                   {long}
             *    - message                      {string}
             *    - messageType                  {int}
             *    - edited                       {boolean}
             *    - editable                     {boolean}
             *    - deletable                    {boolean}
             *    - delivered                    {boolean}
             *    - seen                         {boolean}
             *    - mentioned                    {boolean}
             *    - pinned                       {boolean}
             *    - participant                  {object : ParticipantVO}
             *    - conversation                 {object : ConversationVO}
             *    - replyInfo                    {object : replyInfoVO}
             *    - forwardInfo                  {object : forwardInfoVO}
             *    - metadata                     {string}
             *    - systemMetadata               {string}
             *    - time                         {long}
             *    - timeNanos                    {long}
             */

            if (fromCache || pushMessageVO.time?.toString().length > 14) {
                var time = pushMessageVO.time,
                    timeMiliSeconds = parseInt(pushMessageVO.time / 1000000);
            } else {
                var time = (pushMessageVO.timeNanos)
                        ? (parseInt(parseInt(pushMessageVO.time) / 1000) * 1000000000) + parseInt(pushMessageVO.timeNanos)
                        : (parseInt(pushMessageVO.time)),
                    timeMiliSeconds = parseInt(pushMessageVO.time);
            }

            var message = {
                id: pushMessageVO.id,
                threadId: threadId,
                ownerId: (pushMessageVO.ownerId)
                    ? pushMessageVO.ownerId
                    : undefined,
                uniqueId: pushMessageVO.uniqueId,
                previousId: pushMessageVO.previousId,
                message: pushMessageVO.message,
                messageType: pushMessageVO.messageType,
                edited: pushMessageVO.edited,
                editable: pushMessageVO.editable,
                deletable: pushMessageVO.deletable,
                delivered: pushMessageVO.delivered,
                seen: pushMessageVO.seen,
                mentioned: pushMessageVO.mentioned,
                pinned: pushMessageVO.pinned,
                participant: undefined,
                conversation: undefined,
                replyInfo: undefined,
                forwardInfo: undefined,
                metadata: pushMessageVO.metadata,
                systemMetadata: pushMessageVO.systemMetadata,
                time: time,
                timeMiliSeconds: timeMiliSeconds,
                timeNanos: parseInt(pushMessageVO.timeNanos)
            };

            if (pushMessageVO.participant) {
                message.ownerId = pushMessageVO.participant.id;
            }

            if (pushMessageVO.conversation) {
                message.conversation = formatDataToMakeConversation(pushMessageVO.conversation);
                message.threadId = pushMessageVO.conversation.id;
            }

            if (pushMessageVO.replyInfoVO || pushMessageVO.replyInfo) {
                message.replyInfo = (pushMessageVO.replyInfoVO)
                    ? formatDataToMakeReplyInfo(pushMessageVO.replyInfoVO, threadId)
                    : formatDataToMakeReplyInfo(pushMessageVO.replyInfo, threadId);
            }

            if (pushMessageVO.forwardInfo) {
                message.forwardInfo = formatDataToMakeForwardInfo(pushMessageVO.forwardInfo, threadId);
            }

            if (pushMessageVO.participant) {
                message.participant = formatDataToMakeParticipant(pushMessageVO.participant, threadId);
            }

            // return message;
            return JSON.parse(JSON.stringify(message));
        },

        /**
         * Format Data To Make Pin Message
         *
         * This functions reformats given JSON to proper Object
         *
         * @access private
         *
         * @param {object}  messageContent    Json object of thread taken from chat server
         *
         * @return {object} pin message Object
         */

        formatDataToMakePinMessage = function (threadId, pushMessageVO) {
            /**
             * + PinMessageVO                    {object}
             *    - messageId                    {long}
             *    - time                         {long}
             *    - sender                       {long}
             *    - text                         {string}
             *    - notifyAll                    {boolean}
             */

            var pinMessage = {
                threadId: threadId,
                time: pushMessageVO.time,
                sender: pushMessageVO.sender,
                messageId: pushMessageVO.messageId,
                text: pushMessageVO.text
            };

            if (typeof pushMessageVO.notifyAll === 'boolean') {
                pinMessage.notifyAll = pushMessageVO.notifyAll
            }

            // return pinMessage;
            return JSON.parse(JSON.stringify(pinMessage));
        },

        /**
         * Reformat Thread History
         *
         * This functions reformats given Array of thread Messages
         * into proper chat message object
         *
         * @access private
         *
         * @param {long}    threadId         Id of Thread
         * @param {object}  historyContent   Array of Thread History Messages
         *
         * @return {object} Formatted Thread History
         */
        reformatThreadHistory = function (threadId, historyContent) {
            var returnData = [];

            for (var i = 0; i < historyContent.length; i++) {
                returnData.push(formatDataToMakeMessage(threadId, historyContent[i]));
            }

            return returnData;
        },

        /**
         * Reformat Thread Participants
         *
         * This functions reformats given Array of thread Participants
         * into proper thread participant
         *
         * @access private
         *
         * @param {object}  participantsContent   Array of Thread Participant Objects
         * @param {long}    threadId              Id of Thread
         *
         * @return {object} Formatted Thread Participant Array
         */
        reformatThreadParticipants = function (participantsContent, threadId) {
            var returnData = [];

            for (var i = 0; i < participantsContent.length; i++) {
                returnData.push(formatDataToMakeParticipant(participantsContent[i], threadId));
            }

            return returnData;
        },

        /**
         * Unset Not Seen Duration
         *
         * This functions unsets notSeenDuration property of cached objects
         *
         * @access private
         *
         * @param {object}  content   Object or Array to be modified
         *
         * @return {object}
         */
        unsetNotSeenDuration = function (content) {
            /**
             * Make a copy from original object to modify it's
             * attributes, because we don't want to change
             * the original object
             */
            var temp = cloneObject(content);

            if (temp.hasOwnProperty('notSeenDuration')) {
                temp.notSeenDuration = undefined;
            }

            if (temp.hasOwnProperty('inviter')) {
                temp.inviter.notSeenDuration = undefined;
            }

            if (temp.hasOwnProperty('participant')) {
                temp.participant.notSeenDuration = undefined;
            }

            return temp;
        },

        /**
         * Clone Object/Array
         *
         * This functions makes a deep clone of given object or array
         *
         * @access private
         *
         * @param {object}  original   Object or Array to be cloned
         *
         * @return {object} Cloned object
         */
        cloneObject = function (original) {
            var out, value, key;
            out = Array.isArray(original) ? [] : {};

            for (key in original) {
                value = original[key];
                out[key] = (typeof value === 'object' && value !== null)
                    ? cloneObject(value)
                    : value;
            }

            return out;
        },

        /**
         * Get Treads.
         *
         * This functions gets threads list
         *
         * @access private
         *
         * @param {int}       count                 count of threads to be received
         * @param {int}       offset                offset of select query
         * @param {array}     threadIds             An array of thread ids to be received
         * @param {string}    name                  Search term to look up in thread Titles
         * @param {long}      creatorCoreUserId     SSO User Id of thread creator
         * @param {long}      partnerCoreUserId     SSO User Id of thread partner
         * @param {long}      partnerCoreContactId  Contact Id of thread partner
         * @param {function}  callback              The callback function to call after
         *
         * @return {object} Instant sendMessage result
         */
        getThreads = function (params, callback) {
            var count = 50,
                offset = 0,
                content = {},
                whereClause = {},
                returnCache = false;

            if (params) {
                if (parseInt(params.count) > 0) {
                    count = params.count;
                }

                if (parseInt(params.offset) > 0) {
                    offset = params.offset;
                }

                if (typeof params.threadName === 'string') {
                    content.name = whereClause.name = params.threadName;
                }

                if (Array.isArray(params.threadIds)) {
                    content.threadIds = whereClause.threadIds = params.threadIds;
                }

                if (typeof params.new === 'boolean') {
                    content.new = params.new;
                }

                if (parseInt(params.creatorCoreUserId) > 0) {
                    content.creatorCoreUserId = whereClause.creatorCoreUserId = params.creatorCoreUserId;
                }

                if (parseInt(params.partnerCoreUserId) > 0) {
                    content.partnerCoreUserId = whereClause.partnerCoreUserId = params.partnerCoreUserId;
                }

                if (parseInt(params.partnerCoreContactId) > 0) {
                    content.partnerCoreContactId = whereClause.partnerCoreContactId = params.partnerCoreContactId;
                }

                if (typeof params.isGroup === 'boolean') {
                    content.isGroup = params.isGroup;
                }
                if (typeof params.type === 'number') {
                    content.type = params.type;
                }

                var functionLevelCache = (typeof params.cache == 'boolean') ? params.cache : true;
            }

            content.count = count;
            content.offset = offset;

            var sendMessageParams = {
                chatMessageVOType: chatMessageVOTypes.GET_THREADS,
                typeCode: params.typeCode,
                content: content
            };

            /**
             * Retrive get threads response from server
             */
            return app.messenger.sendMessage(sendMessageParams, {
                onResult: function (result) {
                    var returnData = {
                        hasError: result.hasError,
                        cache: false,
                        errorMessage: result.errorMessage,
                        errorCode: result.errorCode
                    };

                    if (!returnData.hasError) {

                        var messageContent = result.result,
                            messageLength = messageContent.length,
                            resultData = {
                                threads: [],
                                contentCount: result.contentCount,
                                hasNext: messageContent && !(messageLength < count),//(offset + count < result.contentCount && messageLength > 0),
                                nextOffset: offset + messageLength
                            },
                            threadData;

                        for (var i = 0; i < messageLength; i++) {
                            threadData = createThread(messageContent[i], false);
                            if (threadData) {
                                resultData.threads.push(threadData);
                            }
                        }

                        returnData.result = resultData;
                    }

                    callback && callback(returnData);
                    /**
                     * Delete callback so if server pushes response before
                     * cache, cache won't send data again
                     */
                    callback = undefined;

                    if (!returnData.hasError && returnCache) {
                        app.chatEvents.fireEvent('threadEvents', {
                            type: 'THREADS_LIST_CHANGE',
                            result: returnData.result
                        });
                    }
                }
            });
        },

        getAllThreads = function (params, callback) {
            var sendMessageParams = {
                chatMessageVOType: chatMessageVOTypes.GET_THREADS,
                typeCode: params.typeCode,
                content: {}
            };

            if (params) {
                if (typeof params.summary == 'boolean') {
                    sendMessageParams.content.summary = params.summary;
                }
            }

            return app.messenger.sendMessage(sendMessageParams, {
                onResult: function (result) {

                    // if (!result.hasError) {}

                    callback && callback(result);
                }
            });
        },

        /**
         * Get History.
         *
         * This functions gets history of a thread
         *
         * @access private
         *
         * @param {int}       count             Count of threads to be received
         * @param {int}       offset            Offset of select query
         * @param {long}      threadId          Id of thread to get its history
         * @param {long}      id                Id of single message to get
         * @param {long}      userId            Messages of this SSO User
         * @param {int}       messageType       Type of messages to get (types should be set by client)
         * @param {long}      fromTime          Get messages which have bigger time than given fromTime
         * @param {int}       fromTimeNanos     Get messages which have bigger time than given fromTimeNanos
         * @param {long}      toTime            Get messages which have smaller time than given toTime
         * @param {int}       toTimeNanos       Get messages which have smaller time than given toTimeNanos
         * @param {long}      senderId          Messages of this sender only
         * @param {string}    uniqueIds         Array of unique ids to retrieve
         * @param {string}    order             Order of select query (default: DESC)
         * @param {string}    query             Search term to be looked up in messages content
         * @param {object}    metadataCriteria  This JSON will be used to search in message metadata with GraphQL
         * @param {function}  callback          The callback function to call after
         *
         * @return {object} Instant result of sendMessage
         */
        getHistory = function (params, callback) {
            if (parseInt(params.threadId) > 0) {
                var sendMessageParams = {
                        chatMessageVOType: chatMessageVOTypes.GET_HISTORY,
                        typeCode: params.typeCode,
                        content: {},
                        subjectId: params.threadId
                    },
                    whereClause = {},
                    offset = (parseInt(params.offset) > 0) ? parseInt(params.offset) : 0,
                    count = (parseInt(params.count) > 0) ? parseInt(params.count) : config.getHistoryCount,
                    order = (typeof params.order != 'undefined') ? (params.order).toLowerCase() : 'desc',
                    functionLevelCache = (typeof params.cache == 'boolean') ? params.cache : true,
                    messages,
                    dynamicHistoryCount = (params.dynamicHistoryCount && typeof params.dynamicHistoryCount === 'boolean')
                        ? params.dynamicHistoryCount
                        : false,
                    sendingQueue = (params.queues && typeof params.queues.sending === 'boolean')
                        ? params.queues.sending
                        : true,
                    failedQueue = (params.queues && typeof params.queues.failed === 'boolean')
                        ? params.queues.failed
                        : true,
                    uploadingQueue = (params.queues && typeof params.queues.uploading === 'boolean')
                        ? params.queues.uploading
                        : true,
                    sendingQueueMessages = [],
                    failedQueueMessages = [],
                    uploadingQueueMessages = [];

                if (sendingQueue) {
                    getChatSendQueue(parseInt(params.threadId), function (sendQueueMessages) {
                        for (var i = 0; i < sendQueueMessages.length; i++) {
                            var time = new Date().getTime();

                            sendingQueueMessages.push(formatDataToMakeMessage(sendQueueMessages[i].threadId, {
                                uniqueId: sendQueueMessages[i].uniqueId,
                                ownerId: app.store.user.get().id,
                                message: sendQueueMessages[i].content,
                                metadata: sendQueueMessages[i].metadata,
                                systemMetadata: sendQueueMessages[i].systemMetadata,
                                replyInfo: sendQueueMessages[i].replyInfo,
                                forwardInfo: sendQueueMessages[i].forwardInfo,
                                time: time,
                                timeNanos: (time % 1000) * 1000000
                            }));
                        }
                    });
                }

                if (uploadingQueue) {
                    app.messageQueues.getChatUploadQueue(parseInt(params.threadId), function (uploadQueueMessages) {
                        for (var i = 0; i < uploadQueueMessages.length; i++) {
                            uploadQueueMessages[i].message.participant = app.store.user.get();
                            var time = new Date().getTime();
                            uploadQueueMessages[i].message.time = time;
                            uploadQueueMessages[i].message.timeNanos = (time % 1000) * 1000000;
                            uploadingQueueMessages.push(formatDataToMakeMessage(params.threadId, uploadQueueMessages[i].message, false));
                        }
                    });
                }

                app.messageQueues.getChatWaitQueue(parseInt(params.threadId), failedQueue, function (waitQueueMessages) {
                    if (cacheSecret.length > 0) {
                        for (var i = 0; i < waitQueueMessages.length; i++) {
                            var decryptedEnqueuedMessage = Utility.jsonParser(chatDecrypt(waitQueueMessages[i].message, cacheSecret));
                            var time = new Date().getTime();
                            waitQueueMessages[i] = formatDataToMakeMessage(waitQueueMessages[i].threadId,
                                {
                                    uniqueId: decryptedEnqueuedMessage.uniqueId,
                                    ownerId: app.store.user.get().id,
                                    message: decryptedEnqueuedMessage.content,
                                    metadata: decryptedEnqueuedMessage.metadata,
                                    systemMetadata: decryptedEnqueuedMessage.systemMetadata,
                                    replyInfo: decryptedEnqueuedMessage.replyInfo,
                                    forwardInfo: decryptedEnqueuedMessage.forwardInfo,
                                    participant: app.store.user.get(),
                                    time: time,
                                    timeNanos: (time % 1000) * 1000000
                                });
                        }

                        failedQueueMessages = waitQueueMessages;
                    } else {
                        failedQueueMessages = [];
                    }

                    if (dynamicHistoryCount) {
                        var tempCount = count - (sendingQueueMessages.length + failedQueueMessages.length + uploadingQueueMessages.length);
                        sendMessageParams.content.count = (tempCount > 0) ? tempCount : 0;
                    } else {
                        sendMessageParams.content.count = count;
                    }

                    sendMessageParams.content.offset = offset;
                    sendMessageParams.content.order = order;

                    if (parseInt(params.messageId) > 0) {
                        sendMessageParams.content.id = whereClause.id = params.messageId;
                    }

                    if (Array.isArray(params.uniqueIds)) {
                        sendMessageParams.content.uniqueIds = params.uniqueIds;
                    }

                    if (parseInt(params.fromTimeFull) > 0 && params.fromTimeFull.toString().length == 19) {
                        sendMessageParams.content.fromTime = whereClause.fromTime = parseInt(params.fromTimeFull.toString()
                            .substring(0, 13));
                        sendMessageParams.content.fromTimeNanos = whereClause.fromTimeNanos = parseInt(params.fromTimeFull.toString()
                            .substring(10, 19));
                    } else {
                        if (parseInt(params.fromTime) > 0 && parseInt(params.fromTime) < 9999999999999) {
                            sendMessageParams.content.fromTime = whereClause.fromTime = parseInt(params.fromTime);
                        }

                        if (parseInt(params.fromTimeNanos) > 0 && parseInt(params.fromTimeNanos) < 999999999) {
                            sendMessageParams.content.fromTimeNanos = whereClause.fromTimeNanos = parseInt(params.fromTimeNanos);
                        }
                    }

                    if (parseInt(params.toTimeFull) > 0 && params.toTimeFull.toString().length == 19) {
                        sendMessageParams.content.toTime = whereClause.toTime = parseInt(params.toTimeFull.toString()
                            .substring(0, 13));
                        sendMessageParams.content.toTimeNanos = whereClause.toTimeNanos = parseInt(params.toTimeFull.toString()
                            .substring(10, 19));
                    } else {
                        if (parseInt(params.toTime) > 0 && parseInt(params.toTime) < 9999999999999) {
                            sendMessageParams.content.toTime = whereClause.toTime = parseInt(params.toTime);
                        }

                        if (parseInt(params.toTimeNanos) > 0 && parseInt(params.toTimeNanos) < 999999999) {
                            sendMessageParams.content.toTimeNanos = whereClause.toTimeNanos = parseInt(params.toTimeNanos);
                        }
                    }

                    if (typeof params.query != 'undefined') {
                        sendMessageParams.content.query = whereClause.query = params.query;
                    }

                    if (params.allMentioned && typeof params.allMentioned == 'boolean') {
                        sendMessageParams.content.allMentioned = whereClause.allMentioned = params.allMentioned;
                    }

                    if (params.unreadMentioned && typeof params.unreadMentioned == 'boolean') {
                        sendMessageParams.content.unreadMentioned = whereClause.unreadMentioned = params.unreadMentioned;
                    }

                    if (params.messageType && params.messageType.toUpperCase() !== undefined && chatMessageTypes[params.messageType.toUpperCase()] > 0) {
                        sendMessageParams.content.messageType = whereClause.messageType = chatMessageTypes[params.messageType.toUpperCase()];
                    }

                    if (typeof params.metadataCriteria == 'object' && params.metadataCriteria.hasOwnProperty('field')) {
                        sendMessageParams.content.metadataCriteria = whereClause.metadataCriteria = params.metadataCriteria;
                    }

                    if (typeof params.onlyNewMessages === "boolean") {
                        sendMessageParams.content.newMessages = params.onlyNewMessages;
                    }

                    /**
                     * Get Thread Messages From Server
                     */
                    return app.messenger.sendMessage(sendMessageParams, {
                        onResult: function (result) {

                            var returnData = {
                                    hasError: result.hasError,
                                    cache: false,
                                    errorMessage: result.errorMessage,
                                    errorCode: result.errorCode
                                },
                                resultMessagesId = [];

                            if (!returnData.hasError) {
                                var messageContent = result.result,
                                    messageLength = messageContent.length;

                                var history = reformatThreadHistory(params.threadId, messageContent);

                                if (messageLength > 0) {
                                    /**
                                     * Calculating First and Last Messages of result
                                     */
                                    var lastMessage = history[messageContent.length - 1],
                                        firstMessage = history[0];

                                    /**
                                     * Sending Delivery for Last Message of Thread
                                     */
                                    // if (lastMessage.id > 0 && !lastMessage.delivered) {
                                    //     deliver({
                                    //         messageId: lastMessage.id,
                                    //         ownerId: lastMessage.participant.id
                                    //     });
                                    // }
                                }

                                var resultData = {
                                    history: history,
                                    contentCount: result.contentCount,
                                    hasNext: result.result && !(result.result.length < sendMessageParams.content.count),//(sendMessageParams.content.offset + sendMessageParams.content.count < result.contentCount &&messageLength > 0),
                                    nextOffset: sendMessageParams.content.offset + messageLength
                                };

                                returnData.result = resultData;

                                if (sendingQueue) {
                                    returnData.result.sending = sendingQueueMessages;
                                }
                                if (uploadingQueue) {
                                    returnData.result.uploading = uploadingQueueMessages;
                                }
                                if (failedQueue) {
                                    returnData.result.failed = failedQueueMessages;
                                }

                                callback && callback(returnData);
                                callback = undefined;
                            }
                        }
                    });
                });
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] Thread ID is required for Getting history!'
                });
                return;
            }
        },

        /**
         * Update Thread Info
         *
         * This functions updates metadata of thread
         *
         * @access private
         *
         * @param {int}       threadId      Id of thread
         * @param {string}    image         URL og thread image to be set
         * @param {string}    description   Description for thread
         * @param {string}    title         New Title for thread
         * @param {object}    metadata      New Metadata to be set on thread
         * @param {function}  callback      The callback function to call after
         *
         * @return {object} Instant sendMessage result
         */
        updateThreadInfo = function (params, callback) {
            var updateThreadInfoData = {
                    chatMessageVOType: chatMessageVOTypes.UPDATE_THREAD_INFO,
                    typeCode: params.typeCode,
                    content: {},
                    pushMsgType: 3,
                    token: app.sdkParams.token
                },
                threadInfoContent = {},
                fileUploadParams = {},
                metadata = {},
                threadId,
                fileUniqueId = Utility.generateUUID();

            if (params) {
                if (parseInt(params.threadId) > 0) {
                    threadId = parseInt(params.threadId);
                    updateThreadInfoData.subjectId = threadId;
                } else {
                    app.chatEvents.fireEvent('error', {
                        code: 999,
                        message: '[SDK] Thread ID is required for Updating thread info!'
                    });
                }

                if (typeof params.description == 'string') {
                    threadInfoContent.description = params.description;
                }

                if (typeof params.title == 'string') {
                    threadInfoContent.name = params.title;
                }

                if (typeof params.metadata == 'object') {
                    threadInfoContent.metadata = params.metadata;
                } else if (typeof params.metadata == 'string') {
                    try {
                        threadInfoContent.metadata = JSON.parse(params.metadata);
                    } catch (e) {
                        threadInfoContent.metadata = {};
                    }
                } else {
                    threadInfoContent.metadata = {};
                }

                updateThreadInfoData.content = threadInfoContent;

                if (typeof params.image == 'object' && params.image.size > 0) {
                    return chatUploadHandler({
                        threadId: threadId,
                        file: params.image,
                        fileUniqueId: fileUniqueId
                    }, function (uploadHandlerResult, uploadHandlerMetadata, fileType, fileExtension) {
                        fileUploadParams = Object.assign(fileUploadParams, uploadHandlerResult);
                        threadInfoContent.metadata = JSON.stringify(Object.assign(threadInfoContent.metadata, uploadHandlerMetadata));
                        putInChatUploadQueue({
                            message: {
                                chatMessageVOType: chatMessageVOTypes.UPDATE_THREAD_INFO,
                                typeCode: params.typeCode,
                                subjectId: threadId,
                                content: threadInfoContent,
                                metadata: threadInfoContent.metadata,
                                uniqueId: fileUniqueId,
                                pushMsgType: 3,
                                token: app.sdkParams.token
                            },
                            callbacks: callback
                        }, function () {
                            if (imageMimeTypes.indexOf(fileType) >= 0 || imageExtentions.indexOf(fileExtension) >= 0) {
                                app.fileMethods.uploadImageToPodspace(fileUploadParams, function (result) {
                                    if (!result.hasError) {
                                        metadata['fileHash'] = result.result.hashCode;
                                        transferFromUploadQToSendQ(parseInt(params.threadId), fileUniqueId, JSON.stringify(metadata), function () {
                                            chatSendQueueHandler();
                                        });
                                    } else {
                                        app.messageQueues.deleteFromChatUploadQueue({message: {uniqueId: fileUniqueId}});
                                    }
                                });
                            } else {
                                app.chatEvents.fireEvent('error', {
                                    code: 999,
                                    message: '[SDK] Thread picture can be a image type only!'
                                });
                            }
                        });
                    });
                } else if (typeof params.image == 'string' && params.image.length > 5) {
                    threadInfoContent.metadata = JSON.stringify(Object.assign(threadInfoContent.metadata, {fileHash: params.image}));

                    app.fileMethods.getImageDownloadLinkFromPodspace({
                        hashCode: params.image
                    }, function (result) {
                        if (!result.hasError) {
                            threadInfoContent.image = result.downloadUrl;
                        }
                    });

                    return app.messenger.sendMessage({
                        chatMessageVOType: chatMessageVOTypes.UPDATE_THREAD_INFO,
                        typeCode: params.typeCode,
                        subjectId: threadId,
                        content: threadInfoContent,
                        metadata: threadInfoContent.metadata,
                        uniqueId: fileUniqueId,
                        pushMsgType: 3,
                        token: app.sdkParams.token
                    }, {
                        onResult: function (result) {
                            callback && callback(result);
                        }
                    });
                } else {
                    if (Object.keys(threadInfoContent.metadata).length == 0) {
                        delete threadInfoContent.metadata;
                    }

                    return app.messenger.sendMessage({
                        chatMessageVOType: chatMessageVOTypes.UPDATE_THREAD_INFO,
                        typeCode: params.typeCode,
                        subjectId: threadId,
                        content: threadInfoContent,
                        metadata: threadInfoContent.metadata,
                        uniqueId: fileUniqueId,
                        pushMsgType: 3,
                        token: app.sdkParams.token
                    }, {
                        onResult: function (result) {
                            callback && callback(result);
                        }
                    });
                }
            }
        },

        /**
         * Update Thread Info
         *
         * This functions updates metadata of thread
         *
         * @access private
         *
         * @param {int}       threadId      Id of thread
         * @param {string}    image         URL og thread image to be set
         * @param {string}    description   Description for thread
         * @param {string}    title         New Title for thread
         * @param {object}    metadata      New Metadata to be set on thread
         * @param {function}  callback      The callback function to call after
         *
         * @return {object} Instant sendMessage result
         */
        updateChatProfile = function (params, callback) {
            var updateChatProfileData = {
                chatMessageVOType: chatMessageVOTypes.UPDATE_CHAT_PROFILE,
                content: {},
                pushMsgType: 3,
                token: app.sdkParams.token
            };

            if (params) {
                if (typeof params.bio == 'string') {
                    updateChatProfileData.content.bio = params.bio;
                }

                if (typeof params.metadata == 'object') {
                    updateChatProfileData.content.metadata = JSON.stringify(params.metadata);
                } else if (typeof params.metadata == 'string') {
                    updateChatProfileData.content.metadata = params.metadata;
                }
            }

            return app.messenger.sendMessage(updateChatProfileData, {
                onResult: function (result) {
                    callback && callback(result);
                }
            });
        },

        /**
         * Get Participant Roles
         *
         * This functions retrieves roles of an user if they are
         * part of the thread
         *
         * @access private
         *
         * @param {int}       threadId      Id of thread
         * @param {function}  callback      The callback function to call after
         *
         * @return {object} Instant sendMessage result
         */
        getCurrentUserRoles = function (params, callback) {
            var updateChatProfileData = {
                chatMessageVOType: chatMessageVOTypes.GET_PARTICIPANT_ROLES,
                pushMsgType: 3,
                subjectId: params.threadId,
                token: app.sdkParams.token
            };

            return app.messenger.sendMessage(updateChatProfileData, {
                onResult: function (result) {
                    callback && callback(result);
                }
            });
        },

        /**
         * Get Thread Participants
         *
         * Gets participants list of given thread
         *
         * @access pubic
         *
         * @param {int}     threadId        Id of thread which you want to get participants of
         * @param {int}     count           Count of objects to get
         * @param {int}     offset          Offset of select Query
         * @param {string}  name            Search in Participants list (LIKE in name, contactName, email)
         *
         * @return {object} Instant Response
         */
        getThreadParticipants = function (params, callback) {
            var sendMessageParams = {
                    chatMessageVOType: chatMessageVOTypes.THREAD_PARTICIPANTS,
                    typeCode: params.typeCode,
                    content: {},
                    subjectId: params.threadId
                },
                whereClause = {},
                returnCache = false;

            var offset = (parseInt(params.offset) > 0)
                    ? parseInt(params.offset)
                    : 0,
                count = (parseInt(params.count) > 0)
                    ? parseInt(params.count)
                    : config.getHistoryCount;

            sendMessageParams.content.count = count;
            sendMessageParams.content.offset = offset;

            if (typeof params.name === 'string') {
                sendMessageParams.content.name = whereClause.name = params.name;
            }

            if (typeof params.username === 'string') {
                sendMessageParams.content.username = whereClause.username = params.username;
            }
            if (typeof params.cellphoneNumber === 'string') {
                sendMessageParams.content.cellphoneNumber = whereClause.cellphoneNumber = params.cellphoneNumber;
            }

            if (typeof params.admin === 'boolean') {
                sendMessageParams.content.admin = params.admin;
            }

            var functionLevelCache = (typeof params.cache == 'boolean') ? params.cache : true;

            return app.messenger.sendMessage(sendMessageParams, {
                onResult: function (result) {
                    var returnData = {
                        hasError: result.hasError,
                        cache: false,
                        errorMessage: result.errorMessage,
                        errorCode: result.errorCode
                    };

                    if (!returnData.hasError) {
                        var messageContent = result.result,
                            messageLength = messageContent.length,
                            resultData = {
                                participants: reformatThreadParticipants(messageContent, params.threadId),
                                contentCount: result.contentCount,
                                hasNext: (sendMessageParams.content.offset + sendMessageParams.content.count < result.contentCount && messageLength > 0),
                                nextOffset: sendMessageParams.content.offset + messageLength
                            };

                        returnData.result = resultData;
                    }

                    callback && callback(returnData);
                    /**
                     * Delete callback so if server pushes response before
                     * cache, cache won't send data again
                     */
                    callback = undefined;
                }
            });
        },

        /**
         * Deliver
         *
         * This functions sends delivery messages for a message
         *
         * @access private
         *
         * @param {int}    ownerId    Id of Message owner
         * @param {long}   messageId  Id of Message
         *
         * @return {object} Instant sendMessage result
         */
        deliver = function (params) {
            if (app.store.user.get() && params.ownerId !== app.store.user.get().id) {
                return app.messenger.sendMessage({
                    chatMessageVOType: chatMessageVOTypes.DELIVERY,
                    typeCode: params.typeCode,
                    content: params.messageId,
                    pushMsgType: 3
                });
            }
        },

        /**
         * Get Image.
         *
         * This functions gets an uploaded image from File Server.
         *
         * @since 3.9.9
         * @access private
         *
         * @param {long}    imageId         ID of image
         * @param {int}     width           Required width to get
         * @param {int}     height          Required height to get
         * @param {boolean} actual          Required height to get
         * @param {boolean} downloadable    TRUE to be downloadable / FALSE to not
         * @param {string}  hashCode        HashCode of uploaded file
         *
         * @return {object} Image Object
         */
        getImage = function (params, callback) {
            let getImageData = {};

            if (params) {
                if (parseInt(params.imageId) > 0) {
                    getImageData.imageId = params.imageId;
                }

                if (typeof params.hashCode == 'string') {
                    getImageData.hashCode = params.hashCode;
                }

                if (parseInt(params.width) > 0) {
                    getImageData.width = params.width;
                }

                if (parseInt(params.height) > 0) {
                    getImageData.height = params.height;
                }

                if (parseInt(params.actual) > 0) {
                    getImageData.actual = params.actual;
                }

                if (parseInt(params.downloadable) > 0) {
                    getImageData.downloadable = params.downloadable;
                }
            }

            app.httpRequest.httpRequest({
                url: app.sdkParams.SERVICE_ADDRESSES.FILESERVER_ADDRESS + SERVICES_PATH.GET_IMAGE,
                method: 'GET',
                data: getImageData
            }, function (result) {
                if (!result.hasError) {
                    var queryString = '?';
                    for (var i in params) {
                        queryString += i + '=' + params[i] + '&';
                    }
                    queryString = queryString.slice(0, -1);
                    var image = app.sdkParams.SERVICE_ADDRESSES.FILESERVER_ADDRESS + SERVICES_PATH.GET_IMAGE + queryString;
                    callback({
                        hasError: result.hasError,
                        result: image
                    });
                } else {
                    callback({
                        hasError: true
                    });
                }
            });
        },

        /**
         * Get File.
         *
         * This functions gets an uploaded file from File Server.
         *
         * @since 3.9.9
         * @access private
         *
         * @param {long}    fileId          ID of file
         * @param {boolean} downloadable    TRUE to be downloadable / False to not
         * @param {string}  hashCode        HashCode of uploaded file
         *
         * @return {object} File Object
         */
        getFile = function (params, callback) {
            let getFileData = {};

            if (params) {
                if (typeof params.fileId != 'undefined') {
                    getFileData.fileId = params.fileId;
                }

                if (typeof params.hashCode == 'string') {
                    getFileData.hashCode = params.hashCode;
                }

                if (typeof params.downloadable == 'boolean') {
                    getFileData.downloadable = params.downloadable;
                }
            }

            app.httpRequest.httpRequest({
                url: app.sdkParams.SERVICE_ADDRESSES.FILESERVER_ADDRESS +
                    SERVICES_PATH.GET_FILE,
                method: 'GET',
                data: getFileData
            }, function (result) {
                if (!result.hasError) {
                    var queryString = '?';
                    for (var i in params) {
                        queryString += i + '=' + params[i] + '&';
                    }
                    queryString = queryString.slice(0, -1);
                    var file = app.sdkParams.SERVICE_ADDRESSES.FILESERVER_ADDRESS + SERVICES_PATH.GET_FILE + queryString;
                    callback({
                        hasError: result.hasError,
                        result: file
                    });
                } else {
                    callback({
                        hasError: true
                    });
                }
            });
        },

        /**
         * Upload File
         *
         * Upload files to File Server
         *
         * @since 3.9.9
         * @access private
         *
         * @param {string}  fileName        A name for the file
         * @param {file}    file            FILE: the file
         *
         * @link http://docs.pod.land/v1.0.8.0/Developer/CustomPost/605/File
         *
         * @return {object} Uploaded File Object
         */
        uploadFile = function (params, callback) {
            var fileName,
                fileType,
                fileSize,
                fileExtension,
                uploadUniqueId,
                uploadThreadId;

            if (Utility.isNode()) {
                fileName = params.file.split('/')
                    .pop();
                fileType = Mime.getType(params.file);
                fileSize = app.fs.statSync(params.file).size;
                fileExtension = params.file.split('.')
                    .pop();
            } else {
                fileName = params.file.name;
                fileType = params.file.type;
                fileSize = params.file.size;
                fileExtension = params.file.name.split('.')
                    .pop();
            }

            var uploadFileData = {};

            if (params) {
                if (typeof params.file !== 'undefined') {
                    uploadFileData.file = params.file;
                }

                if (params.randomFileName) {
                    uploadFileData.fileName = Utility.generateUUID() + '.' + fileExtension;
                } else {
                    uploadFileData.fileName = fileName;
                }

                uploadFileData.fileSize = fileSize;

                if (parseInt(params.threadId) > 0) {
                    uploadThreadId = params.threadId;
                    uploadFileData.threadId = params.threadId;
                } else {
                    uploadThreadId = 0;
                    uploadFileData.threadId = 0;
                }

                if (typeof params.uniqueId == 'string') {
                    uploadUniqueId = params.uniqueId;
                    uploadFileData.uniqueId = params.uniqueId;
                } else {
                    uploadUniqueId = Utility.generateUUID();
                    uploadFileData.uniqueId = uploadUniqueId;
                }

                if (typeof params.originalFileName == 'string') {
                    uploadFileData.originalFileName = params.originalFileName;
                } else {
                    uploadFileData.originalFileName = fileName;
                }
            }

            app.httpRequest.httpRequest({
                url: app.sdkParams.SERVICE_ADDRESSES.FILESERVER_ADDRESS + SERVICES_PATH.UPLOAD_FILE,
                method: 'POST',
                headers: {
                    '_token_': app.sdkParams.token,
                    '_token_issuer_': 1
                },
                data: uploadFileData,
                uniqueId: uploadUniqueId
            }, function (result) {
                if (!result.hasError) {
                    try {
                        var response = (typeof result.result.responseText == 'string')
                            ? JSON.parse(result.result.responseText)
                            : result.result.responseText;
                        callback({
                            hasError: response.hasError,
                            result: response.result
                        });
                    } catch (e) {
                        callback({
                            hasError: true,
                            errorCode: 999,
                            errorMessage: '[SDK] Problem in Parsing result'
                        });
                    }
                } else {
                    callback({
                        hasError: true,
                        errorCode: result.errorCode,
                        errorMessage: result.errorMessage
                    });
                }
            });

            return {
                uniqueId: uploadUniqueId,
                threadId: uploadThreadId,
                participant: app.store.user.get(),
                content: {
                    caption: params.content,
                    file: {
                        uniqueId: uploadUniqueId,
                        fileName: fileName,
                        fileSize: fileSize,
                        fileObject: params.file
                    }
                }
            };
        },



        /**
         * Upload Image
         *
         * Upload images to Image Server
         *
         * @since 3.9.9
         * @access private
         *
         * @param {string}  fileName        A name for the file
         * @param {file}    image           FILE: the image file  (if its an image file)
         * @param {float}   xC              Crop Start point x    (if its an image file)
         * @param {float}   yC              Crop Start point Y    (if its an image file)
         * @param {float}   hC              Crop size Height      (if its an image file)
         * @param {float}   wC              Crop size Weight      (if its an image file)
         *
         * @link http://docs.pod.land/v1.0.8.0/Developer/CustomPost/215/UploadImage
         *
         * @return {object} Uploaded Image Object
         */
        uploadImage = function (params, callback) {
            var fileName,
                fileType,
                fileSize,
                fileExtension,
                uploadUniqueId,
                uploadThreadId;

            if (Utility.isNode()) {
                fileName = params.image.split('/')
                    .pop();
                fileType = Mime.getType(params.image);
                fileSize = app.fs.statSync(params.image).size;
                fileExtension = params.image.split('.')
                    .pop();
            } else {
                fileName = params.image.name;
                fileType = params.image.type;
                fileSize = params.image.size;
                fileExtension = params.image.name.split('.')
                    .pop();
            }

            if (imageMimeTypes.indexOf(fileType) >= 0 || imageExtentions.indexOf(fileExtension) >= 0) {
                let uploadImageData = {};

                if (params) {
                    if (typeof params.image !== 'undefined') {
                        uploadImageData.image = params.image;
                        uploadImageData.file = params.image;
                    }

                    if (params.randomFileName) {
                        uploadImageData.fileName = Utility.generateUUID() + '.' + fileExtension;
                    } else {
                        uploadImageData.fileName = fileName;
                    }

                    uploadImageData.fileSize = fileSize;

                    if (parseInt(params.threadId) > 0) {
                        uploadThreadId = params.threadId;
                        uploadImageData.threadId = params.threadId;
                    } else {
                        uploadThreadId = 0;
                        uploadImageData.threadId = 0;
                    }

                    if (typeof params.uniqueId == 'string') {
                        uploadUniqueId = params.uniqueId;
                        uploadImageData.uniqueId = params.uniqueId;
                    } else {
                        uploadUniqueId = Utility.generateUUID();
                        uploadImageData.uniqueId = uploadUniqueId;
                    }

                    if (typeof params.originalFileName == 'string') {
                        uploadImageData.originalFileName = params.originalFileName;
                    } else {
                        uploadImageData.originalFileName = fileName;
                    }

                    if (parseInt(params.xC) > 0) {
                        uploadImageData.xC = params.xC;
                    }

                    if (parseInt(params.yC) > 0) {
                        uploadImageData.yC = params.yC;
                    }

                    if (parseInt(params.hC) > 0) {
                        uploadImageData.hC = params.hC;
                    }

                    if (parseInt(params.wC) > 0) {
                        uploadImageData.wC = params.wC;
                    }
                }

                app.httpRequest.httpRequest({
                    url: app.sdkParams.SERVICE_ADDRESSES.FILESERVER_ADDRESS + SERVICES_PATH.UPLOAD_IMAGE,
                    method: 'POST',
                    headers: {
                        '_token_': app.sdkParams.token,
                        '_token_issuer_': 1
                    },
                    data: uploadImageData,
                    uniqueId: uploadUniqueId
                }, function (result) {
                    if (!result.hasError) {
                        try {
                            var response = (typeof result.result.responseText == 'string')
                                ? JSON.parse(result.result.responseText)
                                : result.result.responseText;
                            if (typeof response.hasError !== 'undefined' && !response.hasError) {
                                callback({
                                    hasError: response.hasError,
                                    result: response.result
                                });
                            } else {
                                callback({
                                    hasError: true,
                                    errorCode: response.errorCode,
                                    errorMessage: response.message
                                });
                            }
                        } catch (e) {
                            callback({
                                hasError: true,
                                errorCode: 6300,
                                errorMessage: CHAT_ERRORS[6300]
                            });
                        }
                    } else {
                        callback({
                            hasError: true,
                            errorCode: result.errorCode,
                            errorMessage: result.errorMessage
                        });
                    }
                });

                return {
                    uniqueId: uploadUniqueId,
                    threadId: uploadThreadId,
                    participant: app.store.user.get(),
                    content: {
                        caption: params.content,
                        file: {
                            uniqueId: uploadUniqueId,
                            fileName: fileName,
                            fileSize: fileSize,
                            fileObject: params.file
                        }
                    }
                };
            } else {
                callback({
                    hasError: true,
                    errorCode: 6301,
                    errorMessage: CHAT_ERRORS[6301]
                });
            }
        },

        sendFileMessage = function (params, callbacks) {
            var metadata = {file: {}},
                fileUploadParams = {},
                fileUniqueId = (typeof params.fileUniqueId == 'string' && params.fileUniqueId.length > 0) ? params.fileUniqueId : Utility.generateUUID();
            if (params) {
                if (!params.userGroupHash || params.userGroupHash.length == 0 || typeof (params.userGroupHash) != 'string') {
                    app.chatEvents.fireEvent('error', {
                        code: 6304,
                        message: CHAT_ERRORS[6304]
                    });
                    return;
                } else {
                    fileUploadParams.userGroupHash = params.userGroupHash;
                }

                let msgType = (params.messageType && typeof params.messageType.toUpperCase() !== 'undefined' && chatMessageTypes[params.messageType.toUpperCase()] > 0) ? chatMessageTypes[params.messageType.toUpperCase()] : 1;

                return chatUploadHandler({
                    threadId: params.threadId,
                    file: params.file,
                    fileUniqueId: fileUniqueId
                }, function (uploadHandlerResult, uploadHandlerMetadata, fileType, fileExtension) {
                    fileUploadParams = Object.assign(fileUploadParams, uploadHandlerResult);

                    putInChatUploadQueue({
                        message: {
                            chatMessageVOType: chatMessageVOTypes.MESSAGE,
                            typeCode: params.typeCode,
                            messageType: msgType,
                            subjectId: params.threadId,
                            repliedTo: params.repliedTo,
                            content: params.content,
                            metadata: JSON.stringify(Utility.objectDeepMerger(uploadHandlerMetadata, params.metadata)),
                            systemMetadata: JSON.stringify(params.systemMetadata),
                            uniqueId: fileUniqueId,
                            pushMsgType: 3
                        },
                        callbacks: callbacks
                    }, function () {
                        if ((imageMimeTypes.indexOf(fileType) >= 0 || imageExtentions.indexOf(fileExtension) >= 0)) {
                            app.fileMethods.uploadImageToPodspaceUserGroupNew(fileUploadParams, function (result) {
                                if (!result.hasError) {
                                    // Send onFileUpload callback result
                                    if (typeof callbacks === 'object' && callbacks.hasOwnProperty('onFileUpload')) {
                                        callbacks.onFileUpload && callbacks.onFileUpload({
                                            name: result.result.name,
                                            hashCode: result.result.hash,
                                            parentHash: result.result.parentHash,
                                            size: result.result.size,
                                            actualHeight: result.result.actualHeight,
                                            actualWidth: result.result.actualWidth,
                                            link: `${app.sdkParams.SERVICE_ADDRESSES.PODSPACE_FILESERVER_ADDRESS}/api/v2/images/${result.result.hash}?checkUserGroupAccess=true`
                                        });
                                    }
                                    metadata['name'] = result.result.name;
                                    metadata['fileHash'] = result.result.hash;
                                    metadata['file']['name'] = result.result.name;
                                    metadata['file']['fileHash'] = result.result.hash;
                                    metadata['file']['hashCode'] = result.result.hash;
                                    metadata['file']['parentHash'] = result.result.parentHash;
                                    metadata['file']['size'] = result.result.size;
                                    metadata['file']['actualHeight'] = result.result.actualHeight;
                                    metadata['file']['actualWidth'] = result.result.actualWidth;

                                    metadata['file']['link'] = `${app.sdkParams.SERVICE_ADDRESSES.PODSPACE_FILESERVER_ADDRESS}/api/v2/images/${result.result.hash}?checkUserGroupAccess=true`;//`https://podspace.pod.ir/nzh/drive/downloadImage?hash=${result.result.hashCode}`;
                                    transferFromUploadQToSendQ(parseInt(params.threadId), fileUniqueId, JSON.stringify(metadata), function () {
                                        chatSendQueueHandler();
                                    });
                                } else {
                                    app.chatEvents.fireEvent('error', {
                                        code: result.errorCode,
                                        message: result.errorMessage,
                                        result
                                    });
                                    app.messageQueues.deleteFromChatUploadQueue({message: {uniqueId: fileUniqueId}});
                                }
                            });
                        } else {
                            app.fileMethods.uploadFileToPodspaceUserGroupNew(fileUploadParams, function (result) {
                                if (!result.hasError) {
                                    metadata['fileHash'] = result.result.hash;
                                    metadata['name'] = result.result.name;
                                    metadata['file']['name'] = result.result.name;
                                    metadata['file']['fileHash'] = result.result.hash;
                                    metadata['file']['hashCode'] = result.result.hash;
                                    metadata['file']['parentHash'] = result.result.parentHash;
                                    metadata['file']['size'] = result.result.size;

                                    transferFromUploadQToSendQ(parseInt(params.threadId), fileUniqueId, JSON.stringify(metadata), function () {
                                        chatSendQueueHandler();
                                    });
                                } else {
                                    app.chatEvents.fireEvent('error', {
                                        code: result.errorCode,
                                        message: result.errorMessage,
                                        result
                                    });
                                    app.messageQueues.deleteFromChatUploadQueue({message: {uniqueId: fileUniqueId}});
                                }
                            });
                        }
                    });
                });
            }
        },

        /**
         * Get Chat Send Queue
         *
         * This function returns chat send queue
         *
         * @access private
         *
         * @return {array}  An array of messages on sendQueue
         */
        getChatSendQueue = function (threadId, callback) {
            if (threadId) {
                var tempSendQueue = [];

                for (var i = 0; i < chatSendQueue.length; i++) {
                    if (chatSendQueue[i].threadId == threadId) {
                        tempSendQueue.push(chatSendQueue[i]);
                    }
                }
                callback && callback(tempSendQueue);
            } else {
                callback && callback(chatSendQueue);
            }
        },

        /**
         * Delete an Item from Chat Send Queue
         *
         * This function gets an item and deletes it
         * from Chat Send Queue
         *
         * @access private
         *
         * @return {undefined}
         */
        deleteFromChatSentQueue = function (item, callback) {
            for (var i = 0; i < chatSendQueue.length; i++) {
                if (chatSendQueue[i].message.uniqueId == item.message.uniqueId) {
                    chatSendQueue.splice(i, 1);
                }
            }
            callback && callback();
        },

        /**
         * Push Message Into Send Queue
         *
         * This functions takes a message and puts it
         * into chat's send queue
         *
         * @access private
         *
         * @param {object}    params    The Message and its callbacks to be enqueued
         *
         * @return {undefined}
         */
        putInChatSendQueue = function (params, callback) {
            chatSendQueue.push(params);

            putInChatWaitQueue(params.message, function () {
                callback && callback();
            });
        },

        /**
         * Put an Item inside Chat Wait Queue
         *
         * This function takes an item and puts it
         * inside Chat Wait Queue, either on cached
         * wait queue or the wait queue on RAM memory
         *
         * @access private
         *
         * @return {undefined}
         */
        putInChatWaitQueue = function (item, callback) {
            if (item.uniqueId != '') {
                var waitQueueUniqueId = (typeof item.uniqueId == 'string') ? item.uniqueId : (Array.isArray(item.uniqueId)) ? item.uniqueId[0] : null;

                if (waitQueueUniqueId != null) {
                    item.uniqueId = waitQueueUniqueId;
                    app.store.chatWaitQueue.push(item);
                    callback && callback();
                }
            }
        },

        /**
         * Put an Item inside Chat Upload Queue
         *
         * This function takes an item and puts it
         * inside Chat upload Queue
         *
         * @access private
         *
         * @return {undefined}
         */
        putInChatUploadQueue = function (params, callback) {
            app.store.chatUploadQueue.push(params);
            callback && callback();
        },

        /**
         * Transfer an Item from uploadQueue to sendQueue
         *
         * This function takes an uniqueId, finds that item
         * inside uploadQ. takes it's uploaded metadata and
         * attaches them to the message. Finally removes item
         * from uploadQueue and pushes it inside sendQueue
         *
         * @access private
         *
         * @return {undefined}
         */
        transferFromUploadQToSendQ = function (threadId, uniqueId, metadata, callback) {
            app.messageQueues.getChatUploadQueue(threadId, function (uploadQueue) {
                for (var i = 0; i < uploadQueue.length; i++) {
                    if (uploadQueue[i].message.uniqueId == uniqueId) {
                        try {
                            var message = uploadQueue[i].message,
                                callbacks = uploadQueue[i].callbacks;

                            let oldMetadata = JSON.parse(message.metadata),
                                newMetadata = JSON.parse(metadata);

                            var finalMetaData = Utility.objectDeepMerger(newMetadata, oldMetadata);

                            if (message && typeof message.content === 'object' && typeof message.content.message !== 'undefined') {
                                message.content.message['metadata'] = JSON.stringify(finalMetaData);
                            }

                            if (message && typeof message.content === 'object' && typeof message.content.metadata !== 'undefined') {
                                message.content['metadata'] = JSON.stringify(finalMetaData);
                            }

                            if (message.chatMessageVOType == 21) {
                                app.fileMethods.getImageDownloadLinkFromPodspace({
                                    hashCode: finalMetaData.fileHash
                                }, function (result) {
                                    if (!result.hasError) {
                                        message.content.image = result.downloadUrl;
                                    }
                                });
                            }

                            message.metadata = JSON.stringify(finalMetaData);
                        } catch (e) {
                            console.log(e);
                        }

                        app.messageQueues.deleteFromChatUploadQueue(uploadQueue[i],
                            function () {
                                putInChatSendQueue({
                                    message: message,
                                    callbacks: callbacks
                                }, function () {
                                    callback && callback();
                                });
                            });

                        break;
                    }
                }
            });
        },

        /**
         * Decrypt Encrypted strings using secret key and salt
         *
         * @param string    String to get decrypted
         * @param secret    Cache Secret
         * @param salt      Salt used while string was getting encrypted
         *
         * @return  string  Decrypted string
         */
        chatDecrypt = function (string, secret, salt) {
            var decryptedString = Utility.decrypt(string, secret, salt);
            if (!decryptedString.hasError) {
                return decryptedString.result;
            } else {
                /**
                 * If there is a problem with decrypting cache
                 * Some body is trying to decrypt cache with wrong key
                 * or cacheSecret has been expired, so we should truncate
                 * cache databases to avoid attacks.
                 *
                 * But before deleting cache database we should make
                 * sure that cacheSecret has been retrieved from server
                 * and is ready. If so, and cache is still not decryptable,
                 * there is definitely something wrong with the key; so we are
                 * good to go and delete cache databases.
                 */
                if (secret != 'undefined' && secret != '') {
                    if (db) {
                        db.threads
                            .where('owner')
                            .equals(parseInt(app.store.user.get().id))
                            .count()
                            .then(function (threadsCount) {
                                if (threadsCount > 0) {
                                    clearCacheDatabasesOfUser(function () {
                                        console.log('All cache databases have been cleared.');
                                    });
                                }
                            })
                            .catch(function (e) {
                                console.log(e);
                            });
                    }
                }

                return '{}';
            }
        },

        setRoleToUser = function (params, callback) {
            var setRoleData = {
                chatMessageVOType: chatMessageVOTypes.SET_ROLE_TO_USER,
                typeCode: params.typeCode,
                content: [],
                pushMsgType: 3,
                token: app.sdkParams.token
            };

            if (params) {
                if (parseInt(params.threadId) > 0) {
                    setRoleData.subjectId = params.threadId;
                }

                if (params.admins && Array.isArray(params.admins)) {
                    for (var i = 0; i < params.admins.length; i++) {
                        var temp = {};
                        if (parseInt(params.admins[i].userId) > 0) {
                            temp.userId = params.admins[i].userId;
                        }

                        if (Array.isArray(params.admins[i].roles)) {
                            temp.roles = params.admins[i].roles;
                        }

                        setRoleData.content.push(temp);
                    }

                    setRoleData.content = JSON.stringify(setRoleData.content);
                }
            }

            return app.messenger.sendMessage(setRoleData, {
                onResult: function (result) {
                    callback && callback(result);
                }
            });
        },

        removeRoleFromUser = function (params, callback) {
            var setAdminData = {
                chatMessageVOType: chatMessageVOTypes.REMOVE_ROLE_FROM_USER,
                typeCode: params.typeCode,
                content: [],
                pushMsgType: 3,
                token: app.sdkParams.token
            };

            if (params) {
                if (parseInt(params.threadId) > 0) {
                    setAdminData.subjectId = params.threadId;
                }

                if (params.admins && Array.isArray(params.admins)) {
                    for (var i = 0; i < params.admins.length; i++) {
                        var temp = {};
                        if (parseInt(params.admins[i].userId) > 0) {
                            temp.userId = params.admins[i].userId;
                        }

                        if (Array.isArray(params.admins[i].roles)) {
                            temp.roles = params.admins[i].roles;
                        }

                        setAdminData.content.push(temp);
                    }

                    setAdminData.content = JSON.stringify(setAdminData.content);
                }
            }

            return app.messenger.sendMessage(setAdminData, {
                onResult: function (result) {
                    callback && callback(result);
                }
            });
        },

        unPinMessage = function (params, callback) {
            return app.messenger.sendMessage({
                chatMessageVOType: chatMessageVOTypes.UNPIN_MESSAGE,
                typeCode: params.typeCode,
                subjectId: params.messageId,
                content: JSON.stringify({
                    'notifyAll': (typeof params.notifyAll === 'boolean') ? params.notifyAll : false
                }),
                pushMsgType: 3,
                token: app.sdkParams.token
            }, {
                onResult: function (result) {
                    callback && callback(result);
                }
            });
        },

        chatUploadHandler = function (params, callbacks) {
            if (typeof params.file != 'undefined') {
                var fileName,
                    fileType,
                    fileSize,
                    fileExtension,
                    chatUploadHandlerResult = {},
                    metadata = {file: {}},
                    fileUniqueId = params.fileUniqueId;

                if (Utility.isNode()) {
                    fileName = params.file.split('/').pop();
                    fileType = Mime.getType(params.file);
                    fileSize = app.fs.statSync(params.file).size;
                    fileExtension = params.file.split('.')
                        .pop();
                } else {
                    fileName = params.file.name;
                    fileType = params.file.type;
                    fileSize = params.file.size;
                    fileExtension = params.file.name.split('.')
                        .pop();
                }

                app.chatEvents.fireEvent('fileUploadEvents', {
                    threadId: params.threadId,
                    uniqueId: fileUniqueId,
                    state: 'NOT_STARTED',
                    progress: 0,
                    fileInfo: {
                        fileName: fileName,
                        fileSize: fileSize
                    },
                    fileObject: params.file
                });

                /**
                 * File is a valid Image
                 * Should upload to image server
                 */
                if (imageMimeTypes.indexOf(fileType) >= 0 || imageExtentions.indexOf(fileExtension) >= 0) {
                    chatUploadHandlerResult.image = params.file;

                    if (params.xC >= 0) {
                        fileUploadParams.xC = params.xC;
                    }

                    if (params.yC >= 0) {
                        fileUploadParams.yC = params.yC;
                    }

                    if (params.hC > 0) {
                        fileUploadParams.hC = params.hC;
                    }

                    if (params.wC > 0) {
                        fileUploadParams.wC = params.wC;
                    }
                } else {
                    chatUploadHandlerResult.file = params.file;
                }

                metadata['file']['originalName'] = fileName;
                metadata['file']['mimeType'] = fileType;
                metadata['file']['size'] = fileSize;

                chatUploadHandlerResult.threadId = params.threadId;
                chatUploadHandlerResult.uniqueId = fileUniqueId;
                chatUploadHandlerResult.fileObject = params.file;
                chatUploadHandlerResult.originalFileName = fileName;

                callbacks && callbacks(chatUploadHandlerResult, metadata, fileType, fileExtension);
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 6302,
                    message: CHAT_ERRORS[6302]
                });
            }

            return {
                uniqueId: fileUniqueId,
                threadId: params.threadId,
                participant: app.store.user.get(),
                content: {
                    caption: params.content,
                    file: {
                        uniqueId: fileUniqueId,
                        fileName: fileName,
                        fileSize: fileSize,
                        fileObject: params.file
                    }
                }
            };
        },

        //TODO Change Node Version
        getImageFormUrl = function (url, callback) {
            var img = new Image();
            img.setAttribute('crossOrigin', 'anonymous');
            img.onload = function (a) {
                var canvas = document.createElement("canvas");
                canvas.width = this.width;
                canvas.height = this.height;
                var ctx = canvas.getContext("2d");
                ctx.drawImage(this, 0, 0);

                var dataURI = canvas.toDataURL("image/jpg");

                // convert base64/URLEncoded data component to raw binary data held in a string
                var byteString;
                if (dataURI.split(',')[0].indexOf('base64') >= 0)
                    byteString = atob(dataURI.split(',')[1]);
                else
                    byteString = unescape(dataURI.split(',')[1]);

                // separate out the mime component
                var mimeString = dataURI.split(',')[0].split(':')[1].split(';')[0];

                // write the bytes of the string to a typed array
                var ia = new Uint8Array(byteString.length);
                for (var i = 0; i < byteString.length; i++) {
                    ia[i] = byteString.charCodeAt(i);
                }

                return callback(new Blob([ia], {type: mimeString}));
            }

            img.src = url;
        },

        callReceived = function (params, callback) {
            var receiveCallData = {
                chatMessageVOType: chatMessageVOTypes.RECEIVE_CALL_REQUEST,
                typeCode: params.typeCode,
                pushMsgType: 3,
                token: app.sdkParams.token
            };

            if (params) {
                if (typeof +params.callId === 'number' && params.callId > 0) {
                    receiveCallData.subjectId = +params.callId;
                } else {
                    app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                    // app.chatEvents.fireEvent('error', {
                    //     code: 999,
                    //     message: 'Invalid call id!'
                    // });
                    return;
                }
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] No params have been sent to ReceiveCall()'
                });
                return;
            }

            return app.messenger.sendMessage(receiveCallData, {
                onResult: function (result) {
                    callback && callback(result);
                }
            });
        },

        endCall = function (params, callback) {
            app.sdkParams.consoleLogging && console.log('[SDK][endCall] called...');

            var endCallData = {
                chatMessageVOType: chatMessageVOTypes.END_CALL_REQUEST,
                typeCode: params.typeCode,
                pushMsgType: 3,
                token: app.sdkParams.token
            };

            if (!callRequestController.callEstablishedInMySide) {
                return;
            }

            if (params) {
                if (typeof +params.callId === 'number' && params.callId > 0) {
                    endCallData.subjectId = +params.callId;
                } else {
                    app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                    // app.chatEvents.fireEvent('error', {
                    //     code: 999,
                    //     message: 'Invalid call id!'
                    // });
                    return;
                }
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] No params have been sent to End the call!'
                });
                return;
            }

            callStop();

            return app.messenger.sendMessage(endCallData, {
                onResult: function (result) {
                    callback && callback(result);
                }
            });
        },

        callStop = function () {
            if (callStopQueue.callStarted) {
                callStopQueue.callStarted = false;
            }

            callRequestController.callEstablishedInMySide = false;
            callRequestController.callRequestReceived = false;
            currentCallParams = {};
            currentCallId = null;
        },

        /**
         * Reformat Call Participants
         *
         * This functions reformats given Array of call Participants
         * into proper call participant
         *
         * @access private
         *
         * @param {object}  participantsContent   Array of Call Participant Objects
         * @param {int}    threadId              Id of call
         *
         * @return {object} Formatted Call Participant Array
         */
        reformatCallParticipants = function (participantsContent) {
            var returnData = [];

            for (var i = 0; i < participantsContent.length; i++) {
                returnData.push(formatDataToMakeCallParticipant(participantsContent[i]));
            }

            return returnData;
        },

        formatDataToMakeCallParticipant = function (messageContent) {
            /**
             * + CallParticipantVO                   {object}
             *    - id                           {int}
             *    - joinTime                     {int}
             *    - leaveTime                    {int}
             *    - threadParticipant            {object}
             *    - sendTopic                    {string}
             *    - receiveTopic                 {string}
             *    - brokerAddress                {string}
             *    - active                       {boolean}
             *    - callSession                  {object}
             *    - callStatus                   {int}
             *    - createTime                   {int}
             *    - sendKey                      {string}
             *    - mute                         {boolean}
             */

            var participant = {
                id: messageContent.id,
                joinTime: messageContent.joinTime,
                leaveTime: messageContent.leaveTime,
                sendTopic: messageContent.sendTopic,
                receiveTopic: messageContent.receiveTopic,
                brokerAddress: messageContent.brokerAddress,
                active: messageContent.active,
                callSession: messageContent.callSession,
                callStatus: messageContent.callStatus,
                createTime: messageContent.createTime,
                sendKey: messageContent.sendKey,
                mute: messageContent.mute
            };

            // Add Chat Participant if exist
            if (messageContent.participantVO) {
                participant.participantVO = messageContent.participantVO;
            }

            // Add Call Session if exist
            if (messageContent.callSession) {
                participant.callSession = messageContent.callSession;
            }

            // return participant;
            return JSON.parse(JSON.stringify(participant));
        };

    /******************************************************
     *             P U B L I C   M E T H O D S            *
     ******************************************************/

    this.on = app.chatEvents.on;
    this.off = app.chatEvents.off;

    this.getPeerId = function () {
        return peerId;
    };

    this.getCurrentUser = function () {
        return app.store.user.get();
    };

    this.getUserInfo = getUserInfo;

    this.getThreads = getThreads;

    this.getAllThreads = getAllThreads;

    this.getThreadsLight = function (params, callback) {
        var count = 25,
            offset = 0,
            content = {};

        if (params) {
            if (parseInt(params.count) > 0) {
                count = params.count;
            }

            if (parseInt(params.offset) > 0) {
                offset = params.offset;
            }

            if (typeof params.threadName === 'string') {
                content.name = params.threadName;
            }
            if (typeof params.username === 'string') {
                content.username = params.username;
            }

            if (typeof params.cellphoneNumber === 'string') {
                content.cellphoneNumber = params.cellphoneNumber;
            }

            if (Array.isArray(params.threadIds)) {
                content.threadIds = params.threadIds;
            }

            if (typeof params.new === 'boolean') {
                content.new = params.new;
            }

            if (parseInt(params.creatorCoreUserId) > 0) {
                content.creatorCoreUserId = params.creatorCoreUserId;
            }

            if (parseInt(params.partnerCoreUserId) > 0) {
                content.partnerCoreUserId = params.partnerCoreUserId;
            }

            if (parseInt(params.partnerCoreContactId) > 0) {
                content.partnerCoreContactId = params.partnerCoreContactId;
            }

            if (parseInt(params.fromTime) > 0 && parseInt(params.fromTime) < 9999999999999) {
                content.fromTime = parseInt(params.fromTime);
            }

            if (parseInt(params.toTime) > 0 && parseInt(params.toTime) < 9999999999999) {
                content.toTime = parseInt(params.toTime);
            }

            if (typeof params.isGroup === 'boolean') {
                content.isGroup = params.isGroup;
            }
            if (typeof params.type === 'number') {
                content.type = params.type;
            }
        }

        content.count = count;
        content.offset = offset;

        let sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.GET_THREAD_LIGHT,
            typeCode: app.sdkParams.generalTypeCode, //params.typeCode,
            content: content
        };

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode,
                    uniqueId: result.uniqueId
                };

                if (!returnData.hasError) {
                    let messageContent = result.result,
                        messageLength = messageContent.length,
                        resultData = {
                            threads: [],
                            contentCount: result.contentCount,
                            hasNext: messageContent && !(messageLength < count),//(offset + count < result.contentCount && messageLength > 0),
                            nextOffset: offset * 1 + messageLength * 1
                        },
                        threadData;

                    for (let i = 0; i < messageLength; i++) {
                        threadData = createThread(messageContent[i], false);
                        if (threadData) {
                            resultData.threads.push(threadData);
                        }
                    }

                    // app.store.threads.saveMany(resultData.threads);
                    returnData.result = resultData;
                }

                callback && callback(returnData);
            }
        }, null);
    }

    this.getHistory = getHistory;

    this.getAllMentionedMessages = function (params, callback) {
        return getHistory({
            threadId: params.threadId,
            allMentioned: true,
            typeCode: params.typeCode,
            count: params.count || 50,
            offset: params.offset || 0,
            cache: false,
            queues: {
                sending: false,
                uploading: false
            }
        }, callback);
    };

    this.getUnreadMentionedMessages = function (params, callback) {
        return getHistory({
            threadId: params.threadId,
            unreadMentioned: true,
            typeCode: params.typeCode,
            count: params.count || 50,
            offset: params.offset || 0,
            cache: false,
            queues: {
                sending: false,
                uploading: false
            }
        }, callback);
    };

    this.getAllUnreadMessagesCount = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.ALL_UNREAD_MESSAGE_COUNT,
            typeCode: params.typeCode,
            content: JSON.stringify({
                'mute': (typeof params.countMuteThreads === 'boolean') ? params.countMuteThreads : false
            }),
            pushMsgType: 3,
            token: app.sdkParams.token
        }, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    /**
     * Get Contacts
     *
     * Gets contacts list from chat server
     *
     * @access pubic
     *
     * @param {int}     count           Count of objects to get
     * @param {int}     offset          Offset of select Query
     * @param {string}  query           Search in contacts list to get (search LIKE firstName, lastName, email)
     *
     * @return {object} Instant Response
     */
    this.getContacts = function (params, callback) {
        var count = 50,
            offset = 0,
            content = {},
            whereClause = {},
            returnCache = false;

        if (params) {
            if (parseInt(params.count) > 0) {
                count = parseInt(params.count);
            }

            if (parseInt(params.offset) > 0) {
                offset = parseInt(params.offset);
            }

            if (typeof params.query === 'string') {
                content.query = whereClause.query = params.query;
            }

            if (typeof params.email === 'string') {
                content.email = whereClause.email = params.email;
            }

            if (typeof params.cellphoneNumber === 'string') {
                content.cellphoneNumber = whereClause.cellphoneNumber = params.cellphoneNumber;
            }

            if (typeof params.contactId === 'string') {
                content.id = whereClause.id = params.contactId;
            }

            if (typeof params.uniqueId === 'string') {
                content.uniqueId = whereClause.uniqueId = params.uniqueId;
            }
            if (typeof params.username === 'string') {
                content.username = params.username;
            }
            if (typeof params.coreUserId !== "undefined") {
                content.coreUserId = params.coreUserId;
            }

            var functionLevelCache = (typeof params.cache == 'boolean') ? params.cache : true;
        }

        content.size = count;
        content.offset = offset;

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.GET_CONTACTS,
            typeCode: params.typeCode,
            content: content
        };

        /**
         * Retrieve Contacts from server
         */
        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {

                    var messageContent = result.result,
                        messageLength = messageContent.length,
                        resultData = {
                            contacts: [],
                            contentCount: result.contentCount,
                            hasNext: (offset + count < result.contentCount && messageLength > 0),
                            nextOffset: offset + messageLength
                        },
                        contactData;

                    for (var i = 0; i < messageLength; i++) {
                        contactData = formatDataToMakeContact(messageContent[i]);
                        if (contactData) {
                            resultData.contacts.push(contactData);
                        }
                    }

                    returnData.result = resultData;
                }

                callback && callback(returnData);
                /**
                 * Delete callback so if server pushes response before
                 * cache, cache won't send data again
                 */
                callback = undefined;
            }
        });
    };

    this.getThreadParticipants = getThreadParticipants;

    /**
     * Get Thread Admins
     *
     * Gets admins list of given thread
     *
     * @access pubic
     *
     * @param {int}     threadId        Id of thread which you want to get admins of
     *
     * @return {object} Instant Response
     */
    this.getThreadAdmins = function (params, callback) {
        getThreadParticipants({
            threadId: params.threadId,
            admin: true,
            cache: false
        }, callback);
    };

    this.addParticipants = function (params, callback) {

        /**
         * + AddParticipantsRequest   {object}
         *    - subjectId             {long}
         *    + content               {list} List of CONTACT IDs or inviteeVO Objects
         *    - uniqueId              {string}
         */

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.ADD_PARTICIPANT,
            typeCode: params.typeCode,
            content: []
        };

        if (params) {
            if (parseInt(params.threadId) > 0) {
                sendMessageParams.subjectId = params.threadId;
            }

            if (Array.isArray(params.contactIds)) {
                sendMessageParams.content = params.contactIds;
            }

            if (Array.isArray(params.usernames)) {
                sendMessageParams.content = [];

                for (var i = 0; i < params.usernames.length; i++) {
                    sendMessageParams.content.push({
                        id: params.usernames[i],
                        idType: inviteeVOidTypes.TO_BE_USER_USERNAME
                    });
                }
            }

            if (Array.isArray(params.coreUserids)) {
                sendMessageParams.content = [];

                for (var i = 0; i < params.coreUserids.length; i++) {
                    sendMessageParams.content.push({
                        id: params.coreUserids[i],
                        idType: inviteeVOidTypes.TO_BE_USER_ID
                    });
                }
            }
        }

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            thread: createThread(messageContent)
                        };

                    returnData.result = resultData;
                }

                callback && callback(returnData);
            }
        });
    };

    this.removeParticipants = function (params, callback) {

        /**
         * + RemoveParticipantsRequest    {object}
         *    - subjectId                 {long}
         *    + content                   {list} List of PARTICIPANT IDs from Thread's Participants object
         *       -id                      {long}
         *    - uniqueId                  {string}
         */

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.REMOVE_PARTICIPANT,
            typeCode: params.typeCodes
        };

        if (params) {
            if (parseInt(params.threadId) > 0) {
                sendMessageParams.subjectId = params.threadId;
            }

            if (Array.isArray(params.participantIds)) {
                sendMessageParams.content = params.participantIds;
            }
        }

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            thread: createThread(messageContent)
                        };

                    returnData.result = resultData;
                }

                callback && callback(returnData);
            }
        });
    };

    this.getCurrentUserRoles = getCurrentUserRoles;

    this.leaveThread = function (params, callback) {

        /**
         * + LeaveThreadRequest    {object}
         *    - subjectId          {long}
         *    - uniqueId           {string}
         */

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.LEAVE_THREAD,
            typeCode: params.typeCode
        };

        if (params) {
            if (parseInt(params.threadId) > 0) {
                sendMessageParams.subjectId = params.threadId;
            }
        }

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            thread: createThread(messageContent)
                        };

                    returnData.result = resultData;
                }

                callback && callback(returnData);
            }
        });
    };

    this.createThread = function (params, callback) {

        /**
         * + CreateThreadRequest      {object}
         *    + invitees              {object}
         *       -id                  {string}
         *       -idType              {int} ** inviteeVOidTypes
         *    - title                 {string}
         *    - type                  {int} ** createThreadTypes
         *    - image                 {string}
         *    - description           {string}
         *    - metadata              {string}
         *    - uniqueName            {string}
         *    + message               {object}
         *       -text                {string}
         *       -type                {int}
         *       -repliedTo           {long}
         *       -uniqueId            {string}
         *       -metadata            {string}
         *       -systemMetadata      {string}
         *       -forwardedMessageIds {string}
         *       -forwardedUniqueIds  {string}
         */

        var content = {};

        if (params) {
            if (typeof params.title === 'string') {
                content.title = params.title;
            }

            if (typeof params.type === 'string') {
                var threadType = params.type;
                content.type = createThreadTypes[threadType];
            }

            if (typeof params.uniqueName === 'string') {
                content.uniqueName = params.uniqueName;
            }

            if (Array.isArray(params.invitees)) {
                var tempInvitee;
                content.invitees = [];
                for (var i = 0; i < params.invitees.length; i++) {
                    var tempInvitee = formatDataToMakeInvitee(params.invitees[i]);
                    if (tempInvitee) {
                        content.invitees.push(tempInvitee);
                    }
                }
            }

            if (typeof params.image === 'string') {
                content.image = params.image;
            }

            if (typeof params.description === 'string') {
                content.description = params.description;
            }

            if (typeof params.metadata === 'string') {
                content.metadata = params.metadata;
            } else if (typeof params.metadata === 'object') {
                try {
                    content.metadata = JSON.stringify(params.metadata);
                } catch (e) {
                    console.log(e);
                }
            }

            if (typeof params.message == 'object') {
                content.message = {};

                if (typeof params.message.text === 'string') {
                    content.message.text = params.message.text;
                }

                if (typeof params.message.uniqueId === 'string') {
                    content.message.uniqueId = params.message.uniqueId;
                }

                if (params.message.type > 0) {
                    content.message.messageType = params.message.type;
                }

                if (params.message.repliedTo > 0) {
                    content.message.repliedTo = params.message.repliedTo;
                }

                if (typeof params.message.metadata === 'string') {
                    content.message.metadata = params.message.metadata;
                } else if (typeof params.message.metadata === 'object') {
                    content.message.metadata = JSON.stringify(params.message.metadata);
                }

                if (typeof params.message.systemMetadata === 'string') {
                    content.message.systemMetadata = params.message.systemMetadata;
                } else if (typeof params.message.systemMetadata === 'object') {
                    content.message.systemMetadata = JSON.stringify(params.message.systemMetadata);
                }

                if (Array.isArray(params.message.forwardedMessageIds)) {
                    content.message.forwardedMessageIds = params.message.forwardedMessageIds;
                    content.message.forwardedUniqueIds = [];
                    for (var i = 0; i < params.message.forwardedMessageIds.length; i++) {
                        content.message.forwardedUniqueIds.push(Utility.generateUUID());
                    }
                }

            }
        }

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.CREATE_THREAD,
            typeCode: params.typeCode,
            content: content
        };

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            thread: createThread(messageContent)
                        };

                    returnData.result = resultData;
                }

                callback && callback(returnData);
            }
        });

    };

    this.sendTextMessage = function (params, callbacks) {
        var metadata = {},
            uniqueId;

        if (typeof params.uniqueId != 'undefined') {
            uniqueId = params.uniqueId;
        } else {
            uniqueId = Utility.generateUUID();
        }

        putInChatSendQueue({
            message: {
                chatMessageVOType: chatMessageVOTypes.MESSAGE,
                typeCode: params.typeCode,
                messageType: (params.messageType && params.messageType.toUpperCase() !== undefined && chatMessageTypes[params.messageType.toUpperCase()] > 0) ? chatMessageTypes[params.messageType.toUpperCase()] : chatMessageTypes.TEXT,
                subjectId: params.threadId,
                repliedTo: params.repliedTo,
                content: params.textMessage,
                uniqueId: uniqueId,
                systemMetadata: JSON.stringify(params.systemMetadata),
                metadata: JSON.stringify(metadata),
                pushMsgType: 3
            },
            callbacks: callbacks
        }, function () {
            chatSendQueueHandler();
        });

        return {
            uniqueId: uniqueId,
            threadId: params.threadId,
            participant: app.store.user.get(),
            content: params.content
        };
    };

    this.sendBotMessage = function (params, callbacks) {
        var metadata = {};

        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.BOT_MESSAGE,
            typeCode: params.typeCode,
            subjectId: params.messageId,
            content: params.content,
            uniqueId: params.uniqueId,
            pushMsgType: 3
        }, callbacks);
    };

    this.sendFileMessage = sendFileMessage;

    this.createThreadWithFileMessage = function (params, createThreadCallback, sendFileMessageCallback) {
        /**
         * + CreateThreadRequest      {object}
         *    + invitees              {object}
         *       -id                  {string}
         *       -idType              {int} ** inviteeVOidTypes
         *    - title                 {string}
         *    - type                  {int} ** createThreadTypes
         *    - image                 {string}
         *    - description           {string}
         *    - metadata              {string}
         *    - uniqueName            {string}
         *    + message               {object}
         *       -text                {string}
         *       -type                {int}
         *       -repliedTo           {long}
         *       -uniqueId            {string}
         *       -metadata            {string}
         *       -systemMetadata      {string}
         *       -forwardedMessageIds {string}
         *       -forwardedUniqueIds  {string}
         */
        var content = {};

        if (params) {
            if (typeof params.title === 'string') {
                content.title = params.title;
            }

            if (typeof params.type === 'string') {
                var threadType = params.type;
                content.type = createThreadTypes[threadType];
            }

            if (Array.isArray(params.invitees)) {
                var tempInvitee;
                content.invitees = [];
                for (var i = 0; i < params.invitees.length; i++) {
                    var tempInvitee = formatDataToMakeInvitee(params.invitees[i]);
                    if (tempInvitee) {
                        content.invitees.push(tempInvitee);
                    }
                }
            }

            if (typeof params.description === 'string') {
                content.description = params.description;
            }

            if (typeof params.content === 'string') {
                content.content = params.content;
            }

            if (typeof params.metadata === 'string') {
                content.metadata = params.metadata;
            } else if (typeof params.metadata === 'object') {
                try {
                    content.metadata = JSON.stringify(params.metadata);
                } catch (e) {
                    console.log(e);
                }
            }
        }

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.CREATE_THREAD,
            typeCode: params.typeCode,
            content: content
        };

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            thread: createThread(messageContent)
                        };

                    returnData.result = resultData;
                }

                createThreadCallback && createThreadCallback(returnData);

                sendFileMessage({
                    threadId: returnData.result.thread.id,
                    file: params.file,
                    content: params.caption,
                    messageType: params.messageType,
                    userGroupHash: returnData.result.thread.userGroupHash
                }, sendFileMessageCallback);
            }
        });

    };

    this.sendLocationMessage = function (params, callbacks) {
        var data = {},
            url = app.sdkParams.SERVICE_ADDRESSES.MAP_ADDRESS + SERVICES_PATH.STATIC_IMAGE,
            hasError = false,
            fileUniqueId = Utility.generateUUID();

        if (params) {
            if (typeof params.mapType === 'string') {
                data.type = params.mapType;
            } else {
                data.type = 'standard-night';
            }

            if (parseInt(params.mapZoom) > 0) {
                data.zoom = params.mapZoom;
            } else {
                data.zoom = 15;
            }

            if (parseInt(params.mapWidth) > 0) {
                data.width = params.mapWidth;
            } else {
                data.width = 800;
            }

            if (parseInt(params.mapHeight) > 0) {
                data.height = params.mapHeight;
            } else {
                data.height = 600;
            }

            if (typeof params.mapCenter === 'object') {
                if (parseFloat(params.mapCenter.lat) > 0 && parseFloat(params.mapCenter.lng)) {
                    data.center = params.mapCenter.lat + ',' + parseFloat(params.mapCenter.lng);
                } else {
                    hasError = true;
                    app.chatEvents.fireEvent('error', {
                        code: 6700,
                        message: CHAT_ERRORS[6700],
                        error: undefined
                    });
                }
            } else {
                hasError = true;
                app.chatEvents.fireEvent('error', {
                    code: 6700,
                    message: CHAT_ERRORS[6700],
                    error: undefined
                });
            }

            data.key = app.sdkParams.mapApiKey;
            data.marker = 'red';
        }

        var keys = Object.keys(data);

        if (keys.length > 0) {
            url += '?';

            for (var i = 0; i < keys.length; i++) {
                var key = keys[i];
                url += key + '=' + data[key];
                if (i < keys.length - 1) {
                    url += '&';
                }
            }
        }

        if (!hasError) {
            getImageFormUrl(url, function (blobImage) {
                sendFileMessage({
                    threadId: params.threadId,
                    fileUniqueId: fileUniqueId,
                    file: new File([blobImage], "location.png", {type: "image/png", lastModified: new Date()}),
                    content: params.caption,
                    messageType: 'POD_SPACE_PICTURE',
                    userGroupHash: params.userGroupHash,
                    metadata: {
                        mapLink: `https://maps.neshan.org/@${data.center},${data.zoom}z`
                    }
                });
            });
        }

        return {
            uniqueId: fileUniqueId,
            threadId: params.threadId,
            participant: app.store.user.get(),
            content: params.caption
        };
    };

    this.resendMessage = function (uniqueId, callbacks) {
        // if (hasCache && typeof queueDb == 'object') {
        //     queueDb.waitQ.where('uniqueId')
        //         .equals(uniqueId)
        //         .and(function (item) {
        //             return item.owner == parseInt(app.store.user.get().id);
        //         })
        //         .toArray()
        //         .then(function (messages) {
        //             if (messages.length) {
        //                 putInChatSendQueue({
        //                     message: Utility.jsonParser(chatDecrypt(messages[0].message, cacheSecret)),
        //                     callbacks: callbacks
        //                 }, function () {
        //                     chatSendQueueHandler();
        //                 });
        //             }
        //         })
        //         .catch(function (error) {
        //             fireEvent('error', {
        //                 code: error.code,
        //                 message: error.message,
        //                 error: error
        //             });
        //         });
        // } else {
        for (var i = 0; i < app.store.chatWaitQueue.length; i++) {
            if (app.store.chatWaitQueue[i].message.uniqueId == uniqueId) {
                putInChatSendQueue({
                    message: app.store.chatWaitQueue[i].message,
                    callbacks: callbacks
                }, function () {
                    chatSendQueueHandler();
                });
                break;
            }
        }
        // }
    };

    this.cancelMessage = function (uniqueId, callback) {
        deleteFromChatSentQueue({
            message: {
                uniqueId: uniqueId
            }
        }, function () {
            app.messageQueues.deleteFromChatWaitQueue({
                uniqueId: uniqueId
            }, callback);
        });
    };

    this.clearHistory = function (params, callback) {

        /**
         * + Clear History Request Object    {object}
         *    - subjectId                    {long}
         */

        var clearHistoryParams = {
            chatMessageVOType: chatMessageVOTypes.CLEAR_HISTORY,
            typeCode: params.typeCode
        };

        if (params) {
            if (parseInt(params.threadId) > 0) {
                clearHistoryParams.subjectId = params.threadId;
            }
        }

        return app.messenger.sendMessage(clearHistoryParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var resultData = {
                        thread: result.result
                    };

                    returnData.result = resultData;
                }

                callback && callback(returnData);
            }
        });
    };

    function requestExportChat(stackArr, wantedCount, stepCount, offset, sendData) {
        sendData.content.offset = offset;
        sendData.content.count = stepCount;
        return new Promise(function (resolve, reject) {
            return app.messenger.sendMessage(sendData, {
                onResult: function (result) {

                    var returnData = {
                        hasError: result.hasError,
                        cache: false,
                        errorMessage: result.errorMessage,
                        errorCode: result.errorCode
                    };

                    if (!returnData.hasError) {
                        stackArr.push(...result.result);

                        app.sdkParams.consoleLogging && console.log("[SDK][exportChat] a step passed...");
                        // wantedCount = wantedCount > result.contentCount ? result.contentCount : wantedCount;
                        if (result.result.length < stepCount) {
                            wantedCount = stackArr.length
                        }

                        setTimeout(function () {
                            app.chatEvents.fireEvent('threadEvents', {
                                type: 'EXPORT_CHAT',
                                subType: 'IN_PROGRESS',
                                threadId: sendData.subjectId,
                                percent: Math.floor((stackArr.length / wantedCount) * 100)
                            });

                            if (stackArr.length < wantedCount) {
                                stepCount = wantedCount - stackArr.length < stepCount ? wantedCount - stackArr.length : stepCount;
                                //setTimeout(function () {
                                resolve(requestExportChat(stackArr, wantedCount, stepCount, stackArr.length, sendData));
                                //}, 1000)
                            } else {
                                resolve(stackArr);
                            }
                        });
                    } else {
                        if (result.errorCode !== 21) {
                            app.sdkParams.consoleLogging && console.log("[SDK][exportChat] Problem in one step... . Rerunning the request.", wantedCount, stepCount, stackArr.length, sendData, result);
                            setTimeout(function () {
                                resolve(requestExportChat(stackArr, wantedCount, stepCount, stackArr.length, sendData))
                            }, 2000)
                        } else {
                            reject(result)
                        }
                    }
                }
            });
        })
    }

    this.exportChat = function (params, callback) {
        var stackArr = [], wantedCount = 10000, stepCount = 500, offset = 0;
        var sendData = {
            chatMessageVOType: chatMessageVOTypes.EXPORT_CHAT,
            typeCode: app.sdkParams.generalTypeCode,//params.typeCode,
            content: {
                offset: +params.offset > 0 ? +params.offset : offset,
                count: +params.count > 0 ? +params.count : wantedCount,//config.getHistoryCount,
            },
            subjectId: params.threadId
        };

        if (+params.fromTime > 0 && +params.fromTime < 9999999999999) {
            sendData.content.fromTime = +params.fromTime;
        }

        if (+params.toTime > 0 && +params.toTime < 9999999999999) {
            sendData.content.toTime = +params.toTime;
        }

        if (+params.wantedCount > 0) {
            wantedCount = params.wantedCount;
        }

        if (+params.stepCount > 0) {
            stepCount = params.stepCount;
        }

        if (+params.offset > 0) {
            offset = params.offset;
        }

        // if (params.messageType && typeof params.messageType.toUpperCase() !== 'undefined' && chatMessageTypes[params.messageType.toUpperCase()] > 0) {
        //     sendData.content.messageType = chatMessageTypes[params.messageType.toUpperCase()];
        // }
        sendData.content.messageType = 1;

        if (wantedCount < stepCount)
            stepCount = wantedCount;

        app.sdkParams.consoleLogging && console.log("[SDK][exportChat] Starting...");
        requestExportChat(stackArr, wantedCount, stepCount, offset, sendData).then(function (resultArray) {
            app.sdkParams.consoleLogging && console.log("[SDK][exportChat] Export done..., Now converting...");

            if (!params.exportPath) {
                callback && callback({
                    hasError: false,
                    result: resultArray
                });
                app.chatEvents.fireEvent('threadEvents', {
                    type: 'EXPORT_CHAT',
                    subType: 'DONE',
                    threadId: sendData.subjectId,
                    result: resultArray
                });

                return;
            }

            var str = ''
                , universalBOM = "\uFEFF";

            str += "\u{62A}\u{627}\u{631}\u{6CC}\u{62E} " + ','; //tarikh
            str += " \u{633}\u{627}\u{639}\u{62A} " + ','; //saat
            str += "\u{646}\u{627}\u{645} \u{641}\u{631}\u{633}\u{62A}\u{646}\u{62F}\u{647}" + ',';//name ferestande
            str += "\u{646}\u{627}\u{645} \u{6A9}\u{627}\u{631}\u{628}\u{631}\u{6CC} \u{641}\u{631}\u{633}\u{62A}\u{646}\u{62F}\u{647}" + ','; //name karbariye ferestande
            str += "\u{645}\u{62A}\u{646} \u{67E}\u{6CC}\u{627}\u{645}" + ',';//matne payam
            str += '\r\n';
            var line = '', radif = 1;
            for (var i = 0; i < resultArray.length; i++) {
                line = '';

                if (resultArray[i].messageType !== 1) {
                    continue;
                }

                var sender = '';
                if (resultArray[i].participant.contactName) {
                    sender = resultArray[i].participant.contactName + ',';
                } else {
                    if (resultArray[i].participant.firstName) {
                        sender = resultArray[i].participant.firstName + ' ';
                    }
                    if (resultArray[i].participant.lastName) {
                        sender += resultArray[i].participant.lastName;
                    }
                    sender += ','
                }

                line += new Date(resultArray[i].time).toLocaleDateString('fa-IR') + ',';
                line += new Date(resultArray[i].time).toLocaleTimeString('fa-IR') + ',';
                line += sender;
                line += resultArray[i].participant.username + ',';
                line += '"' + resultArray[i].message.replace(/,/g, "،").replace(/"/g, '”') + '",';
                // line += result[i].message.replaceAll(",", ".").replace(/(\r\n|\n|\r)/gm, " ") + ',';
                str += line + '\r\n';
                radif++;
            }


            app.fs.writeFile(params.exportPath, universalBOM + str, {encoding: 'utf8'}, function (err) {
                if (err) {
                    app.chatEvents.fireEvent('ERROR', {
                        code: null,
                        message: err
                    });
                    callback && callback({
                        hasError: true,
                        code: null,
                        message: err
                    });
                }

                callback && callback({
                    hasError: false,
                    result: resultArray,
                    exportPath: params.exportPath
                });

                app.chatEvents.fireEvent('threadEvents', {
                    type: 'EXPORT_CHAT',
                    subType: 'EXPORTED_TO_DIRECTORY',
                    threadId: sendData.subjectId,
                    exportPath: params.exportPath,
                    result: resultArray
                });


            });

            callback = undefined;
        })
    }

    this.getImage = getImage;

    this.getFile = getFile;

    this.getFileFromPodspace = function (params, callback) {
        return app.fileMethods.getFileFromPodspaceNew(params, callback);
    };

    this.getImageFromPodspace = function (params, callback) {
        return app.fileMethods.getImageFromPodspaceV2(params, callback);
    };

    this.uploadFile = uploadFile;

    this.uploadImage = uploadImage;

    this.podspace = podspace;

    this.cancelFileUpload = function (params, callback) {
        return app.fileMethods.cancelFileUpload(params, callback);
    };

    this.cancelFileDownload = function (params, callback) {
        return app.fileMethods.cancelFileDownload(params, callback);
    };

    this.editMessage = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.EDIT_MESSAGE,
            typeCode: params.typeCode,
            messageType: params.messageType,
            subjectId: params.messageId,
            repliedTo: params.repliedTo,
            content: params.content,
            uniqueId: params.uniqueId,
            metadata: params.metadata,
            systemMetadata: params.systemMetadata,
            pushMsgType: 3
        }, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            editedMessage: formatDataToMakeMessage(undefined, messageContent)
                        };

                    returnData.result = resultData;

                    /**
                     * Update Message on cache
                     */
                    // if (canUseCache && cacheSecret.length > 0) {
                    //     if (db) {
                    //         try {
                    //             var tempData = {},
                    //                 salt = Utility.generateUUID();
                    //             tempData.id = parseInt(resultData.editedMessage.id);
                    //             tempData.owner = parseInt(app.store.user.get().id);
                    //             tempData.threadId = parseInt(resultData.editedMessage.threadId);
                    //             tempData.time = resultData.editedMessage.time;
                    //             tempData.message = Utility.crypt(resultData.editedMessage.message, cacheSecret, salt);
                    //             tempData.data = Utility.crypt(JSON.stringify(unsetNotSeenDuration(resultData.editedMessage)), cacheSecret, salt);
                    //             tempData.salt = salt;
                    //
                    //             /**
                    //              * Insert Message into cache database
                    //              */
                    //             db.messages.put(tempData)
                    //                 .catch(function (error) {
                    //                     fireEvent('error', {
                    //                         code: error.code,
                    //                         message: error.message,
                    //                         error: error
                    //                     });
                    //                 });
                    //         } catch (error) {
                    //             fireEvent('error', {
                    //                 code: error.code,
                    //                 message: error.message,
                    //                 error: error
                    //             });
                    //         }
                    //     } else {
                    //         fireEvent('error', {
                    //             code: 6601,
                    //             message: CHAT_ERRORS[6601],
                    //             error: null
                    //         });
                    //     }
                    // }
                }

                callback && callback(returnData);
            }
        });
    };

    this.deleteMessage = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.DELETE_MESSAGE,
            typeCode: params.typeCode,
            subjectId: params.messageId,
            uniqueId: params.uniqueId,
            content: JSON.stringify({
                'deleteForAll': (typeof params.deleteForAll === 'boolean') ? params.deleteForAll : false
            }),
            pushMsgType: 3
        }, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            deletedMessage: {
                                id: result.result.id,
                                pinned: result.result.pinned,
                                mentioned: result.result.mentioned,
                                messageType: result.result.messageType,
                                edited: result.result.edited,
                                editable: result.result.editable,
                                deletable: result.result.deletable
                            }
                        };

                    returnData.result = resultData;

                    /**
                     * Remove Message from cache
                     */
                    // if (canUseCache) {
                    //     if (db) {
                    //         db.messages.where('id')
                    //             .equals(parseInt(result.result))
                    //             .delete()
                    //             .catch(function (error) {
                    //                 fireEvent('error', {
                    //                     code: 6602,
                    //                     message: CHAT_ERRORS[6602],
                    //                     error: error
                    //                 });
                    //             });
                    //     } else {
                    //         fireEvent('error', {
                    //             code: 6601,
                    //             message: CHAT_ERRORS[6601],
                    //             error: null
                    //         });
                    //     }
                    // }

                }

                callback && callback(returnData);
            }
        });
    };

    this.deleteMultipleMessages = function (params, callback) {
        var messageIdsList = params.messageIds,
            uniqueIdsList = [];

        for (let i in messageIdsList) {
            var messageId = messageIdsList[i];

            var uniqueId = Utility.generateUUID();
            uniqueIdsList.push(uniqueId);

            app.store.messagesCallbacks[uniqueId] = function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        resultData = {
                            deletedMessage: {
                                id: result.result.id,
                                pinned: result.result.pinned,
                                mentioned: result.result.mentioned,
                                messageType: result.result.messageType,
                                edited: result.result.edited,
                                editable: result.result.editable,
                                deletable: result.result.deletable
                            }
                        };

                    returnData.result = resultData;

                    /**
                     * Remove Message from cache
                     */
                    // if (canUseCache) {
                    //     if (db) {
                    //         db.messages.where('id')
                    //             .equals(parseInt(result.result))
                    //             .delete()
                    //             .catch(function (error) {
                    //                 fireEvent('error', {
                    //                     code: 6602,
                    //                     message: CHAT_ERRORS[6602],
                    //                     error: error
                    //                 });
                    //             });
                    //     } else {
                    //         fireEvent('error', {
                    //             code: 6601,
                    //             message: CHAT_ERRORS[6601],
                    //             error: null
                    //         });
                    //     }
                    // }
                }

                callback && callback(returnData);
            };
        }

        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.DELETE_MESSAGE,
            typeCode: params.typeCode,
            content: {
                uniqueIds: uniqueIdsList,
                ids: messageIdsList,
                deleteForAll: (typeof params.deleteForAll === 'boolean') ? params.deleteForAll : false
            },
            pushMsgType: 3
        });
    };

    this.replyTextMessage = function (params, callbacks) {
        var uniqueId;

        if (typeof params.uniqueId != 'undefined') {
            uniqueId = params.uniqueId;
        } else {
            uniqueId = Utility.generateUUID();
        }

        putInChatSendQueue({
            message: {
                chatMessageVOType: chatMessageVOTypes.MESSAGE,
                typeCode: params.typeCode,
                messageType: 1,
                subjectId: params.threadId,
                repliedTo: params.repliedTo,
                content: params.textMessage,
                uniqueId: uniqueId,
                systemMetadata: JSON.stringify(params.systemMetadata),
                metadata: JSON.stringify(params.metadata),
                pushMsgType: 3
            },
            callbacks: callbacks
        }, function () {
            chatSendQueueHandler();
        });

        return {
            uniqueId: uniqueId,
            threadId: params.threadId,
            participant: app.store.user.get(),
            content: params.content
        };
    };

    this.replyFileMessage = function (params, callbacks) {
        var metadata = {file: {}},
            fileUploadParams = {},
            fileUniqueId = Utility.generateUUID();

        if (!params.userGroupHash || params.userGroupHash.length == 0 || typeof (params.userGroupHash) != 'string') {
            app.chatEvents.fireEvent('error', {
                code: 6304,
                message: CHAT_ERRORS[6304]
            });
            return;
        } else {
            fileUploadParams.userGroupHash = params.userGroupHash;
        }

        return chatUploadHandler({
            threadId: params.threadId,
            file: params.file,
            fileUniqueId: fileUniqueId
        }, function (uploadHandlerResult, uploadHandlerMetadata, fileType, fileExtension) {
            fileUploadParams = Object.assign(fileUploadParams, uploadHandlerResult);

            putInChatUploadQueue({
                message: {
                    chatMessageVOType: chatMessageVOTypes.MESSAGE,
                    typeCode: params.typeCode,
                    messageType: (params.messageType && params.messageType.toUpperCase() !== undefined && chatMessageTypes[params.messageType.toUpperCase()] > 0) ? chatMessageTypes[params.messageType.toUpperCase()] : 1,
                    subjectId: params.threadId,
                    repliedTo: params.repliedTo,
                    content: params.content,
                    metadata: JSON.stringify(uploadHandlerMetadata),
                    systemMetadata: JSON.stringify(params.systemMetadata),
                    uniqueId: fileUniqueId,
                    pushMsgType: 3
                },
                callbacks: callbacks
            }, function () {
                if (imageMimeTypes.indexOf(fileType) >= 0 || imageExtentions.indexOf(fileExtension) >= 0) {
                    app.fileMethods.uploadImageToPodspaceUserGroup(fileUploadParams, function (result) {
                        if (!result.hasError) {
                            metadata['name'] = result.result.name;
                            metadata['fileHash'] = result.result.hashCode;
                            metadata['file']['name'] = result.result.name;
                            metadata['file']['fileHash'] = result.result.hashCode;
                            metadata['file']['hashCode'] = result.result.hashCode;
                            metadata['file']['actualHeight'] = result.result.actualHeight;
                            metadata['file']['actualWidth'] = result.result.actualWidth;
                            metadata['file']['parentHash'] = result.result.parentHash;
                            metadata['file']['size'] = result.result.size;
                            metadata['file']['link'] = `https://podspace.pod.ir/nzh/drive/downloadImage?hash=${result.result.hashCode}`;
                            transferFromUploadQToSendQ(parseInt(params.threadId), fileUniqueId, JSON.stringify(metadata), function () {
                                chatSendQueueHandler();
                            });
                        } else {
                            app.messageQueues.deleteFromChatUploadQueue({message: {uniqueId: fileUniqueId}});
                        }
                    });
                } else {
                    app.fileMethods.uploadFileToPodspace(fileUploadParams, function (result) {
                        if (!result.hasError) {
                            metadata['fileHash'] = result.result.hashCode;
                            metadata['name'] = result.result.name;
                            metadata['file']['name'] = result.result.name;
                            metadata['file']['fileHash'] = result.result.hashCode;
                            metadata['file']['hashCode'] = result.result.hashCode;
                            metadata['file']['parentHash'] = result.result.parentHash;
                            metadata['file']['size'] = result.result.size;

                            transferFromUploadQToSendQ(parseInt(params.threadId), fileUniqueId, JSON.stringify(metadata), function () {
                                chatSendQueueHandler();
                            });
                        } else {
                            app.messageQueues.deleteFromChatUploadQueue({message: {uniqueId: fileUniqueId}});
                        }
                    });
                }
            });
        });
    };

    this.forwardMessage = function (params, callbacks) {
        var threadId = params.threadId,
            messageIdsList = params.messageIds,
            uniqueIdsList = [];

        for (let i in messageIdsList) {
            if (!app.store.threadCallbacks[threadId]) {
                app.store.threadCallbacks[threadId] = {};
            }

            var uniqueId = Utility.generateUUID();
            uniqueIdsList.push(uniqueId);

            app.store.threadCallbacks[threadId][uniqueId] = {};

            app.store.sendMessageCallbacks[uniqueId] = {};

            if (callbacks.onSent) {
                app.store.sendMessageCallbacks[uniqueId].onSent = callbacks.onSent;
                app.store.threadCallbacks[threadId][uniqueId].onSent = false;
                app.store.threadCallbacks[threadId][uniqueId].uniqueId = uniqueId;
            }

            if (callbacks.onSeen) {
                app.store.sendMessageCallbacks[uniqueId].onSeen = callbacks.onSeen;
                app.store.threadCallbacks[threadId][uniqueId].onSeen = false;
            }

            if (callbacks.onDeliver) {
                app.store.sendMessageCallbacks[uniqueId].onDeliver = callbacks.onDeliver;
                app.store.threadCallbacks[threadId][uniqueId].onDeliver = false;
            }
        }

        putInChatSendQueue({
            message: {
                chatMessageVOType: chatMessageVOTypes.FORWARD_MESSAGE,
                typeCode: params.typeCode,
                subjectId: params.threadId,
                repliedTo: params.repliedTo,
                content: messageIdsList,
                uniqueId: uniqueIdsList,
                metadata: JSON.stringify(params.metadata),
                pushMsgType: 3
            },
            callbacks: callbacks
        }, function () {
            chatSendQueueHandler();
        });
    };

    this.deliver = deliver;

    this.seen = function (params) {
        if (app.store.user.get() && params.ownerId !== app.store.user.get().id) {
            return app.messenger.sendMessage({
                chatMessageVOType: chatMessageVOTypes.SEEN,
                typeCode: params.typeCode,
                content: params.messageId,
                pushMsgType: 3
            });
        }
    };

    this.startTyping = function (params) {
        var uniqueId = Utility.generateUUID();

        if (parseInt(params.threadId) > 0) {
            var threadId = params.threadId;
        }
        isTypingInterval && clearInterval(isTypingInterval);

        isTypingInterval = setInterval(function () {
            sendSystemMessage({
                content: JSON.stringify({
                    type: systemMessageTypes.IS_TYPING
                }),
                threadId: threadId,
                uniqueId: uniqueId
            });
        }, app.sdkParams.systemMessageIntervalPitch);
    };

    this.stopTyping = function () {
        isTypingInterval && clearInterval(isTypingInterval);
    };

    this.getMessageDeliveredList = function (params, callback) {

        var deliveryListData = {
            chatMessageVOType: chatMessageVOTypes.GET_MESSAGE_DELEVERY_PARTICIPANTS,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token,
            timeout: params.timeout
        };

        if (params) {
            if (parseInt(params.messageId) > 0) {
                deliveryListData.content.messageId = params.messageId;
            }
        }

        return app.messenger.sendMessage(deliveryListData, {
            onResult: function (result) {
                if (typeof result.result == 'object') {
                    for (var i = 0; i < result.result.length; i++) {
                        result.result[i] = formatDataToMakeUser(result.result[i]);
                    }
                }
                callback && callback(result);
            }
        });
    };

    this.getMessageSeenList = function (params, callback) {
        var seenListData = {
            chatMessageVOType: chatMessageVOTypes.GET_MESSAGE_SEEN_PARTICIPANTS,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token,
            timeout: params.timeout
        };

        if (params) {
            if (parseInt(params.messageId) > 0) {
                seenListData.content.messageId = params.messageId;
            }
        }

        return app.messenger.sendMessage(seenListData, {
            onResult: function (result) {
                if (typeof result.result == 'object') {
                    for (var i = 0; i < result.result.length; i++) {
                        result.result[i] = formatDataToMakeUser(result.result[i]);
                    }
                }
                callback && callback(result);
            }
        });
    };

    this.updateThreadInfo = updateThreadInfo;

    this.updateChatProfile = updateChatProfile;

    this.muteThread = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.MUTE_THREAD,
            typeCode: params.typeCode,
            subjectId: params.threadId,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        }, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.unMuteThread = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.UNMUTE_THREAD,
            typeCode: params.typeCode,
            subjectId: params.threadId,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        }, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.joinPublicThread = function (params, callback) {
        var joinThreadData = {
            chatMessageVOType: chatMessageVOTypes.JOIN_THREAD,
            typeCode: params.typeCode,
            content: '',
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof params.uniqueName === 'string' && params.uniqueName.length > 0) {
                joinThreadData.content = params.uniqueName;
            }
        }

        return app.messenger.sendMessage(joinThreadData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.isPublicThreadNameAvailable = function (params, callback) {
        var isNameAvailableData = {
            chatMessageVOType: chatMessageVOTypes.IS_NAME_AVAILABLE,
            typeCode: params.typeCode,
            content: '',
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof params.uniqueName === 'string' && params.uniqueName.length > 0) {
                isNameAvailableData.content = params.uniqueName;
            }
        }

        return app.messenger.sendMessage(isNameAvailableData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.pinThread = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.PIN_THREAD,
            typeCode: params.typeCode,
            subjectId: params.threadId,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        }, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.unPinThread = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.UNPIN_THREAD,
            typeCode: params.typeCode,
            subjectId: params.threadId,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        }, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.pinMessage = function (params, callback) {
        return app.messenger.sendMessage({
            chatMessageVOType: chatMessageVOTypes.PIN_MESSAGE,
            typeCode: params.typeCode,
            subjectId: params.messageId,
            content: JSON.stringify({
                'notifyAll': (typeof params.notifyAll === 'boolean') ? params.notifyAll : false
            }),
            pushMsgType: 3,
            token: app.sdkParams.token
        }, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.unPinMessage = unPinMessage;

    this.spamPrivateThread = function (params, callback) {
        var spamData = {
            chatMessageVOType: chatMessageVOTypes.SPAM_PV_THREAD,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token,
            timeout: params.timeout
        };

        if (params) {
            if (parseInt(params.threadId) > 0) {
                spamData.subjectId = params.threadId;
            }
        }

        return app.messenger.sendMessage(spamData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.block = function (params, callback) {

        var blockData = {
            chatMessageVOType: chatMessageVOTypes.BLOCK,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token,
            timeout: params.timeout
        };

        if (params) {
            if (parseInt(params.contactId) > 0) {
                blockData.content.contactId = params.contactId;
            }

            if (parseInt(params.threadId) > 0) {
                blockData.content.threadId = params.threadId;
            }

            if (parseInt(params.userId) > 0) {
                blockData.content.userId = params.userId;
            }
        }

        return app.messenger.sendMessage(blockData, {
            onResult: function (result) {
                if (typeof result.result == 'object') {
                    result.result = formatDataToMakeBlockedUser(result.result);
                }
                callback && callback(result);
            }
        });
    };

    this.unblock = function (params, callback) {
        var unblockData = {
            chatMessageVOType: chatMessageVOTypes.UNBLOCK,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token,
            content: {},
            timeout: params.timeout
        };

        if (params) {
            if (parseInt(params.blockId) > 0) {
                unblockData.subjectId = params.blockId;
            }

            if (parseInt(params.contactId) > 0) {
                unblockData.content.contactId = params.contactId;
            }

            if (parseInt(params.threadId) > 0) {
                unblockData.content.threadId = params.threadId;
            }

            if (parseInt(params.userId) > 0) {
                unblockData.content.userId = params.userId;
            }
        }

        return app.messenger.sendMessage(unblockData, {
            onResult: function (result) {
                if (typeof result.result == 'object') {
                    result.result = formatDataToMakeBlockedUser(result.result);
                }

                callback && callback(result);
            }
        });
    };

    this.getBlockedList = function (params, callback) {
        var count = 50,
            offset = 0,
            content = {};

        if (params) {
            if (parseInt(params.count) > 0) {
                count = params.count;
            }

            if (parseInt(params.offset) > 0) {
                offset = params.offset;
            }
        }

        content.count = count;
        content.offset = offset;

        var getBlockedData = {
            chatMessageVOType: chatMessageVOTypes.GET_BLOCKED,
            typeCode: params.typeCode,
            content: content,
            pushMsgType: 3,
            token: app.sdkParams.token,
            timeout: params.timeout
        };

        return app.messenger.sendMessage(getBlockedData, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result,
                        messageLength = messageContent.length,
                        resultData = {
                            blockedUsers: [],
                            contentCount: result.contentCount,
                            hasNext: (offset + count < result.contentCount && messageLength > 0),
                            nextOffset: offset + messageLength
                        },
                        blockedUser;

                    for (var i = 0; i < messageLength; i++) {
                        blockedUser = formatDataToMakeBlockedUser(messageContent[i]);
                        if (blockedUser) {
                            resultData.blockedUsers.push(blockedUser);
                        }
                    }

                    returnData.result = resultData;
                }

                callback && callback(returnData);
            }
        });
    };

    this.getUserNotSeenDuration = function (params, callback) {
        var content = {};

        if (params) {
            if (Array.isArray(params.userIds)) {
                content.userIds = params.userIds;
            }
        }

        var getNotSeenDurationData = {
            chatMessageVOType: chatMessageVOTypes.GET_NOT_SEEN_DURATION,
            typeCode: params.typeCode,
            content: content,
            pushMsgType: 3,
            token: app.sdkParams.token,
            timeout: params.timeout
        };

        return app.messenger.sendMessage(getNotSeenDurationData, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    returnData.result = result.result;
                }

                callback && callback(returnData);
            }
        });
    };

    this.addContacts = function (params, callback) {
        var data = {};

        if (params) {
            if (typeof params.firstName === 'string') {
                data.firstName = params.firstName;
            } else {
                data.firstName = '';
            }

            if (typeof params.lastName === 'string') {
                data.lastName = params.lastName;
            } else {
                data.lastName = '';
            }

            if (typeof params.typeCode === 'string') {
                data.typeCode = params.typeCode;
            } else if (app.sdkParams.generalTypeCode) {
                data.typeCode = app.sdkParams.generalTypeCode;
            }

            if (typeof params.ownerId === 'string') {
                data.ownerId = params.ownerId;
            } else if (app.sdkParams.generalTypeCode) {
                data.ownerId = app.sdkParams.typeCodeOwnerId;
            }

            if (typeof params.cellphoneNumber === 'string') {
                data.cellphoneNumber = params.cellphoneNumber;
            } else {
                data.cellphoneNumber = '';
            }

            if (typeof params.email === 'string') {
                data.email = params.email;
            } else {
                data.email = '';
            }

            if (typeof params.username === 'string') {
                data.username = params.username;
            }

            data.uniqueId = Utility.generateUUID();
        }

        var requestParams = {
            url: app.sdkParams.SERVICE_ADDRESSES.PLATFORM_ADDRESS + SERVICES_PATH.ADD_CONTACTS,
            method: 'POST',
            data: data,
            headers: {
                '_token_': app.sdkParams.token,
                '_token_issuer_': 1
            }
        };

        app.httpRequest.httpRequest(requestParams, function (result) {
            if (!result.hasError) {
                var responseData = JSON.parse(result.result.responseText);

                var returnData = {
                    hasError: responseData.hasError,
                    cache: false,
                    errorMessage: responseData.message,
                    errorCode: responseData.errorCode,
                    referenceNumber: responseData.referenceNumber,
                    ott: responseData.ott
                };

                if (!responseData.hasError) {
                    var messageContent = responseData.result,
                        messageLength = responseData.result.length,
                        resultData = {
                            contacts: [],
                            contentCount: messageLength
                        },
                        contactData;

                    for (var i = 0; i < messageLength; i++) {
                        contactData = formatDataToMakeContact(messageContent[i]);
                        if (contactData) {
                            resultData.contacts.push(contactData);
                        }
                    }

                    returnData.result = resultData;
                }

                callback && callback(returnData);

            } else {
                app.chatEvents.fireEvent('error', {
                    code: result.errorCode,
                    message: result.errorMessage + ', referenceNumber:' + result.referenceNumber + ', ott:' + result.ott,
                    error: result
                });
            }
        });
    };

    this.updateContacts = function (params, callback) {
        var data = {};

        if (params) {
            if (parseInt(params.id) > 0) {
                data.id = parseInt(params.id);
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] ID is required for Updating Contact!',
                    error: undefined
                });
            }

            if (typeof params.firstName === 'string') {
                data.firstName = params.firstName;
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] firstName is required for Updating Contact!'
                });
            }

            if (typeof params.lastName === 'string') {
                data.lastName = params.lastName;
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] lastName is required for Updating Contact!'
                });
            }

            if (typeof params.cellphoneNumber === 'string') {
                data.cellphoneNumber = params.cellphoneNumber;
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] cellphoneNumber is required for Updating Contact!'
                });
            }

            if (typeof params.email === 'string') {
                data.email = params.email;
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] email is required for Updating Contact!'
                });
            }
            if (typeof params.typeCode === 'string') {
                data.typeCode = params.typeCode;
            } else if (app.sdkParams.generalTypeCode) {
                data.typeCode = app.sdkParams.generalTypeCode;
            }

            if (typeof params.ownerId === 'string') {
                data.ownerId = params.ownerId;
            } else if (app.sdkParams.typeCodeOwnerId) {
                data.ownerId = app.sdkParams.typeCodeOwnerId;
            }

            data.uniqueId = Utility.generateUUID();
        }

        var requestParams = {
            url: app.sdkParams.SERVICE_ADDRESSES.PLATFORM_ADDRESS +
                SERVICES_PATH.UPDATE_CONTACTS,
            method: 'GET',
            data: data,
            headers: {
                '_token_': app.sdkParams.token,
                '_token_issuer_': 1
            }
        };

        app.httpRequest.httpRequest(requestParams, function (result) {
            if (!result.hasError) {
                var responseData = JSON.parse(result.result.responseText);

                var returnData = {
                    hasError: responseData.hasError,
                    cache: false,
                    errorMessage: responseData.message,
                    errorCode: responseData.errorCode
                };

                if (!responseData.hasError) {
                    var messageContent = responseData.result,
                        messageLength = responseData.result.length,
                        resultData = {
                            contacts: [],
                            contentCount: messageLength
                        },
                        contactData;

                    for (var i = 0; i < messageLength; i++) {
                        contactData = formatDataToMakeContact(messageContent[i]);
                        if (contactData) {
                            resultData.contacts.push(contactData);
                        }
                    }

                    returnData.result = resultData;
                }

                callback && callback(returnData);

            } else {
                app.chatEvents.fireEvent('error', {
                    code: result.errorCode,
                    message: result.errorMessage + ', referenceNumber:' + result.referenceNumber + ', ott:' + result.ott,
                    error: result
                });
            }
        });
    };

    this.removeContacts = function (params, callback) {
        var data = {};

        if (params) {
            if (parseInt(params.id) > 0) {
                data.id = parseInt(params.id);
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] ID is required for Deleting Contact!',
                    error: undefined
                });
            }
        }

        var requestParams = {
            url: app.sdkParams.SERVICE_ADDRESSES.PLATFORM_ADDRESS + SERVICES_PATH.REMOVE_CONTACTS,
            method: 'POST',
            data: data,
            headers: {
                '_token_': app.sdkParams.token,
                '_token_issuer_': 1
            }
        };

        app.httpRequest.httpRequest(requestParams, function (result) {
            if (!result.hasError) {
                var responseData = JSON.parse(result.result.responseText);

                var returnData = {
                    hasError: responseData.hasError,
                    cache: false,
                    errorMessage: responseData.message,
                    errorCode: responseData.errorCode
                };

                if (!responseData.hasError) {
                    returnData.result = responseData.result;
                }

                callback && callback(returnData);

            } else {
                app.chatEvents.fireEvent('error', {
                    code: result.errorCode,
                    message: result.errorMessage + ', referenceNumber:' + result.referenceNumber + ', ott:' + result.ott,
                    error: result
                });
            }
        });
    };

    this.searchContacts = function (params, callback) {
        var data = {
                size: 50,
                offset: 0
            },
            whereClause = {},
            returnCache = false;

        if (params) {
            if (typeof params.firstName === 'string') {
                data.firstName = whereClause.firstName = params.firstName;
            }

            if (typeof params.lastName === 'string') {
                data.lastName = whereClause.lastName = params.lastName;
            }

            if (parseInt(params.cellphoneNumber) > 0) {
                data.cellphoneNumber = whereClause.cellphoneNumber = params.cellphoneNumber;
            }

            if (typeof params.email === 'string') {
                data.email = whereClause.email = params.email;
            }

            if (typeof params.query === 'string') {
                data.q = whereClause.q = params.query;
            }

            if (typeof params.uniqueId === 'string') {
                data.uniqueId = whereClause.uniqueId = params.uniqueId;
            }

            if (parseInt(params.id) > 0) {
                data.id = whereClause.id = params.id;
            }

            if (parseInt(params.typeCode) > 0) {
                data.typeCode = whereClause.typeCode = params.typeCode;
            }

            if (parseInt(params.size) > 0) {
                data.size = params.size;
            }

            if (parseInt(params.offset) > 0) {
                data.offset = params.offset;
            }

            var functionLevelCache = (typeof params.cache == 'boolean') ? params.cache : true;
        }

        var requestParams = {
            url: app.sdkParams.SERVICE_ADDRESSES.PLATFORM_ADDRESS + SERVICES_PATH.SEARCH_CONTACTS,
            method: 'POST',
            data: data,
            headers: {
                '_token_': app.sdkParams.token,
                '_token_issuer_': 1
            }
        };

        /**
         * Get Search Contacts Result From Server
         */
        app.httpRequest.httpRequest(requestParams, function (result) {
            if (!result.hasError) {
                var responseData = JSON.parse(result.result.responseText);

                var returnData = {
                    hasError: responseData.hasError,
                    cache: false,
                    errorMessage: responseData.message,
                    errorCode: responseData.errorCode
                };

                if (!responseData.hasError) {
                    var messageContent = responseData.result,
                        messageLength = responseData.result.length,
                        resultData = {
                            contacts: [],
                            contentCount: messageLength
                        },
                        contactData;

                    for (var i = 0; i < messageLength; i++) {
                        contactData = formatDataToMakeContact(messageContent[i]);
                        if (contactData) {
                            resultData.contacts.push(contactData);
                        }
                    }

                    returnData.result = resultData;
                }

                callback && callback(returnData);
                /**
                 * Delete callback so if server pushes response before
                 * cache, cache won't send data again
                 */
                callback = undefined;

                if (!returnData.hasError && returnCache) {
                    app.chatEvents.fireEvent('contactEvents', {
                        type: 'CONTACTS_SEARCH_RESULT_CHANGE',
                        result: returnData.result
                    });
                }
            } else {
                app.chatEvents.fireEvent('error', {
                    code: result.errorCode,
                    message: result.errorMessage,
                    error: result
                });
            }
        });
    };

    this.createBot = function (params, callback) {
        var createBotData = {
            chatMessageVOType: chatMessageVOTypes.CREATE_BOT,
            typeCode: params.typeCode,
            content: '',
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof params.botName === 'string' && params.botName.length > 0) {
                if (params.botName.substr(-3) === "BOT") {
                    createBotData.content = params.botName;
                } else {
                    app.chatEvents.fireEvent('error', {
                        code: 999,
                        message: '[SDK] Bot name should end in "BOT", ex. "testBOT"'
                    });
                    return;
                }
            } else {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] Insert a bot name to create one!'
                });
                return;
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] Insert a bot name to create one!'
            });
            return;
        }

        return app.messenger.sendMessage(createBotData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.defineBotCommand = function (params, callback) {
        var defineBotCommandData = {
            chatMessageVOType: chatMessageVOTypes.DEFINE_BOT_COMMAND,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        }, commandList = [];

        if (params) {
            if (typeof params.botName !== 'string' || params.botName.length == 0) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] You need to insert a botName!'
                });
                return;
            }

            if (!Array.isArray(params.commandList) || !params.commandList.length) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] Bot Commands List has to be an array of strings.'
                });
                return;
            } else {
                for (var i = 0; i < params.commandList.length; i++) {
                    commandList.push('/' + params.commandList[i].trim());
                }
            }

            defineBotCommandData.content = {
                botName: params.botName.trim(),
                commandList: commandList
            };

        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to create bot commands'
            });
            return;
        }

        return app.messenger.sendMessage(defineBotCommandData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.startBot = function (params, callback) {
        var startBotData = {
            chatMessageVOType: chatMessageVOTypes.START_BOT,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof +params.threadId !== 'number' || params.threadId < 0) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] Enter a valid Thread Id for Bot to start in!'
                });
                return;
            }

            if (typeof params.botName !== 'string' || params.botName.length == 0) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] You need to insert a botName!'
                });
                return;
            }

            startBotData.subjectId = +params.threadId;

            startBotData.content = JSON.stringify({
                botName: params.botName.trim()
            });

        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to create bot commands'
            });
            return;
        }

        return app.messenger.sendMessage(startBotData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.stopBot = function (params, callback) {
        var stopBotData = {
            chatMessageVOType: chatMessageVOTypes.STOP_BOT,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        }, commandList = [];

        if (params) {
            if (typeof +params.threadId !== 'number' || params.threadId < 0) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] Enter a valid Thread Id for Bot to stop on!'
                });
                return;
            }

            if (typeof params.botName !== 'string' || params.botName.length == 0) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] You need to insert a botName!'
                });
                return;
            }

            stopBotData.subjectId = +params.threadId;

            stopBotData.content = JSON.stringify({
                botName: params.botName.trim()
            });

        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to create bot commands'
            });
            return;
        }

        return app.messenger.sendMessage(stopBotData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.getBotCommandsList = function (params, callback) {
        var getBotCommandsListData = {
            chatMessageVOType: chatMessageVOTypes.BOT_COMMANDS,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof params.botName !== 'string' || params.botName.length == 0) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] You need to insert a botName!'
                });
                return;
            }

            getBotCommandsListData.content = JSON.stringify({
                botName: params.botName.trim()
            });

        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to get bot commands'
            });
            return;
        }

        return app.messenger.sendMessage(getBotCommandsListData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.getThreadAllBots = function (params, callback) {
        var getThreadBotsData = {
            chatMessageVOType: chatMessageVOTypes.THREAD_ALL_BOTS,
            typeCode: params.typeCode,
            content: {},
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof +params.threadId !== 'number' || params.threadId < 0) {
                app.chatEvents.fireEvent('error', {
                    code: 999,
                    message: '[SDK] Enter a valid Thread Id to get all Bots List!'
                });
                return;
            }

            getThreadBotsData.subjectId = +params.threadId;

        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to get thread\' bots list!'
            });
            return;
        }

        return app.messenger.sendMessage(getThreadBotsData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.mapReverse = function (params, callback) {
        var data = {};

        if (params) {
            if (parseFloat(params.lat) > 0) {
                data.lat = params.lat;
            }

            if (parseFloat(params.lng) > 0) {
                data.lng = params.lng;
            }

            data.uniqueId = Utility.generateUUID();
        }

        var requestParams = {
            url: app.sdkParams.SERVICE_ADDRESSES.MAP_ADDRESS + SERVICES_PATH.REVERSE,
            method: 'GET',
            data: data,
            headers: {
                'Api-Key': app.sdkParams.mapApiKey
            }
        };

        app.httpRequest.httpRequest(requestParams, function (result) {
            if (!result.hasError) {
                var responseData = JSON.parse(result.result.responseText);

                var returnData = {
                    hasError: result.hasError,
                    cache: result.cache,
                    errorMessage: result.message,
                    errorCode: result.errorCode,
                    result: responseData
                };

                callback && callback(returnData);

            } else {
                app.chatEvents.fireEvent('error', {
                    code: result.errorCode,
                    message: result.errorMessage,
                    error: result
                });
            }
        });
    };

    this.mapSearch = function (params, callback) {
        var data = {};

        if (params) {
            if (typeof params.term === 'string') {
                data.term = params.term;
            }

            if (parseFloat(params.lat) > 0) {
                data.lat = params.lat;
            }

            if (parseFloat(params.lng) > 0) {
                data.lng = params.lng;
            }

            data.uniqueId = Utility.generateUUID();
        }

        var requestParams = {
            url: app.sdkParams.SERVICE_ADDRESSES.MAP_ADDRESS + SERVICES_PATH.SEARCH,
            method: 'GET',
            data: data,
            headers: {
                'Api-Key': app.sdkParams.mapApiKey
            }
        };

        app.httpRequest.httpRequest(requestParams, function (result) {
            if (!result.hasError) {
                var responseData = JSON.parse(result.result.responseText);

                var returnData = {
                    hasError: result.hasError,
                    cache: result.cache,
                    errorMessage: result.message,
                    errorCode: result.errorCode,
                    result: responseData
                };

                callback && callback(returnData);

            } else {
                app.chatEvents.fireEvent('error', {
                    code: result.errorCode,
                    message: result.errorMessage,
                    error: result
                });
            }
        });
    };

    this.mapRouting = function (params, callback) {
        var data = {};

        if (params) {
            if (typeof params.alternative === 'boolean') {
                data.alternative = params.alternative;
            } else {
                data.alternative = true;
            }

            if (typeof params.origin === 'object') {
                if (parseFloat(params.origin.lat) > 0 && parseFloat(params.origin.lng)) {
                    data.origin = params.origin.lat + ',' + parseFloat(params.origin.lng);
                } else {
                    console.log('No origin has been selected!');
                }
            }

            if (typeof params.destination === 'object') {
                if (parseFloat(params.destination.lat) > 0 && parseFloat(params.destination.lng)) {
                    data.destination = params.destination.lat + ',' + parseFloat(params.destination.lng);
                } else {
                    console.log('No destination has been selected!');
                }
            }

            data.uniqueId = Utility.generateUUID();
        }

        var requestParams = {
            url: app.sdkParams.SERVICE_ADDRESSES.MAP_ADDRESS + SERVICES_PATH.ROUTING,
            method: 'GET',
            data: data,
            headers: {
                'Api-Key': app.sdkParams.mapApiKey
            }
        };

        app.httpRequest.httpRequest(requestParams, function (result) {
            if (!result.hasError) {
                var responseData = JSON.parse(result.result.responseText);

                var returnData = {
                    hasError: result.hasError,
                    cache: result.cache,
                    errorMessage: result.message,
                    errorCode: result.errorCode,
                    result: responseData
                };

                callback && callback(returnData);

            } else {
                app.chatEvents.fireEvent('error', {
                    code: result.errorCode,
                    message: result.errorMessage,
                    error: result
                });
            }
        });
    };

    this.createSelfThread = function (params, callback) {
        var content = {
            type: createThreadTypes['SELF']
        };

        if (params) {
            if (typeof params.description === 'string') {
                content.description = params.description;
            }

            if (typeof params.metadata === 'string') {
                content.metadata = params.metadata;
            } else if (typeof params.metadata === 'object') {
                try {
                    content.metadata = JSON.stringify(params.metadata);
                } catch (e) {
                    app.sdkParams.consoleLogging && console.log(e);
                }
            }

            if (typeof params.message == 'object') {
                content.message = {};

                if (typeof params.message.text === 'string') {
                    content.message.text = params.message.text;
                }

                if (typeof params.message.uniqueId === 'string') {
                    content.message.uniqueId = params.message.uniqueId;
                }

                if (params.message.type > 0) {
                    content.message.messageType = params.message.type;
                }

                if (params.message.repliedTo > 0) {
                    content.message.repliedTo = params.message.repliedTo;
                }

                if (typeof params.message.metadata === 'string') {
                    content.message.metadata = params.message.metadata;
                } else if (typeof params.message.metadata === 'object') {
                    content.message.metadata = JSON.stringify(params.message.metadata);
                }

                if (typeof params.message.systemMetadata === 'string') {
                    content.message.systemMetadata = params.message.systemMetadata;
                } else if (typeof params.message.systemMetadata === 'object') {
                    content.message.systemMetadata = JSON.stringify(params.message.systemMetadata);
                }

                if (Array.isArray(params.message.forwardedMessageIds)) {
                    content.message.forwardedMessageIds = params.message.forwardedMessageIds;
                    content.message.forwardedUniqueIds = [];
                    for (var i = 0; i < params.message.forwardedMessageIds.length; i++) {
                        content.message.forwardedUniqueIds.push(Utility.generateUUID());
                    }
                }

            }
        }

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.CREATE_THREAD,
            typeCode: app.sdkParams.generalTypeCode,//params.typeCode,
            content: content
        };

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };

                if (!returnData.hasError) {
                    var messageContent = result.result;

                    returnData.result = {
                        thread: createThread(messageContent)
                    };
                }

                callback && callback(returnData);
            }
        });
    };

    this.mapStaticImage = function (params, callback) {
        var data = {},
            url = app.sdkParams.SERVICE_ADDRESSES.MAP_ADDRESS + SERVICES_PATH.STATIC_IMAGE,
            hasError = false;

        if (params) {
            if (typeof params.type === 'string') {
                data.type = params.type;
            } else {
                data.type = 'standard-night';
            }

            if (parseInt(params.zoom) > 0) {
                data.zoom = params.zoom;
            } else {
                data.zoom = 15;
            }

            if (parseInt(params.width) > 0) {
                data.width = params.width;
            } else {
                data.width = 800;
            }

            if (parseInt(params.height) > 0) {
                data.height = params.height;
            } else {
                data.height = 600;
            }

            if (typeof params.center === 'object') {
                if (parseFloat(params.center.lat) > 0 && parseFloat(params.center.lng)) {
                    data.center = params.center.lat + ',' + parseFloat(params.center.lng);
                } else {
                    hasError = true;
                    app.chatEvents.fireEvent('error', {
                        code: 6700,
                        message: CHAT_ERRORS[6700],
                        error: undefined
                    });
                }
            } else {
                hasError = true;
                app.chatEvents.fireEvent('error', {
                    code: 6700,
                    message: CHAT_ERRORS[6700],
                    error: undefined
                });
            }

            data.key = app.sdkParams.mapApiKey;
        }

        var keys = Object.keys(data);

        if (keys.length > 0) {
            url += '?';

            for (var i = 0; i < keys.length; i++) {
                var key = keys[i];
                url += key + '=' + data[key];
                if (i < keys.length - 1) {
                    url += '&';
                }
            }
        }

        var returnData = {
            hasError: hasError,
            cache: false,
            errorMessage: (hasError) ? CHAT_ERRORS[6700] : '',
            errorCode: (hasError) ? 6700 : undefined,
            result: {
                link: (!hasError) ? url : ''
            }
        };

        callback && callback(returnData);
    };

    this.setAdmin = function (params, callback) {
        setRoleToUser(params, callback);
    };

    this.removeAdmin = function (params, callback) {
        removeRoleFromUser(params, callback);
    };

    this.setAuditor = function (params, callback) {
        setRoleToUser(params, callback);
    };

    this.removeAuditor = function (params, callback) {
        removeRoleFromUser(params, callback);
    };

    this.clearChatServerCaches = clearChatServerCaches;

    this.getChatState = function () {
        return chatFullStateObject;
    };

    this.reconnect = function () {
        asyncClient.reconnectSocket();
    };

    this.setToken = function (newToken) {
        if (typeof newToken != 'undefined') {
            app.sdkParams.token = newToken;
            app.chatEvents.updateToken(app.sdkParams.token);
            if (!app.store.user.get() || !app.store.user.get().id) {
                getUserAndUpdateSDKState();
            }
        }
    };

    this.generateUUID = Utility.generateUUID;

    let requestLimiter = {
        pingChatServer: {
            limit: 1000 * 60,
            lastTime: 0
        }
    };

    this.pingChatServer = function (callback) {
        if (requestLimiter.pingChatServer.lastTime < (new Date().getTime() - (requestLimiter.pingChatServer.limit))) {
            requestLimiter.pingChatServer.lastTime = new Date().getTime();
            app.messenger.sendMessage({
                chatMessageVOType: chatMessageVOTypes.PING,
                pushMsgType: 3,
                uniqueId: Utility.generateUUID()
            }, function (result) {
                app.store.messagesCallbacks[result.uniqueId] && delete app.store.messagesCallbacks[result.uniqueId]
                callback && callback(result);
            });
        } else {
            let result = {
                hasError: true,
                errorCode: null,
                errorMessage: CHAT_ERRORS[6800].replace("{time}", requestLimiter.pingChatServer.limit)
            }
            callback && callback(result)
        }
    };

    /** Call public methods */
    this.startRecordingCall = function (params, callback) {
        var recordCallData = {
            chatMessageVOType: chatMessageVOTypes.RECORD_CALL,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token,
            content: {}
        };

        if (params) {
            if (typeof +params.callId === 'number' && params.callId > 0) {
                recordCallData.subjectId = +params.callId;
            } else {
                // app.chatEvents.fireEvent('error', {
                //     code: 999,
                //     message: 'Invalid Call id!'
                // });
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                return;
            }

            if (params.destinated === true) {
                recordCallData.chatMessageVOType = chatMessageVOTypes.DESTINATED_RECORD_CALL;
                recordCallData.content.recordType = typeof +params.recordType === 'number' ? params.recordType : 1;
                recordCallData.content.tags = Array.isArray(params.tags) ? params.tags : null;
                recordCallData.content.threadId = typeof +params.threadId === 'number' ? params.threadId : null;
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to Record call!'
            });
            return;
        }

        return app.messenger.sendMessage(recordCallData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.stopRecordingCall = function (params, callback) {
        var stopRecordingCallData = {
            chatMessageVOType: chatMessageVOTypes.END_RECORD_CALL,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof +params.callId === 'number' && params.callId > 0) {
                stopRecordingCallData.subjectId = +params.callId;
            } else {
                // app.chatEvents.fireEvent('error', {
                //     code: 999,
                //     message: 'Invalid Call id!'
                // });
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                return;
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to Stop Recording the call!'
            });
            return;
        }

        return app.messenger.sendMessage(stopRecordingCallData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.startCall = function (params, callback) {
        var startCallData = {
            chatMessageVOType: chatMessageVOTypes.CALL_REQUEST,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        }, content = {
            creatorClientDto: {}
        };

        if (params) {
            /*if (typeof params.type === 'string' && callTypes.hasOwnProperty(params.type.toUpperCase())) {
                    content.type = callTypes[params.type.toUpperCase()];
                } else {
                    content.type = 0x0; // Defaults to AUDIO Call
                }*/

            content.type = 0x0;

            //TODO: Check for mute
            content.creatorClientDto.mute = true;//(params.mute && typeof params.mute === 'boolean') ? params.mute : false;
            content.mute = true;//(params.mute && typeof params.mute === 'boolean') ? params.mute : false;

            if (params.clientType
                && typeof params.clientType === 'string'
                && callClientType[params.clientType.toUpperCase()] > 0) {
                content.creatorClientDto.clientType = callClientType[params.clientType.toUpperCase()];
            } else {
                content.creatorClientDto.clientType = callClientType.NODE;
            }

            if (typeof +params.threadId === 'number' && +params.threadId > 0) {
                content.threadId = +params.threadId;
            } else {
                if (Array.isArray(params.invitees)) {
                    content.invitees = [];//params.invitees;
                    for (var i = 0; i < params.invitees.length; i++) {
                        var tempInvitee = params.invitees[i];

                        if (tempInvitee && typeof tempInvitee.idType === "string") {
                            tempInvitee.idType = inviteeVOidTypes[tempInvitee.idType];
                            content.invitees.push(tempInvitee);
                        }
                    }
                } else {
                    app.errorHandler.raiseError(errorList.INVITEES_LIST_REQUIRED, callback, true, {});
                    // app.chatEvents.fireEvent('error', {
                    //     code: 999,
                    //     message: 'Invitees list is empty! Send an array of invitees to start a call with, Or send a Thread Id to start a call with current participants'
                    // });
                    return;
                }
            }

            if (params.threadInfo
                && (params.threadInfo.metadata
                    || params.threadInfo.uniqueName)
            ) {
                content.createCallThreadRequest = params.threadInfo
            }

            startCallData.content = JSON.stringify(content);
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to start call!'
            });
            return;
        }

        callRequestController.callRequestReceived = true;
        callRequestController.callEstablishedInMySide = true;

        if (callNoAnswerTimeout) {
            //TODO: Remove timeout when call ends fast
            setTimeout(function (metaData) {
                //Reject the call if participant didn't answer
                if (!callStopQueue.callStarted) {
                    app.chatEvents.fireEvent("callEvents", {
                        type: "CALL_NO_ANSWER_TIMEOUT",
                        message: "Call request timed out, Participant(s) didn't answer",
                    });

                    metaData.callInstance.rejectCall({
                        callId: currentCallId
                    });
                }
            }, callNoAnswerTimeout, {callInstance: currentModuleInstance});
        }

        return app.messenger.sendMessage(startCallData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.startGroupCall = function (params, callback) {
        var startCallData = {
            chatMessageVOType: chatMessageVOTypes.GROUP_CALL_REQUEST,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        }, content = {
            creatorClientDto: {}
        };

        if (params) {
            /*if (typeof params.type === 'string' && callTypes.hasOwnProperty(params.type.toUpperCase())) {
                    content.type = callTypes[params.type.toUpperCase()];
                } else {
                    content.type = 0x0; // Defaults to AUDIO Call
                }*/
            content.type = 0x0;

            content.creatorClientDto.mute = true;//(typeof params.mute === 'boolean') ? params.mute : false;
            content.mute = true;//(params.mute && typeof params.mute === 'boolean') ? params.mute : false;

            if (params.clientType && typeof params.clientType === 'string' && callClientType[params.clientType.toUpperCase()] > 0) {
                content.creatorClientDto.clientType = callClientType[params.clientType.toUpperCase()];
            } else {
                content.creatorClientDto.clientType = callClientType.NODE;
            }

            if (typeof +params.threadId === 'number' && params.threadId > 0) {
                content.threadId = +params.threadId;
            } else {
                if (Array.isArray(params.invitees)) {
                    content.invitees = [];//params.invitees;
                    for (var i = 0; i < params.invitees.length; i++) {
                        var tempInvitee = params.invitees[i];

                        if (tempInvitee && typeof tempInvitee.idType === "string") {
                            tempInvitee.idType = inviteeVOidTypes[tempInvitee.idType];
                            content.invitees.push(tempInvitee);
                        }
                    }
                } else {

                    app.errorHandler.raiseError(errorList.INVITEES_LIST_REQUIRED, callback, true, {});
                    return;
                }
            }

            if (params.threadInfo
                && (params.threadInfo.title
                    || params.threadInfo.description
                    || params.threadInfo.metadata
                    || params.threadInfo.uniqueName)
            ) {
                content.createCallThreadRequest = params.threadInfo
            }

            startCallData.content = JSON.stringify(content);
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to start call!'
            });
            return;
        }

        callRequestController.callRequestReceived = true;
        callRequestController.callEstablishedInMySide = true;

        return app.messenger.sendMessage(startCallData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.callReceived = callReceived;

    this.terminateCall = function (params, callback) {
        var terminateCallData = {
            chatMessageVOType: chatMessageVOTypes.TERMINATE_CALL,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        }, content = {};

        if (params) {
            if (typeof +params.callId === 'number' && params.callId > 0) {
                terminateCallData.subjectId = +params.callId;
            } else {
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                return;
            }

            terminateCallData.content = JSON.stringify(content);
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to terminate the call!'
            });
            return;
        }

        return app.messenger.sendMessage(terminateCallData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.acceptCall = function (params, callback) {
        var acceptCallData = {
            chatMessageVOType: chatMessageVOTypes.ACCEPT_CALL,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        }, content = {};

        if (params) {
            if (typeof +params.callId === 'number' && params.callId > 0) {
                acceptCallData.subjectId = +params.callId;
            } else {
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                return;
            }

            content.mute = true;//(typeof params.mute === 'boolean') ? params.mute : false;

            content.video = false;//(typeof params.video === 'boolean') ? params.video : false;

            content.videoCall = false;// content.video;

            if (params.clientType && typeof params.clientType === 'string' && callClientType[params.clientType.toUpperCase()] > 0) {
                content.clientType = callClientType[params.clientType.toUpperCase()];
            } else {
                content.clientType = callClientType.NODE;
            }

            acceptCallData.content = JSON.stringify(content);

            if (params.joinCall) {
                callRequestController.callRequestReceived = true;
                currentCallId = params.callId;
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to accept the call!'
            });
            return;
        }
        callRequestController.callEstablishedInMySide = true;
        return app.messenger.sendMessage(acceptCallData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.rejectCall = this.cancelCall = function (params, callback) {
        var rejectCallData = {
            chatMessageVOType: chatMessageVOTypes.REJECT_CALL,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof +params.callId === 'number' && params.callId > 0) {
                rejectCallData.subjectId = +params.callId;
            } else {
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                // app.chatEvents.fireEvent('error', {
                //     code: 999,
                //     message: 'Invalid call id!'
                // });
                return;
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to reject the call!'
            });
            return;
        }

        return app.messenger.sendMessage(rejectCallData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.endCall = endCall;

    this.getCallsList = function (params, callback) {
        var getCallListData = {
            chatMessageVOType: chatMessageVOTypes.GET_CALLS,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        }, content = {};

        if (params) {
            if (typeof params.count === 'number' && params.count >= 0) {
                content.count = +params.count;
            } else {
                content.count = 50;
            }

            if (typeof params.offset === 'number' && params.offset >= 0) {
                content.offset = +params.offset;
            } else {
                content.offset = 0;
            }

            if (typeof params.creatorCoreUserId === 'number' && params.creatorCoreUserId > 0) {
                content.creatorCoreUserId = +params.creatorCoreUserId;
            }

            if (typeof params.creatorSsoId === 'number' && params.creatorSsoId > 0) {
                content.creatorSsoId = +params.creatorSsoId;
            }

            if (typeof params.name === 'string') {
                content.name = params.name;
            }

            if (typeof params.type === 'string' && callTypes.hasOwnProperty(params.type.toUpperCase())) {
                content.type = callTypes[params.type.toUpperCase()];
            }

            if (Array.isArray(params.callIds)) {
                content.callIds = params.callIds;
            }

            if (typeof params.threadId === 'number' && +params.threadId > 0) {
                content.threadId = +params.threadId;
            }

            if (typeof params.contactType === 'string') {
                content.contactType = params.contactType;
            }

            if (typeof params.uniqueId === 'string') {
                content.uniqueId = params.uniqueId;
            }

            getCallListData.content = JSON.stringify(content);
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to End the call!'
            });
            return;
        }

        return app.messenger.sendMessage(getCallListData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.getCallsToJoin = function (params, callback) {
        var getCallListData = {
            chatMessageVOType: chatMessageVOTypes.GET_CALLS_TO_JOIN,
            pushMsgType: 3,
            token: app.sdkParams.token
        }, content = {};

        if (params) {
            if (typeof params.count === 'number' && params.count >= 0) {
                content.count = +params.count;
            } else {
                content.count = 50;
            }

            if (typeof params.offset === 'number' && params.offset >= 0) {
                content.offset = +params.offset;
            } else {
                content.offset = 0;
            }

            if (typeof params.creatorSsoId === 'number' && params.creatorSsoId > 0) {
                content.creatorSsoId = +params.creatorSsoId;
            }

            if (typeof params.name === 'string') {
                content.name = params.name;
            }

            if (typeof params.type === 'string' && callTypes.hasOwnProperty(params.type.toUpperCase())) {
                content.type = callTypes[params.type.toUpperCase()];
            }

            if (Array.isArray(params.threadIds)) {
                content.threadIds = params.threadIds;
            }

            if (typeof params.uniqueId === 'string') {
                content.uniqueId = params.uniqueId;
            }

            getCallListData.content = JSON.stringify(content);
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] Invalid params'
            });
            return;
        }

        return app.messenger.sendMessage(getCallListData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.getCallParticipants = function (params, callback) {
        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.ACTIVE_CALL_PARTICIPANTS,
            typeCode: params.typeCode,
            content: {}
        };

        if (params) {
            if (isNaN(params.callId)) {
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                return;
            } else {
                var callId = +params.callId;
                sendMessageParams.subjectId = callId;

                var offset = (parseInt(params.offset) > 0)
                        ? parseInt(params.offset)
                        : 0,
                    count = (parseInt(params.count) > 0)
                        ? parseInt(params.count)
                        : config.getHistoryCount;

                sendMessageParams.content.count = count;
                sendMessageParams.content.offset = offset;

                return app.messenger.sendMessage(sendMessageParams, {
                    onResult: function (result) {
                        var returnData = {
                            hasError: result.hasError,
                            cache: false,
                            errorMessage: result.errorMessage,
                            errorCode: result.errorCode
                        };

                        if (!returnData.hasError) {
                            var messageContent = result.result,
                                messageLength = messageContent.length,
                                resultData = {
                                    participants: reformatCallParticipants(messageContent),
                                    contentCount: result.contentCount,
                                    hasNext: (sendMessageParams.content.offset + sendMessageParams.content.count < result.contentCount && messageLength > 0),
                                    nextOffset: sendMessageParams.content.offset * 1 + messageLength * 1
                                };

                            returnData.result = resultData;
                        }

                        callback && callback(returnData);
                        /**
                         * Delete callback so if server pushes response before
                         * cache, cache won't send data again
                         */
                        callback = undefined;

                        if (!returnData.hasError) {
                            app.chatEvents.fireEvent('callEvents', {
                                type: 'CALL_PARTICIPANTS_LIST_CHANGE',
                                threadId: callId,
                                result: returnData.result
                            });
                        }
                    }
                });
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to Get Call Participants!'
            });
            return;
        }
    };

    this.addCallParticipants = function (params, callback) {
        /**
         * + AddCallParticipantsRequest     {object}
         *    - subjectId                   {int}
         *    + content                     {list} List of CONTACT IDs or inviteeVO Objects
         *    - uniqueId                    {string}
         */

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.ADD_CALL_PARTICIPANT,
            typeCode: params.typeCode,
            content: []
        };

        if (params) {
            if (typeof params.callId === 'number' && params.callId > 0) {
                sendMessageParams.subjectId = params.callId;
            }

            if (Array.isArray(params.contactIds)) {
                sendMessageParams.content = params.contactIds;
            }

            if (Array.isArray(params.usernames)) {
                sendMessageParams.content = [];
                for (var i = 0; i < params.usernames.length; i++) {
                    sendMessageParams.content.push({
                        id: params.usernames[i],
                        idType: inviteeVOidTypes.TO_BE_USER_USERNAME
                    });
                }
            }

            if (Array.isArray(params.coreUserids)) {
                sendMessageParams.content = [];
                for (var i = 0; i < params.coreUserids.length; i++) {
                    sendMessageParams.content.push({
                        id: params.coreUserids[i],
                        idType: inviteeVOidTypes.TO_BE_CORE_USER_ID
                    });
                }
            }
        }

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };
                if (!returnData.hasError) {
                    // TODO : What is the result?!
                    var messageContent = result.result;
                    returnData.result = messageContent;
                }
                callback && callback(returnData);
            }
        });
    };

    this.removeCallParticipants = function (params, callback) {
        /**
         * + removeCallParticipantsRequest     {object}
         *    - subjectId                   {int}
         *    + content                     {list} List of Participants UserIds
         */

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.REMOVE_CALL_PARTICIPANT,
            typeCode: params.typeCode,
            content: []
        };

        if (params) {
            if (typeof params.callId === 'number' && params.callId > 0) {
                sendMessageParams.subjectId = params.callId;
            }

            if (Array.isArray(params.userIds)) {
                sendMessageParams.content = params.userIds;
            }
        }

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };
                if (!returnData.hasError) {
                    // TODO : What is the result?!
                    var messageContent = result.result;
                    returnData.result = messageContent;
                }
                callback && callback(returnData);
            }
        });
    };

    this.muteCallParticipants = function (params, callback) {
        /**
         * + muteCallParticipantsRequest     {object}
         *    - subjectId                   {int}
         *    + content                     {list} List of Participants UserIds
         */

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.MUTE_CALL_PARTICIPANT,
            typeCode: params.typeCode,
            content: []
        };

        if (params) {
            if (typeof params.callId === 'number' && params.callId > 0) {
                sendMessageParams.subjectId = params.callId;
            }

            if (Array.isArray(params.userIds)) {
                sendMessageParams.content = params.userIds;
            }
        }

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };
                if (!returnData.hasError) {
                    // TODO : What is the result?!
                    var messageContent = result.result;
                    returnData.result = messageContent;
                }
                callback && callback(returnData);
            }
        });
    };

    this.unMuteCallParticipants = function (params, callback) {
        /**
         * + unMuteCallParticipantsRequest     {object}
         *    - subjectId                   {int}
         *    + content                     {list} List of Participants UserIds
         */

        var sendMessageParams = {
            chatMessageVOType: chatMessageVOTypes.UNMUTE_CALL_PARTICIPANT,
            typeCode: params.typeCode,
            content: []
        };

        if (params) {
            if (typeof params.callId === 'number' && params.callId > 0) {
                sendMessageParams.subjectId = params.callId;
            }

            if (Array.isArray(params.userIds)) {
                sendMessageParams.content = params.userIds;
            }
        }
        var myId = app.store.user.get().id;

        return app.messenger.sendMessage(sendMessageParams, {
            onResult: function (result) {
                var returnData = {
                    hasError: result.hasError,
                    cache: false,
                    errorMessage: result.errorMessage,
                    errorCode: result.errorCode
                };
                if (!returnData.hasError) {
                    // TODO : What is the result?!
                    var messageContent = result.result;
                    returnData.result = messageContent;
                }
                callback && callback(returnData);
            }
        });
    };

    this.turnOnVideoCall = function (params, callback) {
        var turnOnVideoData = {
            chatMessageVOType: chatMessageVOTypes.TURN_ON_VIDEO_CALL,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof +params.callId === 'number' && params.callId > 0) {
                turnOnVideoData.subjectId = +params.callId;
            } else {
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                // app.chatEvents.fireEvent('error', {
                //     code: 999,
                //     message: 'Invalid call id!'
                // });
                return;
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to turn on the video call!'
            });
            return;
        }

        return app.messenger.sendMessage(turnOnVideoData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.turnOffVideoCall = function (params, callback) {
        var turnOffVideoData = {
            chatMessageVOType: chatMessageVOTypes.TURN_OFF_VIDEO_CALL,
            typeCode: params.typeCode,
            pushMsgType: 3,
            token: app.sdkParams.token
        };

        if (params) {
            if (typeof +params.callId === 'number' && params.callId > 0) {
                turnOffVideoData.subjectId = +params.callId;
            } else {
                app.errorHandler.raiseError(errorList.INVALID_CALLID, callback, true, {});
                return;
            }
        } else {
            app.chatEvents.fireEvent('error', {
                code: 999,
                message: '[SDK] No params have been sent to turn off the video call!'
            });
            return;
        }


        return app.messenger.sendMessage(turnOffVideoData, {
            onResult: function (result) {
                callback && callback(result);
            }
        });
    };

    this.callStop = callStop;

    this.chatStickerTypes = chatStickerTypes;

    this.addReaction = function (params, callback) {
        let sendData = {
            chatMessageVOType: chatMessageVOTypes.ADD_REACTION,
            subjectId: params.threadId,
            typeCode: params.typeCode,
            content: {
                messageId: params.messageId,
                reaction: params.reaction
            },
            token: app.sdkParams.token
        };

        return app.messenger.sendMessage(sendData, function(result){
            callback && callback(result);
        });
    }

    function onAddReaction(uniqueId, messageContent, contentCount) {
        if (app.store.messagesCallbacks[uniqueId]) {
            app.store.messagesCallbacks[uniqueId](Utility.createReturnData(false, '', 0, messageContent, contentCount, uniqueId));
        }

        app.chatEvents.fireEvent('messageEvents', {
            type: 'ADD_REACTION',
            result: messageContent
        });
    }

    this.messageHasReaction = function (
        {
            threadId,
            messageId,
            typeCode
        }, callback) {

        let count = 1,
            offset = 0;

        let sendData = {
            chatMessageVOType: chatMessageVOTypes.REACTION_LIST,
            subjectId: threadId,
            typeCode, //params.typeCode,
            content: {
                messageId,
                count,
                offset
            },
            token: app.sdkParams.token,
            uniqueId: Utility.generateUUID()
        };

        app.messenger.sendMessage(sendData, function (result) {
            if(!result.hasError) {
                callback && callback({hasError: false, hasReaction: !!result.result.reactionVOList.length});
            } else {
                callback && callback(result);

            }
        });

        return {
            uniqueId: sendData.uniqueId
        }
    }

    this.version = function () {
        console.log("%c[SDK] Version: podchat@" + buildConfig.version, "color:green; font-size:13px")
        console.log("%c[SDK] Build date:" + buildConfig.date, "color:green;font-size:13px")
        console.log("%c[SDK] Additional info: " + buildConfig.VersionInfo, "color:green;font-size:13px")
        return buildConfig;
    };

    this.logout = function () {
        // clearChatServerCaches();

        // Delete all event callbacks
        app.chatEvents.clearEventCallbacks();
        app.store.messagesCallbacks = {};
        app.store.sendMessageCallbacks = {};
        app.store.threadCallbacks = {};
        app.messenger.stopChatPing();

        asyncClient.logout();
    };

    this.setUserObject = function (userObject) {
        if (
            !userObject.id || typeof userObject.id != "number"
            || userObject.coreUserId && typeof userObject.coreUserId != "number"
        ) {
            app.errorHandler.raiseError(errorList.INVALID_USER_OBJECT, null, true, {});
        }
        app.store.user.setUser(userObject);
        // app.store.user.get() = userObject;
        app.messenger.chatState = true;
        app.chatEvents.fireEvent('chatReady');
    }

    init();
}

module.exports = Chat;

export default Chat;